﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_DATETIME_H_
#define INCLUDE_NN_NLIB_DATETIME_H_

#include "nn/nlib/Config.h"

#ifdef NLIB_CXX11_STDLIB_CHRONO
#include <chrono>  // NOLINT
#endif

NLIB_NAMESPACE_BEGIN

// writes the duration in milliseconds from 2000/01/01.
NLIB_VIS_PUBLIC errno_t GetNintendoTime(int64_t* t) NLIB_NOEXCEPT NLIB_NONNULL;

inline errno_t GetNintendoTime(int64_t* t) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(t);
    nlib_time time;
    errno_t e = nlib_epochtime(&time);
    if (e != 0) return e;
    // 1970/01/01 -> 2000/01/01(in milliseconds)
    *t = time / 10000 - (10957LL * 24LL * 60LL * 60LL * 1000LL);
    return 0;
}

// returns in milliseconds after the system boot.
inline uint64_t GetTickTime() NLIB_NOEXCEPT {
    nlib_time t;
    (void)nlib_ticktime(&t);
    return static_cast<uint64_t>(t / 10000);
}

// NOTE:
// TimeValue has a 64bit value, and 1 represents 100ns, 1ms is 10000.
// We can represent the time from A.D 1 to 9999 within 64bits.
struct TimeValue NLIB_FINAL {
    NLIB_CEXPR TimeValue() NLIB_NOEXCEPT : tick(0) {}
    NLIB_CEXPR explicit TimeValue(int64_t tick_) NLIB_NOEXCEPT : tick(tick_) {}

    TimeValue& operator+=(const TimeValue& rhs) NLIB_NOEXCEPT {
        tick += rhs.tick;
        return *this;
    }
    TimeValue& operator-=(const TimeValue& rhs) NLIB_NOEXCEPT {
        tick -= rhs.tick;
        return *this;
    }
    void Abs() NLIB_NOEXCEPT {
        if (tick < 0) tick = -tick;
    }
    TimeValue operator-() const NLIB_NOEXCEPT { return TimeValue(-tick); }
    TimeValue operator+() const NLIB_NOEXCEPT { return *this; }

 public:
    int64_t tick;  // nlib_time, nlib_duration
};

inline bool operator==(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    return lhs.tick == rhs.tick;
}
inline bool operator<(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    return lhs.tick < rhs.tick;
}
inline bool operator!=(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}
inline bool operator>(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    return rhs < lhs;
}
inline bool operator<=(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    return !(lhs > rhs);
}
inline bool operator>=(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    return !(lhs < rhs);
}

inline TimeValue operator+(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    TimeValue result(lhs);
    result += rhs;
    return result;
}

inline TimeValue operator-(const TimeValue& lhs, const TimeValue& rhs) NLIB_NOEXCEPT {
    TimeValue result(lhs);
    result -= rhs;
    return result;
}

// TimeValue(tick_1d) is 1 day
NLIB_CEXPR const int64_t tick_1d = (10LL * 1000 * 1000 * 60 * 60 * 24);

class TimeSpan NLIB_FINAL {
 public:
    NLIB_CEXPR TimeSpan() NLIB_NOEXCEPT : m_Value(0) {}
    explicit TimeSpan(const TimeValue& rhs) NLIB_NOEXCEPT : m_Value(rhs) {}
    NLIB_CEXPR explicit TimeSpan(int days) NLIB_NOEXCEPT
        : m_Value(static_cast<int64_t>(days) * tick_1d) {}
    NLIB_CEXPR TimeSpan(int days, int seconds) NLIB_NOEXCEPT
        : m_Value(days* tick_1d + seconds*(10LL * 1000 * 1000)) {}
    // TimeSpan(0, 0, 1) for 1msec
    NLIB_CEXPR TimeSpan(int days, int seconds,
                        int milliseconds, int microseconds = 0) NLIB_NOEXCEPT
        : m_Value(days* tick_1d + seconds*(10LL * 1000 * 1000) + milliseconds*(10LL * 1000) +
                  microseconds * 10LL) {}

    NLIB_VIS_PUBLIC void Get(int* days, int* seconds,
                             int* milliseconds, int* microseconds) NLIB_NOEXCEPT;

    TimeValue ToTimeValue() const NLIB_NOEXCEPT { return m_Value; }
    int ToDays() const NLIB_NOEXCEPT {
        return static_cast<int>(m_Value.tick / (10LL * 1000 * 1000 * 60 * 60 * 24));
    }
    int ToHours() const NLIB_NOEXCEPT {
        return static_cast<int>(m_Value.tick / (10LL * 1000 * 1000 * 60 * 60));
    }
    int64_t ToMinutes() const NLIB_NOEXCEPT {
        return m_Value.tick / (10 * 1000 * 1000 * 60);
    }
    int64_t ToSeconds() const NLIB_NOEXCEPT {
        return m_Value.tick / (10 * 1000 * 1000);
    }
    int64_t ToMilliSeconds() const NLIB_NOEXCEPT {
        return m_Value.tick / (10 * 1000);
    }
    int64_t ToMicroSeconds() const NLIB_NOEXCEPT {
        return m_Value.tick / (10);
    }
#ifdef NLIB_HAS_STDHEADER_TIME
    template <class TIMEVAL>
    void ToTimeVal(TIMEVAL* tv) const NLIB_NOEXCEPT {
        int64_t usec = ToMicroSeconds();
        int64_t sec = usec / (1000 * 1000);
        tv->tv_sec = static_cast<time_t>(sec);
        tv->tv_usec = static_cast<int32_t>(usec - sec * (1000 * 1000));
    }
    template <class TIMESPEC>
    void ToTimeSpec(TIMESPEC* tp) const NLIB_NOEXCEPT {
        int64_t usec = ToMicroSeconds();
        int64_t sec = usec / (1000 * 1000);
        tp->tv_sec = static_cast<time_t>(sec);
        tp->tv_nsec = static_cast<long>(usec - sec * (1000 * 1000)) * 1000;  // NOLINT
    }
#endif
#ifdef NLIB_CXX11_STDLIB_CHRONO
    template <class Rep, class Period>
    TimeSpan(const std::chrono::duration<Rep, Period>& rhs) {  // NOLINT
        auto tmp = std::chrono::duration_cast<std::chrono::microseconds>(rhs);
        m_Value.tick = static_cast<nlib_duration>(tmp.count() * 10);
    }
    template <class T>
    T ToChrono() const {
        auto tmp = std::chrono::microseconds(m_Value.tick / 10);
        return std::chrono::duration_cast<T>(tmp);
    }
#endif

 public:
    TimeSpan& operator+=(const TimeSpan& rhs) NLIB_NOEXCEPT {
        m_Value += rhs.m_Value;
        return *this;
    }
    TimeSpan& operator-=(const TimeSpan& rhs) NLIB_NOEXCEPT {
        m_Value -= rhs.m_Value;
        return *this;
    }
    TimeSpan& operator*=(int i) NLIB_NOEXCEPT {
        m_Value.tick *= i;
        return *this;
    }
    TimeSpan& operator*=(double d) NLIB_NOEXCEPT {
        m_Value.tick = static_cast<int64_t>(m_Value.tick * d);
        return *this;
    }
    void Abs() NLIB_NOEXCEPT { m_Value.Abs(); }
    TimeSpan operator-() const NLIB_NOEXCEPT { return TimeSpan(-m_Value); }
    TimeSpan operator+() const NLIB_NOEXCEPT { return *this; }
    bool operator==(const TimeSpan& rhs) const NLIB_NOEXCEPT {
        return m_Value == rhs.m_Value;
    }
    bool operator!=(const TimeSpan& rhs) const NLIB_NOEXCEPT {
        return m_Value != rhs.m_Value;
    }
    bool operator<(const TimeSpan& rhs) const NLIB_NOEXCEPT {
        return m_Value < rhs.m_Value;
    }
    bool operator>(const TimeSpan& rhs) const NLIB_NOEXCEPT {
        return m_Value > rhs.m_Value;
    }
    bool operator<=(const TimeSpan& rhs) const NLIB_NOEXCEPT {
        return m_Value <= rhs.m_Value;
    }
    bool operator>=(const TimeSpan& rhs) const NLIB_NOEXCEPT {
        return m_Value >= rhs.m_Value;
    }

 private:
    TimeValue m_Value;
};

inline TimeSpan operator+(const TimeSpan& lhs, const TimeSpan& rhs) NLIB_NOEXCEPT {
    TimeSpan result(lhs);
    result += rhs;
    return result;
}

inline TimeSpan operator-(const TimeSpan& lhs, const TimeSpan& rhs) NLIB_NOEXCEPT {
    TimeSpan result(lhs);
    result -= rhs;
    return result;
}

inline TimeSpan operator*(int i, const TimeSpan& rhs) NLIB_NOEXCEPT {
    TimeSpan result(rhs);
    result *= i;
    return result;
}

inline TimeSpan operator*(double d, const TimeSpan& rhs) NLIB_NOEXCEPT {
    TimeSpan result(rhs);
    result *= d;
    return result;
}

inline TimeSpan operator*(const TimeSpan& lhs, int i) NLIB_NOEXCEPT {
    TimeSpan result(lhs);
    result *= i;
    return result;
}

inline TimeSpan operator*(const TimeSpan& lhs, double d) NLIB_NOEXCEPT {
    TimeSpan result(lhs);
    result *= d;
    return result;
}

// has year/month/day/hour/min/sec/msec/usec
struct DateTimeParams {
    int year;
    int month;
    int day;
    int hour;
    int min;
    int sec;
    int msec;
    int usec;
};

class DateTime NLIB_FINAL {
 public:
    // WEEK_SUNDAY -> WEEK_SATURDAY
    enum Week {
        WEEK_SUNDAY = 0,
        WEEK_MONDAY,
        WEEK_TUESDAY,
        WEEK_WEDNESDAY,
        WEEK_THURSDAY,
        WEEK_FRIDAY,
        WEEK_SATURDAY,
        WEEK_MAX
    };

 public:
    NLIB_CEXPR DateTime() NLIB_NOEXCEPT : m_Value(INT64_MIN) {}
    // You can omit min/sec/msec/usec.
    NLIB_VIS_PUBLIC errno_t Init(int year, int month, int day, int hour = 0,
                                 int min = 0, int sec = 0, int msec = 0,
                                 int usec = 0) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Init(const DateTimeParams& rhs) NLIB_NOEXCEPT;
    // nlib_epochtime(&t); dt.Init(TimeValue(t)); for example
    errno_t Init(const TimeValue& tv) NLIB_NOEXCEPT {
        m_Value = tv;
        return IsValid() ? 0 : EINVAL;
    }
    // you can pass null pointers
    NLIB_VIS_PUBLIC errno_t GetDay(int* year, int* month, int* day) const NLIB_NOEXCEPT;
    // you can pass null pointers
    NLIB_VIS_PUBLIC errno_t GetTime(int* hour, int* min, int* sec,
                                    int* msec, int* usec) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Get(DateTimeParams* rhs) const NLIB_NOEXCEPT;

    TimeValue ToTimeValue() const NLIB_NOEXCEPT { return m_Value; }
    NLIB_VIS_PUBLIC bool IsValid() const NLIB_NOEXCEPT;

    // 01-01 becomes 0.
    NLIB_VIS_PUBLIC errno_t GetDayOfYear(int* nth) const NLIB_NOEXCEPT;
    // Sunday -> 0(WEEK_SUNDAY), Saturday -> 6(WEEK_SATURDAY)
    NLIB_VIS_PUBLIC errno_t GetDayOfWeek(Week* week) const NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t Add(const TimeSpan& rhs) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddYears(int value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddMonths(int value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddDays(double value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddHours(double value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddMinutes(double value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddSeconds(double value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddMilliSeconds(double value) NLIB_NOEXCEPT;
    // returns EINVAL on failure
    NLIB_VIS_PUBLIC errno_t AddMicroSeconds(double value) NLIB_NOEXCEPT;

    // "1994-11-05T13:15:30Z" for example
    NLIB_VIS_PUBLIC errno_t ToW3cDtf(char(&str)[32], const TimeSpan& zone) const NLIB_NOEXCEPT;
    errno_t ToW3cDtf(char (&str)[32]) const NLIB_NOEXCEPT { return ToW3cDtf(str, TimeSpan()); }

    // "Tue, 02 Sep 2014 10:51:34 +0900" for example
    NLIB_VIS_PUBLIC errno_t ToRfc2822(char(&str)[32], const TimeSpan& zone) const NLIB_NOEXCEPT;
    errno_t ToRfc2822(char (&str)[32]) const NLIB_NOEXCEPT { return ToRfc2822(str, TimeSpan()); }

    // "Wed Jan 02 02:03:45 1980\n" for example
    NLIB_VIS_PUBLIC errno_t ToAscTime(char(&str)[26]) const NLIB_NOEXCEPT;

    NLIB_SAFE_BOOL(DateTime, IsValid())

 public:
    NLIB_VIS_PUBLIC DateTime& operator+=(const TimeSpan& rhs) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC DateTime& operator-=(const TimeSpan& rhs) NLIB_NOEXCEPT;
    bool operator==(const DateTime& rhs) const NLIB_NOEXCEPT {
        return m_Value == rhs.m_Value;
    }
    bool operator!=(const DateTime& rhs) const NLIB_NOEXCEPT {
        return m_Value != rhs.m_Value;
    }
    bool operator<(const DateTime& rhs) const NLIB_NOEXCEPT {
            return m_Value < rhs.m_Value;
    }
    bool operator>(const DateTime& rhs) const NLIB_NOEXCEPT {
                return m_Value > rhs.m_Value;
    }
    bool operator<=(const DateTime& rhs) const NLIB_NOEXCEPT {
                    return m_Value <= rhs.m_Value;
    }
    bool operator>=(const DateTime& rhs) const NLIB_NOEXCEPT {
                        return m_Value >= rhs.m_Value;
    }

#ifdef NLIB_CXX11_STDLIB_CHRONO
    template <class Clock, class Duration>
    errno_t Init(const std::chrono::time_point<Clock, Duration>& rhs) {
        auto tmp = std::chrono::duration_cast<std::chrono::microseconds>(rhs.time_since_epoch());
        m_Value.tick = static_cast<nlib_duration>(tmp.count() * 10);
        return 0;
    }
    template <class T>
    T ToChrono() const {
        auto tmp = std::chrono::microseconds(m_Value.tick / 10);
        return T(tmp);
    }
#endif

 public:
    static bool IsLeapYear(int year) NLIB_NOEXCEPT {
        return ((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0);
    }
    NLIB_VIS_PUBLIC static errno_t GetNow(DateTime* p) NLIB_NOEXCEPT;
    // January has 31 days for example
    NLIB_VIS_PUBLIC static int GetDaysInMonth(int year, int month) NLIB_NOEXCEPT;
    // can parse W3CDTF, RFC2822, or ASCTIME
    NLIB_VIS_PUBLIC static errno_t Parse(const char* str, DateTime* dt,
                                         TimeSpan* delta) NLIB_NOEXCEPT NLIB_NONNULL;

 private:
    TimeValue m_Value;
};

inline errno_t DateTime::Init(const DateTimeParams& rhs) NLIB_NOEXCEPT {
    return this->Init(rhs.year, rhs.month, rhs.day, rhs.hour, rhs.min, rhs.sec, rhs.msec, rhs.usec);
}

inline errno_t DateTime::Add(const TimeSpan& rhs) NLIB_NOEXCEPT {
    m_Value += rhs.ToTimeValue();
    return IsValid() ? 0 : ERANGE;
}

inline errno_t DateTime::AddDays(double value) NLIB_NOEXCEPT {
    TimeSpan d(1);
    d *= value;
    return this->Add(d);
}

inline errno_t DateTime::AddHours(double value) NLIB_NOEXCEPT {
    TimeSpan d(0, 60 * 60);
    d *= value;
    return this->Add(d);
}

inline errno_t DateTime::AddMinutes(double value) NLIB_NOEXCEPT {
    TimeSpan d(0, 60);
    d *= value;
    return this->Add(d);
}

inline errno_t DateTime::AddSeconds(double value) NLIB_NOEXCEPT {
    TimeSpan d(0, 1);
    d *= value;
    return this->Add(d);
}

inline errno_t DateTime::AddMilliSeconds(double value) NLIB_NOEXCEPT {
    TimeSpan d(0, 0, 1);
    d *= value;
    return this->Add(d);
}

inline errno_t DateTime::AddMicroSeconds(double value) NLIB_NOEXCEPT {
    TimeSpan d(0, 0, 0, 1);
    d *= value;
    return this->Add(d);
}

inline DateTime& DateTime::operator+=(const TimeSpan& rhs) NLIB_NOEXCEPT {
    this->Add(rhs);
    return *this;
}

inline DateTime& DateTime::operator-=(const TimeSpan& rhs) NLIB_NOEXCEPT {
    *this += (-rhs);
    return *this;
}

inline DateTime operator+(const DateTime& lhs, const TimeSpan& rhs) NLIB_NOEXCEPT {
    DateTime tmp(lhs);
    tmp += rhs;
    return tmp;
}

inline DateTime operator-(const DateTime& lhs, const TimeSpan& rhs) NLIB_NOEXCEPT {
    DateTime tmp(lhs);
    tmp -= rhs;
    return tmp;
}

inline TimeSpan operator-(const DateTime& lhs, const DateTime& rhs) NLIB_NOEXCEPT {
    return TimeSpan(lhs.ToTimeValue() - rhs.ToTimeValue());
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_DATETIME_H_
