﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_INPUTSTREAM_H_
#define INCLUDE_NN_NLIB_INPUTSTREAM_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

// code snippets:
//   InputStream& stream = 'stream that inherits InputStream';
//   if (!stream.Read(dataptr, nbyte)) { error = stream.GetErrorValue(); ..... }
//   while ((c = stream.Read()) >= 0) { c is [0x00 - 0xFF] }
//   if (!stream) { error = stream.GetErrorValue(); ..... }
class NLIB_VIS_PUBLIC InputStream {
 public:
    NLIB_CEXPR InputStream() NLIB_NOEXCEPT : m_Pos(0),
                                             m_Cur(0),
                                             m_End(0),
                                             m_Buf(NULL),
                                             m_BufSize(0),
                                             m_ErrorId(0) {}
    virtual ~InputStream() NLIB_NOEXCEPT {
        // NOTE:
        // cannot call Close_() from this destructor.
        // the destructor of the derived class has to call Close_() if needed.
    }
    // returns true if there has been no errors
    // you can also write the codes such that 'if (!!stream) { ... }'
    bool IsOk() const NLIB_NOEXCEPT { return m_ErrorId == 0; }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return m_ErrorId; }
    // use Pos64() if you need a 64bit value on 32bit arch
    size_t Pos() const NLIB_NOEXCEPT { return static_cast<size_t>(Pos64()); }
    uint64_t Pos64() const NLIB_NOEXCEPT { return m_Pos + m_Cur; }
    bool IsEos() NLIB_NOEXCEPT {
        if (this->IsOk() && m_Cur == m_End) {
            CheckBuffer();
            return this->IsOk() && m_Cur == m_End;
        }
        return false;
    }
    // < 0 if EOS or error, otherwise returns a byte and ++pos
    int Read() NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(m_Cur == m_End)) {
            CheckBuffer();
            if (NLIB_UNLIKELY(m_Cur == m_End)) return -1;
        }
        return m_Buf[m_Cur++];
    }
    // < 0 if EOS or error, otherwise returns a byte and pos is unchanged
    NLIB_CHECK_RESULT int Peek() NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(m_Cur == m_End)) {
            CheckBuffer();
            if (NLIB_UNLIKELY(m_Cur == m_End)) return -1;
        }
        return m_Buf[m_Cur];
    }
    // returns skipped number of bytes
    size_t Skip(size_t nBytes) NLIB_NOEXCEPT;
    size_t Read(void* ptr, size_t nBytes) NLIB_NOEXCEPT NLIB_NONNULL {
#ifndef NLIB_NONNULL_ENABLED
        if (!ptr) {
            this->SetError(EINVAL);
            return 0;
        }
#endif
        if (nBytes > RSIZE_MAX) {  // INT01-C
            this->SetError(EINVAL);
            return 0;
        }
        if (m_Cur + nBytes <= m_End) {
            nlib_memcpy(ptr, nBytes, &m_Buf[m_Cur], nBytes);
            m_Cur += nBytes;
            return nBytes;
        }
        return this->Read_(ptr, nBytes);
    }
    bool Close() NLIB_NOEXCEPT;
    NLIB_SAFE_BOOL(InputStream, IsOk())

 protected:
    void ResetBuffer(void* p, size_t nBytes) NLIB_NOEXCEPT {
        m_Buf = reinterpret_cast<unsigned char*>(p);
        m_BufSize = static_cast<int>(nBytes);
    }
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (m_ErrorId == 0) m_ErrorId = e;
    }

 private:
    virtual size_t FillBuffer_(void* p, size_t nBytes) NLIB_NOEXCEPT = 0;
    virtual bool Close_() NLIB_NOEXCEPT = 0;
    virtual void* GetWorkBuffer_(size_t* nBytes) NLIB_NOEXCEPT;
    size_t Read_(void* ptr, size_t nBytes) NLIB_NOEXCEPT;

    // By default, Skip_() reads data on m_Buf.
    // you can override it if you can implement more efficient code.
    virtual size_t Skip_(size_t nBytes) NLIB_NOEXCEPT;

 private:
    void CheckBuffer() NLIB_NOEXCEPT;
    bool GetWorkBuffer() NLIB_NOEXCEPT;

 private:
    uint64_t m_Pos;
    size_t m_Cur;
    size_t m_End;
    unsigned char* m_Buf;
    size_t m_BufSize;
    mutable errno_t m_ErrorId;
    // m_ErrorId is changed only from derived classes.
    // FillBuffer_() or Close_() changes m_ErrorId in other words.

    NLIB_DISALLOW_COPY_AND_ASSIGN(InputStream);
};

class NLIB_VIS_PUBLIC NullInputStream NLIB_FINAL : public InputStream {
 public:
    NullInputStream() NLIB_NOEXCEPT;
    virtual ~NullInputStream() NLIB_NOEXCEPT {}

 private:
    char m_DummyBuf[256];
    virtual size_t FillBuffer_(void* p, size_t nBytes) NLIB_NOEXCEPT NLIB_OVERRIDE;
    virtual bool Close_() NLIB_NOEXCEPT NLIB_OVERRIDE { return true; }
    virtual size_t Skip_(size_t nBytes) NLIB_NOEXCEPT NLIB_OVERRIDE;
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_INPUTSTREAM_H_
