﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_NLIST_H_
#define INCLUDE_NN_NLIB_NLIST_H_

#include <string.h>
#include <utility>
#include <algorithm>
#include <iterator>
#include <memory>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/TypeTraits.h"

NLIB_NAMESPACE_BEGIN

template <class T, class AL>
class Nlist;

namespace nlist {

class NLIB_VIS_PUBLIC NlistBase {
 protected:
    struct Blk {
        Blk* next;
        void* item;
    };

 protected:
    NLIB_CEXPR NlistBase() NLIB_NOEXCEPT : m_CurBlk(NULL),
                                           m_CurLevel(0),
                                           m_CurIdx(0),
                                           m_FirstBlk(NULL) {}
    static size_t BlkBaseIdx(size_t level) NLIB_NOEXCEPT {
        return 8 * ((1 << level) - 1);
    }
    static size_t BlkSize(size_t level) NLIB_NOEXCEPT {
        return 8 * (1 << level);
    }
    Blk* LastBlock(size_t* level) const NLIB_NOEXCEPT;
    bool ConfirmBack() const NLIB_NOEXCEPT;
    void ClearSimple() NLIB_NOEXCEPT {
        m_CurLevel = 0;
        m_CurIdx = 0;
        m_CurBlk = m_FirstBlk;
    }
    const Blk* AtSub(size_t n, size_t* base) const NLIB_NOEXCEPT;

 protected:
    mutable Blk* m_CurBlk;
    mutable size_t m_CurLevel;
    mutable size_t m_CurIdx;
    mutable Blk* m_FirstBlk;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(NlistBase);
};

template <class T>
class NlistConstIterator;
template <class T>
class NlistIterator;

template <class T>
class NlistBaseT : public NlistBase {
 public:
    typedef T& reference;
    typedef const T& const_reference;
    typedef NlistIterator<T> iterator;
    typedef NlistConstIterator<T> const_iterator;
    typedef size_t size_type;
    typedef T value_type;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef ptrdiff_t difference_type;

 protected:
    NLIB_CEXPR NlistBaseT() {}
    static T* GetItem(Blk* blk) NLIB_NOEXCEPT {
        return reinterpret_cast<T*>(blk->item);
    }
    static const T* GetItem(const Blk* blk) NLIB_NOEXCEPT {
        return reinterpret_cast<const T*>(blk->item);
    }
    T* At(size_type n) const NLIB_NOEXCEPT {
        size_type base;
        const Blk* p = this->AtSub(n, &base);
        return p ? const_cast<T*>(&(GetItem(p)[n - base])) : NULL;
    }
    iterator Begin() NLIB_NOEXCEPT { return iterator(this, TrueType()); }
    iterator End() NLIB_NOEXCEPT { return iterator(this, FalseType()); }
    const_iterator Begin() const NLIB_NOEXCEPT {
        return const_iterator(this, TrueType());
    }
    const_iterator End() const NLIB_NOEXCEPT {
        return const_iterator(this, FalseType());
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(NlistBaseT);
    friend class NlistConstIterator<T>;
    friend class NlistIterator<T>;
};

template <class T>
class NlistConstIterator
    : public std::iterator<std::forward_iterator_tag, T> {
 private:
    typedef NlistBaseT<T> ContainerType;
    typedef typename ContainerType::Blk Blk;

 public:
    typedef typename std::iterator<std::forward_iterator_tag, T> BaseType;
    typedef typename BaseType::iterator_category iterator_category;
    typedef typename BaseType::value_type value_type;
    typedef typename BaseType::difference_type difference_type;
    typedef typename BaseType::pointer pointer;
    typedef typename BaseType::reference reference;

 public:
    NLIB_CEXPR NlistConstIterator() NLIB_NOEXCEPT : m_Idx(0),
                                                    m_ElemPtr(NULL),
                                                    m_BlkEnd(NULL),
                                                    m_Blk(NULL) {}

 protected:
    NlistConstIterator(const ContainerType* p, TrueType) NLIB_NOEXCEPT {
        ContainerType* rhs = const_cast<ContainerType*>(p);
        m_Idx = 0;
        m_Blk = rhs->m_FirstBlk;
        m_ElemPtr = GetItem(m_Blk);
        m_BlkEnd = m_ElemPtr + 8;
    }
    NlistConstIterator(const ContainerType* p, FalseType) NLIB_NOEXCEPT {
        ContainerType* rhs = const_cast<ContainerType*>(p);
        m_Idx = rhs->BlkBaseIdx(rhs->m_CurLevel) + rhs->m_CurIdx;
        m_Blk = rhs->m_CurBlk;
        m_ElemPtr = &(GetItem(m_Blk)[rhs->m_CurIdx]);
        m_BlkEnd = GetItem(m_Blk) + rhs->BlkSize(rhs->m_CurLevel);
        if (m_ElemPtr == m_BlkEnd) this->Normalize();
    }

 public:
    const T& operator*() const NLIB_NOEXCEPT { return *m_ElemPtr; }
    const T* operator->() const NLIB_NOEXCEPT { return m_ElemPtr; }
    NlistConstIterator& operator++() NLIB_NOEXCEPT {
        NLIB_ASSERT(m_ElemPtr != m_BlkEnd);
        ++m_Idx;
        if (++m_ElemPtr == m_BlkEnd) this->Normalize();
        return *this;
    }
    NlistConstIterator operator++(int) NLIB_NOEXCEPT {  // NOLINT
        NlistConstIterator rval(*this);
        this->operator++();
        return rval;
    }
    bool operator==(const NlistConstIterator& rhs) const NLIB_NOEXCEPT {
        return m_Idx == rhs.m_Idx && m_ElemPtr == rhs.m_ElemPtr;
    }
    bool operator!=(const NlistConstIterator& rhs) const NLIB_NOEXCEPT {
        return !operator==(rhs);
    }
    void Advance(difference_type n) NLIB_NOEXCEPT;
    difference_type Distance(NlistConstIterator to) const NLIB_NOEXCEPT {
        return to.m_Idx - m_Idx;
    }

 private:
    static T* GetItem(Blk* blk) NLIB_NOEXCEPT {
        return reinterpret_cast<T*>(blk->item);
    }
    static const T* GetItem(const Blk* blk) NLIB_NOEXCEPT {
        return reinterpret_cast<const T*>(blk->item);
    }
    bool Normalize() NLIB_NOEXCEPT;

 private:
    size_t m_Idx;
    T* m_ElemPtr;
    T* m_BlkEnd;
    Blk* m_Blk;
    friend class NlistBaseT<T>;
};

template <class T>
bool NlistConstIterator<T>::Normalize() NLIB_NOEXCEPT {
    if (m_Blk && m_Blk->next) {
        Blk* next = m_Blk->next;
        size_t next_size = 2 * static_cast<size_t>(m_BlkEnd - GetItem(m_Blk));
        T* next_elemptr = GetItem(next);
        T* next_end = next_elemptr + next_size;
        m_Blk = next;
        m_ElemPtr = next_elemptr;
        m_BlkEnd = next_end;
        return true;
    }
    return false;
}

template <class T>
void NlistConstIterator<T>::Advance(difference_type n) NLIB_NOEXCEPT {
    while (m_ElemPtr + n >= m_BlkEnd) {
        n -= m_BlkEnd - m_ElemPtr;
        m_Idx += m_BlkEnd - m_ElemPtr;
        m_ElemPtr = m_BlkEnd;
        if (!this->Normalize()) return;
    }
    m_ElemPtr += n;
    m_Idx += n;
}

template <class T>
class NlistIterator : public NlistConstIterator<T> {
 private:
    typedef NlistBaseT<T> ContainerType;
    typedef typename ContainerType::Blk Blk;
    typedef NlistConstIterator<T> MyBase;

 public:
    typedef typename NlistConstIterator<T>::BaseType BaseType;
    typedef typename BaseType::iterator_category iterator_category;
    typedef typename BaseType::value_type value_type;
    typedef typename BaseType::difference_type difference_type;
    typedef typename BaseType::pointer pointer;
    typedef typename BaseType::reference reference;

 public:
    NLIB_CEXPR NlistIterator() NLIB_NOEXCEPT : MyBase() {}

 private:
    NlistIterator(const ContainerType* p, TrueType) NLIB_NOEXCEPT : MyBase(p, TrueType()) {}
    NlistIterator(const ContainerType* p, FalseType) NLIB_NOEXCEPT : MyBase(p, FalseType()) {}

 public:
    T& operator*() const NLIB_NOEXCEPT {
        const MyBase* tmp = static_cast<const MyBase*>(this);
        return const_cast<T&>(**tmp);
    }
    T* operator->() const NLIB_NOEXCEPT {
        const MyBase* tmp = static_cast<const MyBase*>(this);
        return const_cast<T*>(tmp->operator->());
    }
    NlistIterator& operator++() NLIB_NOEXCEPT {
        MyBase* tmp = static_cast<MyBase*>(this);
        ++(*tmp);
        return *this;
    }
    NlistIterator operator++(int) NLIB_NOEXCEPT {  // NOLINT
        NlistIterator rval(*this);
        ++(*this);
        return rval;
    }
    bool operator==(const NlistIterator& rhs) const NLIB_NOEXCEPT {
        const MyBase* tmp = static_cast<const MyBase*>(this);
        const MyBase* tmp2 = static_cast<const MyBase*>(&rhs);
        return *tmp == *tmp2;
    }
    bool operator!=(const NlistIterator& rhs) const NLIB_NOEXCEPT {
        return !operator==(rhs);
    }
    difference_type Distance(NlistIterator to) const NLIB_NOEXCEPT {
        const MyBase* tmp = static_cast<const MyBase*>(this);
        const MyBase* tmp2 = static_cast<const MyBase*>(&to);
        return tmp->Distance(*tmp2);
    }

 private:
    friend class NlistBaseT<T>;
};

template<bool is_empty, class AL>
class NlistAlloc {
 public:
    NlistAlloc() NLIB_NOEXCEPT : m_Alloc(AL()) {}
    explicit NlistAlloc(const AL& al) NLIB_NOEXCEPT : m_Alloc(al) {}

 protected:
    AL& _GetNlistAlloc() const NLIB_NOEXCEPT {
        return const_cast<AL&>(m_Alloc);
    }
    void _SwapNlistAlloc(NlistAlloc& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(m_Alloc, rhs.m_Alloc);
    }

 private:
    AL m_Alloc;
};

template<class AL>
class NlistAlloc <true, AL> {
 public:
    NlistAlloc() NLIB_NOEXCEPT {}
    explicit NlistAlloc(const AL& al) NLIB_NOEXCEPT { NLIB_UNUSED(al); }

 protected:
    AL _GetNlistAlloc() const NLIB_NOEXCEPT { return AL(); }
    void _SwapNlistAlloc(NlistAlloc& rhs) NLIB_NOEXCEPT {  // NOLINT
        NLIB_UNUSED(rhs);
    }
};

}  // namespace nlist

template <class T, class AL = std::allocator<char> >
class Nlist : public nlist::NlistBaseT<T>,
              public nlist::NlistAlloc<IsEmpty<AL>::value, AL> {
    // NOTE:
    // To develop custom allocators:
    //
    // Nlist uses only the subset of std::allocator<T>.
    //  1) rebind() is not used. the custom allocator does not have to be a class template.
    //  2) construct(), destroy() are not used.
    //
    // You have to implement:
    //  a) void* allocate(size_t nBytes);
    //  b) void deallocate(void* ptr, size_t nBytes);
    // allocate has to return 8 bytes aligned memory.
    // And the custom allocator cannot have non-static data members.
    /*
    class MyAllocator
    {
    public:
        MyAllocator();
        MyAllocator(const MyAllocator& rhs);
        void* allocate(size_t nBytes, void*);
        void deallocate(void* p, size_t nBytes);
    };
    */

 private:
    typedef typename nlist::NlistBaseT<T> BaseType;
    typedef typename nlist::NlistAlloc<IsEmpty<AL>::value, AL> AllocType;
    using BaseType::m_CurBlk;
    using BaseType::m_CurIdx;
    using BaseType::m_CurLevel;
    using BaseType::m_FirstBlk;
    using BaseType::ConfirmBack;
    using BaseType::ClearSimple;
    using BaseType::BlkBaseIdx;
    using BaseType::BlkSize;
    using BaseType::GetItem;
    using BaseType::Begin;
    using BaseType::End;
    typedef typename BaseType::Blk Blk;

 public:
    typedef typename BaseType::value_type value_type;
    typedef AL allocator_type;
    typedef typename BaseType::size_type size_type;
    typedef typename BaseType::difference_type difference_type;
    typedef typename BaseType::reference reference;
    typedef typename BaseType::reference const_reference;
    typedef typename BaseType::pointer pointer;
    typedef typename BaseType::const_pointer const_pointer;
    typedef typename BaseType::iterator iterator;
    typedef typename BaseType::const_iterator const_iterator;

 public:
    Nlist() NLIB_NOEXCEPT : AllocType(AL()) {}
    explicit Nlist(const AL& al) NLIB_NOEXCEPT : AllocType(al) {}
    ~Nlist() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_2(Nlist, BaseType, AllocType)
    size_type size() const NLIB_NOEXCEPT {
        return BlkBaseIdx(m_CurLevel) + m_CurIdx;
    }
    size_type capacity() const NLIB_NOEXCEPT {
        size_type lv;
        return (NULL != this->LastBlock(&lv)) ? BlkBaseIdx(lv + 1) : 0;
    }
    bool empty() const NLIB_NOEXCEPT { return m_CurLevel == 0 && m_CurIdx == 0; }
    bool resize(size_type n) NLIB_NOEXCEPT {
        return Resize(n, typename IsTriviallyDefaultConstructible<T>::type());
    }
    bool reserve(size_type n) NLIB_NOEXCEPT;
    pointer push_back() {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = MyCtor(p, typename IsTriviallyDefaultConstructible<T>::type());  // may throw
        ++m_CurIdx;
        return ptr;
    }
#ifdef NLIB_CXX11_RVALUE_REFERENCES
    pointer push_back(T&& rhs) NLIB_NOEXCEPT {  // NOLINT
#if defined(NLIB_HAS_NATIVE_TYPETRAITS) && !defined(NLIB_HAS_TR1_TYPETRAITS)
        NLIB_STATIC_ASSERT(std::is_nothrow_move_constructible<T>::value);
#endif
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(std::forward<T>(rhs));
        ++m_CurIdx;
        return ptr;
    }
#endif
    pointer push_back(const T& rhs) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(rhs);  // may throw
        ++m_CurIdx;
        return ptr;
    }
    bool pop_back() NLIB_NOEXCEPT {
        return this->pop_back_(typename IsTriviallyDestructible<T>::type());
    }

    void swap(Nlist& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(m_CurBlk, rhs.m_CurBlk);
        swap(m_CurLevel, rhs.m_CurLevel);
        swap(m_CurIdx, rhs.m_CurIdx);
        swap(m_FirstBlk, rhs.m_FirstBlk);
        AllocType::_SwapNlistAlloc(rhs);
    }
    void clear() NLIB_NOEXCEPT {
        Clear(typename IsTriviallyDestructible<T>::type());
    }
    allocator_type get_allocator() const NLIB_NOEXCEPT {
        return AllocType::_GetNlistAlloc();
    }

    const_reference back() const {
        if (m_CurIdx == 0) ConfirmBack();
        return GetItem(m_CurBlk)[m_CurIdx - 1];
    }
    reference back() {
        if (m_CurIdx == 0) ConfirmBack();
        return GetItem(m_CurBlk)[m_CurIdx - 1];
    }
    const_reference front() const { return GetItem(m_FirstBlk)[0]; }
    reference front() { return GetItem(m_FirstBlk)[0]; }
    const_reference operator[](size_type idx) const {
        NLIB_ASSERT(idx < this->size());
        return *this->At(idx);
    }
    reference operator[](size_type idx) {
        NLIB_ASSERT(idx < this->size());
        return *this->At(idx);
    }

    iterator begin() NLIB_NOEXCEPT {
        if (!m_FirstBlk && !ConfirmFirstBlk()) return iterator();
        return Begin();
    }
    const_iterator begin() const NLIB_NOEXCEPT {
        if (!m_FirstBlk && !ConfirmFirstBlk()) return const_iterator();
        return Begin();
    }
    const_iterator cbegin() const NLIB_NOEXCEPT {
        if (!m_FirstBlk && !ConfirmFirstBlk()) return const_iterator();
        return Begin();
    }

    iterator end() NLIB_NOEXCEPT {
        if (!m_FirstBlk && !ConfirmFirstBlk()) return iterator();
        return End();
    }
    const_iterator end() const NLIB_NOEXCEPT {
        if (!m_FirstBlk && !ConfirmFirstBlk()) return const_iterator();
        return End();
    }
    const_iterator cend() const NLIB_NOEXCEPT {
        if (!m_FirstBlk && !ConfirmFirstBlk()) return const_iterator();
        return End();
    }

    void shrink_to_fit() NLIB_NOEXCEPT {
        if (!m_CurBlk) return;
        Blk* p = m_CurBlk->next;
        m_CurBlk->next = NULL;
        DeallocBlkList(p, static_cast<int>(m_CurLevel + 1));
    }
#if defined(NLIB_CXX11_RVALUE_REFERENCES) && defined(NLIB_CXX11_VARIADIC_TEMPLATES)
    template<class... Args>
    pointer EmplaceBack(Args&&... args) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(std::forward<Args>(args)...);  // may throw
        ++m_CurIdx;
        return ptr;
    }
#else
    template <class A1>
    pointer EmplaceBack(const A1& a1) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1));
        ++m_CurIdx;
        return ptr;
    }
    template <class A1, class A2>
    pointer EmplaceBack(const A1& a1, const A2& a2) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2));
        ++m_CurIdx;
        return ptr;
    }
    template <class A1, class A2, class A3>
    pointer EmplaceBack(const A1& a1, const A2& a2, const A3& a3) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2),
                                const_cast<typename RemoveConst<A3>::type&>(a3));
        ++m_CurIdx;
        return ptr;
    }
    template <class A1, class A2, class A3, class A4>
    pointer EmplaceBack(const A1& a1, const A2& a2, const A3& a3, const A4& a4) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2),
                                const_cast<typename RemoveConst<A3>::type&>(a3),
                                const_cast<typename RemoveConst<A4>::type&>(a4));
        ++m_CurIdx;
        return ptr;
    }
    template <class A1, class A2, class A3, class A4, class A5>
    pointer
        EmplaceBack(const A1& a1, const A2& a2, const A3& a3, const A4& a4, const A5& a5) {
        void* p = this->PushBack();
        if (!p) return NULL;
        pointer ptr = new (p) T(const_cast<typename RemoveConst<A1>::type&>(a1),
                                const_cast<typename RemoveConst<A2>::type&>(a2),
                                const_cast<typename RemoveConst<A3>::type&>(a3),
                                const_cast<typename RemoveConst<A4>::type&>(a4),
                                const_cast<typename RemoveConst<A5>::type&>(a5));
        ++m_CurIdx;
        return ptr;
    }
#endif
    void Clobber() NLIB_NOEXCEPT {
        m_CurBlk = NULL;
        m_CurLevel = 0;
        m_CurIdx = 0;
        m_FirstBlk = NULL;
    }

 private:
    // The # of elements in the block increases as 8, 16, 32, 64, ....
    Blk* AllocBlk(size_t n) const NLIB_NOEXCEPT;
    void DeallocBlkList(Blk* p, int lv) const NLIB_NOEXCEPT;
    void DestroyBlk(Blk*, size_t, TrueType) const NLIB_NOEXCEPT {}
    void DestroyBlk(Blk* p, size_t n, FalseType) const NLIB_NOEXCEPT;
    void DestroyBlk(Blk* p, size_t n) const NLIB_NOEXCEPT {
        DestroyBlk(p, n, typename IsTriviallyDestructible<T>::type());
    }
    bool ConfirmFirstBlk() const NLIB_NOEXCEPT;
    void* PushBack() NLIB_NOEXCEPT;
    void Clear(FalseType) NLIB_NOEXCEPT;
    void Clear(TrueType) NLIB_NOEXCEPT { ClearSimple(); }
    bool Resize(size_t n, FalseType) NLIB_NOEXCEPT;
    bool Resize(size_t n, TrueType) NLIB_NOEXCEPT;
    T* MyCtor(void* p, TrueType) NLIB_NOEXCEPT {
        return reinterpret_cast<T*>(p);
    }
    T* MyCtor(void* p, FalseType) {
        T* ptr = new (p) T;  // may throw
        return ptr;
    }
    bool pop_back_(FalseType) NLIB_NOEXCEPT {
        if (m_CurIdx == 0) {
            if (!ConfirmBack()) return false;
        }
        reference item = GetItem(m_CurBlk)[--m_CurIdx];
        NLIB_UNUSED(item);
        item.~T();
        return true;
    }
    bool pop_back_(TrueType) NLIB_NOEXCEPT {
        if (m_CurIdx == 0) {
            if (!ConfirmBack()) return false;
        }
        --m_CurIdx;
        return true;
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Nlist);
};

template <class T, class AL>
inline Nlist<T, AL>::~Nlist() NLIB_NOEXCEPT {
    this->clear();
    this->DeallocBlkList(m_FirstBlk, 0);
}

namespace nlist {

template <class T, class AL>
class ResizeRewinder {
 public:
    ResizeRewinder(Nlist<T, AL>& l) NLIB_NOEXCEPT : m_Count(0), m_List(l) {}  // NOLINT
    void PushBack(size_t n) {
        for (; m_Count < n; ++m_Count) {
            m_List.push_back();  // may throw in the default constructor of T
        }
        m_Count = 0;
    }
    ~ResizeRewinder() NLIB_NOEXCEPT {
        for (; m_Count > 0; --m_Count) {
            m_List.pop_back();
        }
    }

 private:
    size_t m_Count;
    Nlist<T, AL>& m_List;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ResizeRewinder);
};

}  // namespace nlist

template <class T, class AL>
bool Nlist<T, AL>::Resize(size_t n, FalseType) NLIB_NOEXCEPT {
    // NOTE: This can be optimized.
    size_t sz = this->size();
    if (n == sz) return true;
    if (n < sz) {
        for (size_t i = sz; i > n; --i) {
            this->pop_back();
        }
    } else {
        if (!this->reserve(n)) return false;
        nlist::ResizeRewinder<T, AL> rewinder(*this);
        rewinder.PushBack(n);
    }
    return true;
}

template <class T, class AL>
bool Nlist<T, AL>::Resize(size_t n, TrueType) NLIB_NOEXCEPT {
    size_type sz = this->size();
    if (n == sz) return true;
    if (n < sz) {
        for (size_type i = sz; i > n; --i) {
            this->pop_back();
        }
    } else {
        if (!this->reserve(n)) return false;
        // NOTE:
        // Just set m_CurLevel, m_CurBlk, m_CurIdx
        // because we don't have to invoke the constuctor of T.
        size_t lower = this->BlkBaseIdx(m_CurLevel);
        size_t upper = this->BlkBaseIdx(m_CurLevel + 1);
        while (upper > n) {
            ++m_CurLevel;
            m_CurBlk = m_CurBlk->next;
            lower = upper;
            upper = this->BlkBaseIdx(m_CurLevel + 1);
        }
        m_CurIdx = static_cast<size_t>(n - lower);
    }
    return true;
}

template <class T, class AL>
bool Nlist<T, AL>::reserve(size_type n) NLIB_NOEXCEPT {
    if (n == 0 || n <= BlkBaseIdx(m_CurLevel + 1)) return true;
    if (!m_FirstBlk && !this->ConfirmFirstBlk()) return false;
    size_t lv;
    Blk* blk = this->LastBlock(&lv);
    for (;;) {
        if (n <= BlkBaseIdx(lv + 1)) return true;
        ++lv;
        Blk* p = this->AllocBlk(BlkSize(lv));
        if (!p) return false;
        blk->next = p;
        blk = p;
    }
}

template <class T, class AL>
typename Nlist<T, AL>::Blk* Nlist<T, AL>::AllocBlk(size_t n) const NLIB_NOEXCEPT {
    NLIB_STATIC_ASSERT(NLIB_ALIGNOF(T) == 1 || NLIB_ALIGNOF(T) == 2 ||
                       NLIB_ALIGNOF(T) == 4 || NLIB_ALIGNOF(T) == 8);
    NLIB_STATIC_ASSERT(sizeof(Blk) % 8 == 0);
    size_t sz = sizeof(Blk) + n * sizeof(T);
    void* pp;
    NLIB_TRY {
        pp = AllocType::_GetNlistAlloc().allocate(sz);  // must be 8 bytes aligned
    }
#ifdef NLIB_EXCEPTION_ENABLED
    NLIB_CATCH(const std::bad_alloc&) { return NULL; }
#endif
    if (!pp) return NULL;
    Blk* p = reinterpret_cast<Blk*>(pp);
    p->next = NULL;
    p->item = reinterpret_cast<T*>(p + 1);
    return p;
}

template <class T, class AL>
void Nlist<T, AL>::DeallocBlkList(Blk* p, int lv) const NLIB_NOEXCEPT {
    while (p) {
        size_t n = BlkSize(lv);
        Blk* pp = p;
        p = pp->next;

        // DeallocBlock
        AllocType::_GetNlistAlloc().deallocate(reinterpret_cast<char*>(pp),
                                               sizeof(Blk) + n * sizeof(T));
        ++lv;
    }
}

template <class T, class AL>
void Nlist<T, AL>::DestroyBlk(Blk* p, size_t n, FalseType) const NLIB_NOEXCEPT {
    // NOTE:
    // Do not use std::allocator::destroy() because destroy() is not required.
    // invoke the destructor directly.
    T* base = reinterpret_cast<T*>(p->item);
    NLIB_UNUSED(base);
    for (size_t i = 0; i < n; ++i) {
        (base + i)->~T();
    }
}

template <class T, class AL>
bool Nlist<T, AL>::ConfirmFirstBlk() const NLIB_NOEXCEPT {
    Blk* p = AllocBlk(BlkSize(0));
    if (!p) return false;
    m_FirstBlk = p;
    m_CurBlk = p;
    m_CurIdx = 0;
    m_CurLevel = 0;
    return true;
}

template <class T, class AL>
void* Nlist<T, AL>::PushBack() NLIB_NOEXCEPT {
    if (!m_FirstBlk && !ConfirmFirstBlk()) return NULL;
    if (m_CurIdx == BlkSize(m_CurLevel)) {
        if (!m_CurBlk->next) {
            Blk* p = AllocBlk(BlkSize(m_CurLevel + 1));
            if (!p) return NULL;
            m_CurBlk->next = p;
        }
        ++m_CurLevel;
        m_CurBlk = m_CurBlk->next;
        m_CurIdx = 0;
    }

    // NOTE:
    // If m_CurIdx is incremented here,
    // The state will be inconsisitent when T's constructor throws an exception.
    T& item = GetItem(m_CurBlk)[m_CurIdx];
    return &item;
}

template <class T, class AL>
void Nlist<T, AL>::Clear(FalseType) NLIB_NOEXCEPT {
    int lv = 0;
    Blk* p = m_FirstBlk;
    if (!p) return;
    while (p != m_CurBlk) {
        size_t cnt = BlkSize(lv);
        DestroyBlk(p, cnt);
        ++lv;
        p = p->next;
    }
    if (p == m_CurBlk) {
        DestroyBlk(p, m_CurIdx);
    }
    ClearSimple();
}

template <class T, class AL1, class AL2>
inline bool operator==(const Nlist<T, AL1>& lhs,
                                   const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return lhs.size() == rhs.size() && std::equal(lhs.begin(), lhs.end(), rhs.begin());
}

template <class T, class AL1, class AL2>
inline bool operator!=(const Nlist<T, AL1>& lhs,
                                   const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}

template <class T, class AL1, class AL2>
inline bool operator<(const Nlist<T, AL1>& lhs,
                                  const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return std::lexicographical_compare(lhs.begin(), lhs.end(), rhs.begin(), rhs.end());
}

template <class T, class AL1, class AL2>
inline bool operator>(const Nlist<T, AL1>& lhs,
                                  const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return (rhs < lhs);
}

template <class T, class AL1, class AL2>
inline bool operator<=(const Nlist<T, AL1>& lhs,
                                   const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return (!(rhs < lhs));
}

template <class T, class AL1, class AL2>
inline bool operator>=(const Nlist<T, AL1>& lhs,
                                   const Nlist<T, AL2>& rhs) NLIB_NOEXCEPT {
    return (!(lhs < rhs));
}

template <class AL = std::allocator<char> >
class StringList {
    typedef Nlist<char*, AL> ContainerType;

 public:
    typedef AL allocator_type;
    typedef typename ContainerType::iterator iterator;
    typedef typename ContainerType::const_iterator const_iterator;

 public:
    StringList() NLIB_NOEXCEPT : m_List() {}
    explicit StringList(const AL& al) : m_List(al) {}
    ~StringList() NLIB_NOEXCEPT;
    size_t size() const NLIB_NOEXCEPT { return m_List.size(); }
    size_t capacity() const NLIB_NOEXCEPT { return m_List.capacity(); }
    bool empty() const NLIB_NOEXCEPT { return m_List.empty(); }
    bool reserve(size_t n) { return m_List.reserve(n); }
    bool resize(size_t n) { return m_List.resize(n); }

    char** push_back(const char* str);
    template <class Iterator>
    char** push_back(Iterator first, Iterator last);

    bool pop_back() NLIB_NOEXCEPT;
    void swap(StringList& rhs) NLIB_NOEXCEPT { m_List.swap(rhs.m_List); }  // NOLINT
    void clear() NLIB_NOEXCEPT { m_List.clear(); }
    allocator_type get_allocator() const { return m_List.get_allocator(); }

    const char* back() const { return m_List.back(); }
    char* back() { return m_List.back(); }
    const char* front() const { return m_List.front(); }
    char* front() { return m_List.front(); }
    const char* operator[](size_t idx) const { return m_List[idx]; }
    char* operator[](size_t idx) { return m_List[idx]; }
    iterator begin() NLIB_NOEXCEPT { return m_List.begin(); }
    const_iterator begin() const NLIB_NOEXCEPT { return m_List.begin(); }
    const_iterator cbegin() NLIB_NOEXCEPT { return m_List.cbegin(); }

    iterator end() NLIB_NOEXCEPT { return m_List.end(); }
    const_iterator end() const NLIB_NOEXCEPT { return m_List.end(); }
    const_iterator cend() NLIB_NOEXCEPT { return m_List.cend(); }

    void shrink_to_fit() { m_List.shrink_to_fit(); }
    void Clobber() NLIB_NOEXCEPT { m_List.clobber(); }

    char** Replace(iterator it, const char* str);
    char** Replace(size_t idx, const char* str) {
        if (idx >= m_List.size()) return NULL;
        iterator it = m_List.begin();
        it.Advance(idx);
        return this->Replace(it, str);
    }

 private:
    void Dealloc(const char* ptr) const {
        if (ptr) {
            // NOTE: note that size info is not supplied.
            this->get_allocator().deallocate(const_cast<char*>(ptr), 0);
        }
    }
    char* StrDup(const char* ptr) const;
    char* Alloc(size_t nBytes) const {
        return this->get_allocator().allocate(nBytes);
    }

 private:
    ContainerType m_List;
    NLIB_DISALLOW_COPY_AND_ASSIGN(StringList);
};

template <class AL>
char* StringList<AL>::StrDup(const char* ptr) const {
    size_t nBytes = nlib_strlen(ptr) + 1;
    void* p = this->Alloc(nBytes);
    if (!p) return NULL;
    nlib_memcpy(p, nBytes, ptr, nBytes);
    return reinterpret_cast<char*>(p);
}

template <class AL>
StringList<AL>::~StringList() NLIB_NOEXCEPT {
    const_iterator it = this->cbegin();
    const_iterator it_end = this->cend();
    while (it != it_end) {
        Dealloc(*it);
        ++it;
    }
}

template <class AL>
char** StringList<AL>::push_back(const char* str) {
    char* dup;
    if (str) {
        dup = this->StrDup(str);
        if (!dup) return NULL;
    } else {
        dup = NULL;
    }
    char** ptrStr = m_List.push_back(dup);
    if (!ptrStr) this->Dealloc(dup);
    return ptrStr;
}

template <class AL>
template <class Iterator>
char** StringList<AL>::push_back(Iterator first, Iterator last) {
    typename std::iterator_traits<Iterator>::difference_type n = std::distance(first, last);
    char* str = this->Alloc(n + 1);
    if (!str) return NULL;
    std::copy(first, last, str);
    str[n] = '\0';
    char** rval = m_List.push_back(str);
    if (!rval) this->Dealloc(str);
    return rval;
}

template <class AL>
bool StringList<AL>::pop_back() NLIB_NOEXCEPT {
    if (m_List.empty()) return false;
    Dealloc(m_List.back());
    return m_List.pop_back();
}

template <class AL>
char** StringList<AL>::Replace(iterator it, const char* str) {
    char* dup;
    if (str) {
        dup = this->StrDup(str);
        if (!dup) return NULL;
    } else {
        dup = NULL;
    }
    Dealloc(*it);
    *it = dup;
    return &*it;
}

template <class AL>
bool AppendString(Nlist<char, AL>* obj, const char* str) {
    size_t n = nlib_strlen(str);
    if (!obj->reserve(obj->size() + n)) return false;
    while (*str) {
        obj->push_back(*str);
        ++str;
    }
    return true;
}

NLIB_NAMESPACE_END

namespace std {

template <class T, class Distance>
inline void advance(::nlib_ns::nlist::NlistIterator<T>& pos,  // NOLINT
                    Distance n) {
    pos.Advance(static_cast<size_t>(n));
}

template <class T>
inline std::ptrdiff_t distance(const ::nlib_ns::nlist::NlistIterator<T>& first,
                               const ::nlib_ns::nlist::NlistIterator<T>& last) {
    return first.Distance(last);
}

}  // namespace std

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#else
NLIB_DEFINE_STD_SWAP_T_BEGIN1(std)  // NOLINT
#endif

NLIB_DEFINE_STD_SWAP_T2(T, AL, NLIB_NS::Nlist)  // NOLINT
NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::StringList)  // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#else
NLIB_DEFINE_STD_SWAP_T_END1(std)  // NOLINT
#endif

#endif  // INCLUDE_NN_NLIB_NLIST_H_
