﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_OUTPUTSTREAM_H_
#define INCLUDE_NN_NLIB_OUTPUTSTREAM_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

// code snippets:
//   OutputStream& stream = 'stream that inherits OutputStream';
//   if (!stream.Write(dataptr, nbyte)) { error = stream.GetErrorValue(); .... }
//   for (.....) { if (!stream.Write(byte)) { error = stream.GetErrorValue(); .... } }
//   # you should Flush() and Close() explicitly
//   if (!stream.Flush()) { error = stream.GetErrorValue(); .... } }
//   if (!stream.Close()) { error = stream.GetErrorValue(); .... } }
class NLIB_VIS_PUBLIC OutputStream {
 public:
    enum BufferingMode {
        BUFFERINGMODE_BLOCKBUFFERED = 0,  // default
        BUFFERINGMODE_LINEBUFFERED,       // if console output
        BUFFERINGMODE_UNBUFFERED
    };

 public:
    NLIB_CEXPR OutputStream() NLIB_NOEXCEPT : m_Pos(0),
                                              m_Cur(0),
                                              m_Buf(NULL),
                                              m_BufSize(0),
                                              m_ErrorId(0),
                                              m_BufferingMode(BUFFERINGMODE_BLOCKBUFFERED) {}
    virtual ~OutputStream() NLIB_NOEXCEPT {
        // NOTE:
        // cannot call Close_() from this destructor.
        // the destructor of the derived class has to call Close_() if needed.
    }
    // use Pos64() if you need a 64bit value on 32bit arch
    size_t Pos() const NLIB_NOEXCEPT { return static_cast<size_t>(Pos64()); }
    uint64_t Pos64() const NLIB_NOEXCEPT { return m_Pos + m_Cur; }
    bool Write(int b) NLIB_NOEXCEPT {
        // NOTE:
        // if the stream is closed,
        // Write() will succeed before the internal buffer becomes full.
        if (NLIB_UNLIKELY(m_Cur == m_BufSize)) {
            if (NLIB_UNLIKELY(!this->Flush_(false))) return false;
            if (NLIB_UNLIKELY(m_BufSize == 0)) {
                this->SetError(EIO);
                return false;
            }
        }
        m_Buf[m_Cur++] = static_cast<unsigned char>(b & 0xff);
        return true;
    }
    bool Write(const void* p, size_t n) NLIB_NOEXCEPT NLIB_NONNULL {
#ifndef NLIB_NONNULL_ENABLED
        if (!p) {
            this->SetError(EINVAL);
            return false;
        }
#endif
        if (n > RSIZE_MAX) {
            this->SetError(EINVAL);
            return false;
        }
        if (m_Cur + n <= m_BufSize) {
            nlib_memcpy(&m_Buf[m_Cur], n, p, n);
            m_Cur += n;
            return true;
        }
        return this->Write_(p, n);
    }
    bool WriteGather(const nlib_fd_iovec* iov, int iovcnt) NLIB_NOEXCEPT NLIB_NONNULL {
#ifndef NLIB_NONNULL_ENABLED
        if (!iov) {
            this->SetError(EINVAL);
            return false;
        }
#endif
        if (iovcnt < 0) {
            this->SetError(EINVAL);
            return false;
        }
        return this->WriteGather_(iov, iovcnt);
    }
    bool Flush() NLIB_NOEXCEPT { return Flush_(true); }
    bool Close() NLIB_NOEXCEPT;
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return m_ErrorId; }
    // returns true if there has been no errors
    // you can also write the codes such that 'if (!!stream) { ... }'
    bool IsOk() const NLIB_NOEXCEPT { return m_ErrorId == 0; }
    BufferingMode GetBufferingMode() const NLIB_NOEXCEPT { return m_BufferingMode; }
    NLIB_SAFE_BOOL(OutputStream, IsOk())

 protected:
    void ResetBuffer(void* p, size_t nBytes) NLIB_NOEXCEPT {
        m_Buf = reinterpret_cast<unsigned char*>(p);
        m_BufSize = static_cast<int>(nBytes);
    }
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (m_ErrorId == 0) m_ErrorId = e;
    }

 private:
    virtual bool PushBuffer_(const void* p, size_t nBytes, bool doFlush) NLIB_NOEXCEPT = 0;
    virtual bool Close_() NLIB_NOEXCEPT = 0;
    virtual void* GetWorkBuffer_(size_t* nBytes) NLIB_NOEXCEPT;
    virtual bool WriteGather_(const nlib_fd_iovec* iov, int iovcnt) NLIB_NOEXCEPT;
    bool Write_(const void* p, size_t n) NLIB_NOEXCEPT;

 private:
    bool Flush_(bool flushDevice) NLIB_NOEXCEPT;
    bool GetWorkBuffer_() NLIB_NOEXCEPT;

 private:
    uint64_t m_Pos;
    size_t m_Cur;
    unsigned char* m_Buf;
    size_t m_BufSize;
    mutable errno_t m_ErrorId;

 protected:
    // NOTE:
    // BufferingMode is for the codes which use OutputStream.
    // OutputStream only shows the buffering mode which the user specified.
    // for example, TextWriter (the user of OutputStream) has to decide
    // when to call Flush() looking at GetBufferingMode().
    BufferingMode m_BufferingMode;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(OutputStream);
};

class NLIB_VIS_PUBLIC NullOutputStream NLIB_FINAL : public OutputStream {
 public:
    NullOutputStream() NLIB_NOEXCEPT;
    virtual ~NullOutputStream() NLIB_NOEXCEPT NLIB_OVERRIDE {}

 private:
    unsigned char m_DummyBuf[256];
    virtual bool PushBuffer_(const void* p, size_t, bool doFlush) NLIB_NOEXCEPT NLIB_OVERRIDE;
    virtual bool Close_() NLIB_NOEXCEPT NLIB_OVERRIDE { return true; }
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_OUTPUTSTREAM_H_
