﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_REALLOCOUTPUTSTREAM_H_
#define INCLUDE_NN_NLIB_REALLOCOUTPUTSTREAM_H_

#include <stdlib.h>

#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/OutputStream.h"

NLIB_NAMESPACE_BEGIN

template <class T>
class FreeDeleter {
 public:
    typedef void (*FreeFunc)(void* ptr);
    FreeDeleter() NLIB_NOEXCEPT : m_Free(& ::nlib_free) {}
    explicit FreeDeleter(FreeFunc freefunc) NLIB_NOEXCEPT : m_Free(freefunc) {}
    template <class U>
    FreeDeleter(const FreeDeleter<U>& rhs) NLIB_NOEXCEPT {
        m_Free = rhs.m_Free;
    }
    void operator()(T* p) const NLIB_NOEXCEPT {
        if (p && m_Free) {
            (*m_Free)(p);
        }
    }
    FreeFunc GetFunc() const NLIB_NOEXCEPT { return m_Free; }

 private:
    FreeFunc m_Free;
};

template <class T>
class ReallocDeleter {
 public:
    typedef void* (*ReallocFunc)(void* ptr, size_t n);
    ReallocDeleter() NLIB_NOEXCEPT : m_Realloc(::nlib_realloc) {}
    explicit ReallocDeleter(ReallocFunc reallocfunc) NLIB_NOEXCEPT : m_Realloc(reallocfunc) {}
    template <class U>
    ReallocDeleter(const ReallocDeleter<U>& rhs) NLIB_NOEXCEPT {
        m_Realloc = rhs.m_Realloc;
    }
    void operator()(T* p) const NLIB_NOEXCEPT {
        if (p && m_Realloc) {
            (*m_Realloc)(p, 0);
        }
    }
    ReallocFunc GetFunc() const NLIB_NOEXCEPT { return m_Realloc; }

 private:
    ReallocFunc m_Realloc;
};

NLIB_STATIC_ASSERT(!IsEmpty<FreeDeleter<int> >::value && !IsEmpty<ReallocDeleter<int> >::value);

#ifdef _MSC_VER
class ReallocOutputStream NLIB_FINAL : public OutputStream {
#else
class NLIB_VIS_PUBLIC ReallocOutputStream NLIB_FINAL : public OutputStream {
#endif

 public:
    typedef UniquePtr<uint8_t[], ReallocDeleter<uint8_t> > UniquePtrType;
    typedef UniquePtr<char[], ReallocDeleter<char> > CharPtrType;
    typedef void* (*ReallocFunc)(void* ptr, size_t n);

    ReallocOutputStream() NLIB_NOEXCEPT : m_Data(), m_Capacity(0) {}
    NLIB_VIS_PUBLIC explicit ReallocOutputStream(ReallocFunc func) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_VIS_PUBLIC virtual ~ReallocOutputStream() NLIB_NOEXCEPT NLIB_OVERRIDE;

    uint8_t* Data() NLIB_NOEXCEPT { return m_Data.get(); }
    const uint8_t* Data() const NLIB_NOEXCEPT { return m_Data.get(); }

    bool Reserve(size_t n) NLIB_NOEXCEPT { return this->Expand(n) == 0; }
    void ShrinkToFit() NLIB_NOEXCEPT {
        if (this->ExpandOrShrink(this->Pos()) == 0) {
            this->ResetBuffer(m_Data.get() + m_Capacity, 0);
        }
    }

    NLIB_VIS_PUBLIC size_t Release(UniquePtrType* ptr) NLIB_NOEXCEPT;
    // UniquePtrType Reelase(size_t* n);

    // Appends '\0' and, returns the memory after ShrinkToFit().
    NLIB_VIS_PUBLIC bool ReleaseAsCstring(CharPtrType* ptr) NLIB_NOEXCEPT;

 private:
    errno_t Expand(size_t newCap) NLIB_NOEXCEPT {
        return (newCap > m_Capacity) ? this->ExpandOrShrink(newCap) : 0;
    }
    NLIB_VIS_PUBLIC errno_t ExpandOrShrink(size_t newCap) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC virtual bool PushBuffer_(const void* p, size_t nBytes,
                                             bool doFlush) NLIB_NOEXCEPT NLIB_OVERRIDE;
    NLIB_VIS_PUBLIC virtual bool Close_() NLIB_NOEXCEPT NLIB_OVERRIDE;
    NLIB_VIS_PUBLIC virtual void* GetWorkBuffer_(size_t* nBytes) NLIB_NOEXCEPT NLIB_OVERRIDE;

 private:
    UniquePtrType m_Data;
    size_t m_Capacity;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocOutputStream);
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_REALLOCOUTPUTSTREAM_H_
