﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_REALLOCVEC_H_
#define INCLUDE_NN_NLIB_REALLOCVEC_H_

#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <iterator>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/TypeTraits.h"

NLIB_NAMESPACE_BEGIN

template <class T>
class ReallocVec {
 public:
    typedef void* (*ReallocFunc)(void*, size_t);
    typedef T value_type;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef T& reference;
    typedef const T& const_reference;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef T* iterator;
    typedef const T* const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    NLIB_CEXPR ReallocVec() NLIB_NOEXCEPT : m_Vec(NULL),
                                            m_Cur(0),
                                            m_Size(0),
                                            m_Realloc(::nlib_realloc) {}
    NLIB_CEXPR explicit ReallocVec(ReallocFunc func) NLIB_NOEXCEPT NLIB_NONNULL;
    ~ReallocVec() NLIB_NOEXCEPT {
        if (m_Vec) {
            m_Realloc(m_Vec, 0);
        }
        NLIB_STATIC_ASSERT(IsPod<T>::value);
    }
    NLIB_MOVE_MEMBER_HELPER_4(ReallocVec, m_Vec, m_Cur, m_Size, m_Realloc)

    ReallocFunc GetRealloc() const NLIB_NOEXCEPT { return m_Realloc; }
    bool push_back(const T& v) NLIB_NOEXCEPT {
        if (m_Cur == m_Size) {
            size_t newSize = (m_Size == 0) ? 1 : m_Size * 2;
            if (newSize < m_Size) return false;
            if (!this->Expand(newSize)) return false;
        }
        m_Vec[m_Cur++] = v;
        return true;
    }
    bool pop_back() NLIB_NOEXCEPT {
        if (m_Cur == 0) return false;
        --m_Cur;
        return true;
    }
    T& operator[](size_t idx) { return m_Vec[idx]; }
    const T& operator[](size_t idx) const { return m_Vec[idx]; }
    T& front() { return *begin(); }
    const T& front() const { return *begin(); }
    T& back() { return *(begin() + m_Cur - 1); }
    const T& back() const { return *(begin() + m_Cur - 1); }

    iterator begin() NLIB_NOEXCEPT { return m_Vec; }
    const_iterator begin() const NLIB_NOEXCEPT { return m_Vec; }
    const_iterator cbegin() const NLIB_NOEXCEPT { return m_Vec; }
    iterator end() NLIB_NOEXCEPT { return begin() + m_Cur; }
    const_iterator end() const NLIB_NOEXCEPT { return begin() + m_Cur; }
    const_iterator cend() const NLIB_NOEXCEPT { return begin() + m_Cur; }
    reverse_iterator rbegin() NLIB_NOEXCEPT { return reverse_iterator(end()); }
    const_reverse_iterator rbegin() const NLIB_NOEXCEPT {
        return const_reverse_iterator(end());
    }
    reverse_iterator rend() NLIB_NOEXCEPT { return reverse_iterator(begin()); }
    const_reverse_iterator rend() const NLIB_NOEXCEPT {
        return const_reverse_iterator(begin());
    }
    const_reverse_iterator crbegin() const NLIB_NOEXCEPT {
        return const_reverse_iterator(end());
    }
    const_reverse_iterator crend() const NLIB_NOEXCEPT {
        return const_reverse_iterator(begin());
    }
    void swap(ReallocVec& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_Vec, rhs.m_Vec);
        swap(m_Cur, rhs.m_Cur);
        swap(m_Size, rhs.m_Size);
        swap(m_Realloc, rhs.m_Realloc);
    }
    size_t size() const NLIB_NOEXCEPT { return m_Cur; }
    size_t capacity() const NLIB_NOEXCEPT { return m_Size; }
    bool empty() const NLIB_NOEXCEPT { return m_Cur == 0; }
    bool reserve(size_t n) NLIB_NOEXCEPT {
        return (n > m_Size) ? this->Expand(n) : true;
    }
    bool resize(size_t n) NLIB_NOEXCEPT {
        if (!this->reserve(n)) return false;
        m_Cur = n;
        return true;
    }
    void clear() NLIB_NOEXCEPT { m_Cur = 0; }
    void shrink_to_fit() NLIB_NOEXCEPT {
        if (m_Cur != m_Size) {
            if (m_Cur > 0) {
                this->Expand(m_Cur);
            } else {
                m_Realloc(m_Vec, 0);
                m_Vec = NULL;
                m_Size = 0;
            }
        }
    }
    iterator insert(iterator pos, const T& val) {
        ptrdiff_t idx = pos - begin();
        if (!this->push_back(val)) return NULL;
        std::rotate(begin() + idx, &back(), end());
        return begin() + idx;
    }

 private:
    bool Expand(size_t newSize) NLIB_NOEXCEPT;

    T* m_Vec;
    size_t m_Cur;
    size_t m_Size;
    ReallocFunc m_Realloc;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocVec);
};

template <class T>
inline NLIB_CEXPR ReallocVec<T>::ReallocVec(ReallocFunc func) NLIB_NOEXCEPT : m_Vec(NULL),
                                                                              m_Cur(0),
                                                                              m_Size(0),
                                                                              m_Realloc(func) {}

template <class T>
bool ReallocVec<T>::Expand(size_t newSize) NLIB_NOEXCEPT {
    T* ptr = m_Vec;
    void* newPtr = m_Realloc(ptr, newSize * sizeof(T));
    if (!newPtr) return false;
    m_Vec = reinterpret_cast<T*>(newPtr);
    m_Size = newSize;
    return true;
}

template <class T>
class ReallocQueue {
 public:
    typedef typename ReallocVec<T>::ReallocFunc ReallocFunc;
    ReallocQueue() NLIB_NOEXCEPT : m_FrontIdx(0) {}
    explicit ReallocQueue(ReallocFunc func) NLIB_NOEXCEPT : m_FrontIdx(0),
                                                            m_FrontVec(func),
                                                            m_BackVec(func) {}
    ~ReallocQueue() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_3(ReallocQueue, m_FrontIdx, m_FrontVec, m_BackVec)
    ReallocFunc GetRealloc() const NLIB_NOEXCEPT {
        return m_FrontVec.GetRealloc();
    }
    size_t size() const NLIB_NOEXCEPT {
        return m_FrontVec.size() - m_FrontIdx + m_BackVec.size();
    }
    bool empty() const NLIB_NOEXCEPT {
        return m_FrontIdx == m_FrontVec.size() && m_BackVec.empty();
    }
    bool push(const T& val) NLIB_NOEXCEPT { return m_BackVec.push_back(val); }
    bool pop() NLIB_NOEXCEPT {
        if (m_FrontIdx == m_FrontVec.size()) {
            if (m_BackVec.empty()) return false;
            m_FrontVec.clear();
            m_FrontVec.swap(m_BackVec);
            m_FrontIdx = 0;
        }
        ++m_FrontIdx;
        return true;
    }
    T& front() {
        if (m_FrontIdx == m_FrontVec.size()) {
            NLIB_ASSERT(!m_BackVec.empty());
            m_FrontVec.clear();
            m_FrontVec.swap(m_BackVec);
            m_FrontIdx = 0;
        }
        return m_FrontVec[m_FrontIdx];
    }
    bool pop(T* v) NLIB_NOEXCEPT {
        if (m_FrontIdx == m_FrontVec.size()) {
            if (m_BackVec.empty()) return false;
            m_FrontVec.clear();
            m_FrontVec.swap(m_BackVec);
            m_FrontIdx = 0;
        }
        if (v) *v = m_FrontVec[m_FrontIdx];
        ++m_FrontIdx;
        return true;
    }
    void swap(ReallocQueue& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_FrontIdx, rhs.m_FrontIdx);
        m_FrontVec.swap(rhs.m_FrontVec);
        m_BackVec.swap(rhs.m_BackVec);
    }

 private:
    size_t m_FrontIdx;
    ReallocVec<T> m_FrontVec;
    ReallocVec<T> m_BackVec;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocQueue);
};

class ReallocCstringVec {
    typedef ReallocVec<char*> VecType;
    typedef VecType::ReallocFunc ReallocFunc;

 public:
    ReallocCstringVec() NLIB_NOEXCEPT : m_Vec() {}
    explicit ReallocCstringVec(ReallocFunc func) NLIB_NOEXCEPT : m_Vec(func) {}
    ~ReallocCstringVec() NLIB_NOEXCEPT { this->clear(); }
    NLIB_MOVE_MEMBER_HELPER_1(ReallocCstringVec, m_Vec)
    ReallocFunc GetRealloc() const NLIB_NOEXCEPT { return m_Vec.GetRealloc(); }
    bool push_back(const char* str) NLIB_NOEXCEPT {
        return this->push_back(str, str + nlib_strlen(str));
    }
    bool push_back(const char* startCharPtr,
                                      const char* endCharPtr) NLIB_NOEXCEPT {
        size_t n = endCharPtr - startCharPtr;
        char* p = reinterpret_cast<char*>(m_Vec.GetRealloc()(NULL, n + 1));
        if (!p) return false;
        nlib_memcpy(p, n, startCharPtr, n);
        p[n] = '\0';
        if (!m_Vec.push_back(p)) return false;
        return true;
    }
    bool pop_back() NLIB_NOEXCEPT {
        if (m_Vec.empty()) return false;
        m_Vec.GetRealloc()(m_Vec.back(), 0);
        m_Vec.pop_back();
        return true;
    }
    char* operator[](size_t idx)NLIB_NOEXCEPT { return m_Vec[idx]; }
    const char* operator[](size_t idx) const NLIB_NOEXCEPT {
        return m_Vec[idx];
    }
    char* front() NLIB_NOEXCEPT { return m_Vec.front(); }
    const char* front() const NLIB_NOEXCEPT { return m_Vec.front(); }
    char* back() NLIB_NOEXCEPT { return m_Vec.back(); }
    const char* back() const NLIB_NOEXCEPT { return m_Vec.back(); }
    void swap(ReallocCstringVec& rhs) NLIB_NOEXCEPT { m_Vec.swap(rhs.m_Vec); }
    size_t size() const NLIB_NOEXCEPT { return m_Vec.size(); }
    size_t capacity() const NLIB_NOEXCEPT { return m_Vec.capacity(); }
    bool empty() const NLIB_NOEXCEPT { return m_Vec.empty(); }
    bool reserve(size_t n) NLIB_NOEXCEPT { return m_Vec.reserve(n); }
    void clear() NLIB_NOEXCEPT {
        VecType::iterator it = m_Vec.begin();
        VecType::iterator end = m_Vec.end();
        for (; it != end; ++it) {
            m_Vec.GetRealloc()(*it, 0);
        }
        m_Vec.clear();
    }

 private:
    VecType m_Vec;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ReallocCstringVec);
};

NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(NLIB_NS::ReallocCstringVec)
#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#else
NLIB_DEFINE_STD_SWAP_T_BEGIN1(std)  // NOLINT
#endif

NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::ReallocVec)  // NOLINT
NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::ReallocQueue)  // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#else
NLIB_DEFINE_STD_SWAP_T_END1(std)  // NOLINT
#endif

#endif  // INCLUDE_NN_NLIB_REALLOCVEC_H_
