﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TAGGEDTEXTPARSER_H_
#define INCLUDE_NN_NLIB_TAGGEDTEXTPARSER_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

class TaggedTextParser NLIB_FINAL {
 public:
    enum Event {
        NONE = -1,
        START_ELEMENT = 1,
        END_ELEMENT,
        CHARACTERS,
        END_DOCUMENT,
        SYNTAX_ERROR
    };
    explicit TaggedTextParser(char* taggedText) {
        // change taggedText in parsing
        m_Text = taggedText;
        m_Cur = reinterpret_cast<unsigned char*>(taggedText);
        m_Chr = *taggedText;
        *m_Cur++ = '\0';
        m_IsEmptyElemTag = false;
    }
    const char* GetText() const {
        // tag name if START_ELEMENT, END_ELEMENT
        // trimmed string if CHARACTERS
        return m_Text;
    }
    Event Next() {
    NLIB_TAGGEDTEXTPARSER_NEXT:
        if (m_IsEmptyElemTag) {
            m_IsEmptyElemTag = false;
            return END_ELEMENT;
        }
        // SkipSpace
        if (IsSpc(m_Chr)) {
            for (;;) {
                if (!IsSpc(*m_Cur)) {
                    m_Chr = *m_Cur++;
                    break;
                }
                ++m_Cur;
            }
        }
        if (m_Chr == '\0') return END_DOCUMENT;
        if (m_Chr == '<') {
            unsigned char* p = m_Cur;
            int c = *p;
            if (c == '?' || c == '!') {
                // skip until '>' if '<!' or '<?'.
                ++p;
                unsigned char* p2 = GotoGT(p);
                if (*p2 == '\0') goto NLIB_TAGGEDTEXTPARSER_ERROR;
                ++p2;  // skip '>'
                m_Chr = *p2;
                m_Cur = *p2 != '\0' ? p2 + 1 : p2;
                goto NLIB_TAGGEDTEXTPARSER_NEXT;
            } else if (c == '/') {
                // ETag
                ++p;  // move after '/'
                // skip until space, '/' or '>'.
                unsigned char* p2 = SpotName(p);
                // skip until '>', even if there are attributes.
                unsigned char* p3 = GotoGT(p2);
                if (*p3 == '\0') goto NLIB_TAGGEDTEXTPARSER_ERROR;

                *p2 = '\0';
                m_Text = reinterpret_cast<const char*>(p);
                ++p3;  // skip after '>'
                m_Chr = *p3;
                m_Cur = *p3 != '\0' ? p3 + 1 : p3;
                return END_ELEMENT;
            } else {
                // skip until space, '/' or '>'.
                unsigned char* p2 = SpotName(p);
                // skip until '/' or '>'.
                unsigned char* p3 = GotoSlashGT(p2);
                if (*p3 == '\0') goto NLIB_TAGGEDTEXTPARSER_ERROR;
                if (*p3 == '/') {  // determine '>' or '/>'
                    // EmptyElemTag
                    ++p3;  // move after '/'
                    if (*p3 != '>') goto NLIB_TAGGEDTEXTPARSER_ERROR;
                    m_IsEmptyElemTag = true;
                }
                // STag or EmptyElemTag
                *p2 = '\0';
                m_Text = reinterpret_cast<const char*>(p);
                ++p3;  // move after '>'
                m_Chr = *p3;
                m_Cur = *p3 != '\0' ? p3 + 1 : p3;
                return START_ELEMENT;
            }
        } else {
            // it must be CHARACTERS
            unsigned char* p = m_Cur - 1;
            for (;;) {
                int c = *p;
                if (c == '<' || c == '\0') {
                    m_Text = reinterpret_cast<const char*>(m_Cur - 1);
                    m_Chr = *p;
                    m_Cur = (m_Chr != '\0') ? p + 1 : p;
                    unsigned char* pp = p - 1;
                    while (IsSpc(*pp)) --pp;
                    *(pp + 1) = '\0';  // trim the last space chars
                    return CHARACTERS;
                }
                ++p;
            }
        }
    NLIB_TAGGEDTEXTPARSER_ERROR:
        m_Chr = '\0';
        m_Cur = NULL;
        m_Text = "";
        return SYNTAX_ERROR;
    }

 private:
    static bool IsSpc(int c) {
        // S = ( #x20 | #x9 | #xD | #xA )
        static const char table[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0,  // 0x00 - 0x0F
                                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  // 0x10 - 0x1F
                                     1};
        return (c <= 0x20 && table[c] != 0);
    }
    static unsigned char* SpotName(unsigned char* p) {
        static const char table[] = {
            1, 0, 0, 0, 0, 0, 0, 0, 0,
            1, 1, 0, 0, 1, 0, 0,  // 0x00 - 0x0F
            0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0,  // 0x10 - 0x1F
            1, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 1,  // 0x20 - 0x2F
            0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 1, 0  // 0x30 - 0x3F
        };
        for (;;) {
            int c = *p;
            // (c == '/' || c == '>' || c == '\0' || IsSpc(c))
            if (c < 0x40 && table[c] != 0) return p;
            ++p;
        }
    }
    static unsigned char* GotoGT(unsigned char* p) {
        for (;;) {
            int c = *p;
            if (c == '>' || c == '\0') return p;
            ++p;
        }
    }
    static unsigned char* GotoSlashGT(unsigned char* p) {
        for (;;) {
            int c = *p;
            if (c == '/' || c == '>' || c == '\0') return p;
            ++p;
        }
    }

 private:
    const char* m_Text;
    unsigned char* m_Cur;
    int m_Chr;
    bool m_IsEmptyElemTag;
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_TAGGEDTEXTPARSER_H_
