﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TEXTREADER_H_
#define INCLUDE_NN_NLIB_TEXTREADER_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

class InputStream;

// code snippets:
//   TextReader reader;
//   if (!reader.Init(&stream)) { error }
//   while ((c = reader.Read()) >= 0)
//     c is a codepoint
//   if (!reader) { stream error if !stream, otherwise UTF-8 may be malformed }
//   reader.Close();
//   (stream.Close();)
class NLIB_VIS_PUBLIC TextReader {
 public:
    TextReader() NLIB_NOEXCEPT;
    virtual ~TextReader() NLIB_NOEXCEPT;
    bool Init(InputStream* stream) NLIB_NOEXCEPT NLIB_NONNULL;
    int Read() NLIB_NOEXCEPT {
        if (m_Cur == m_BufEnd) {
            this->FillBuffer();
            if (m_Cur == m_BufEnd) return -1;
        }
        int rval = *reinterpret_cast<unsigned char*>(m_Cur);
        if (!(rval & 0x80)) {
            ++m_Cur;
            if (rval == 0x0A) {
                m_BegLinePos = m_BasePos + (m_Cur - &m_Buf[0]);
                ++m_Line;
            }
            return rval;
        }
        if (!m_Utf32Cache) ConstructUtf32Cache();
        m_Cur += m_NextStep;
        rval = static_cast<int>(m_Utf32Cache);
        m_Utf32Cache = 0;
        return rval;
    }
    int Peek() NLIB_NOEXCEPT {
        if (m_Cur == m_BufEnd) {
            this->FillBuffer();
            if (m_Cur == m_BufEnd) return -1;
        }
        int rval = *reinterpret_cast<unsigned char*>(m_Cur);
        if (!(rval & 0x80)) return rval;
        if (!m_Utf32Cache) ConstructUtf32Cache();
        return static_cast<int>(m_Utf32Cache);
    }
    int SkipWs() NLIB_NOEXCEPT {
        // skip ' ' \t \n (\r)
        if (m_Cur + 1 < m_BufEnd) {
            unsigned char c = *reinterpret_cast<unsigned char*>(m_Cur);
            if (c > ' ') return 0;
            if (c == ' ') {
                ++m_Cur;
                c = *reinterpret_cast<unsigned char*>(m_Cur);
                if (c > ' ') {
                    m_Utf32Cache = 0;
                    return 1;
                }
                return this->SkipWs_(1);
            }
        }
        return this->SkipWs_(0);
    }
    bool ReadUntil(size_t* len, char* buf, size_t n, char delim) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    bool ReadUntil(size_t* len, char (&buf)[N], char delim) NLIB_NOEXCEPT {  // NOLINT
        return this->ReadUntil(len, &buf[0], N, delim);
    }
    template <class T>
    bool ReadUntil(size_t* len, char* buf, size_t n, T pred) NLIB_NOEXCEPT;
    template <class T, size_t N>
    bool ReadUntil(size_t* len, char (&buf)[N], T pred) NLIB_NOEXCEPT {  // NOLINT
        return this->ReadUntil(len, buf, N, pred);
    }
    size_t ReadDecimalString(char* buf, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    size_t ReadDecimalString(char (&buf)[N]) NLIB_NOEXCEPT {
        return this->ReadDecimalString(buf, N);
    }
    bool Proceed(const char* str, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Proceed(char c) NLIB_NOEXCEPT {
        NLIB_ASSERT(!(c & 0x80) && c != 0x0A);
        if (m_Cur == m_BufEnd) {
            this->FillBuffer();
            if (m_Cur == m_BufEnd) return false;
        }
        if (*m_Cur != c) return false;
        ++m_Cur;
        return true;
    }
    bool ProceedEx(const char* str) NLIB_NOEXCEPT;

    int ReadAsUtf8(char* b0, char* b1, char* b2, char* b3) NLIB_NOEXCEPT {
        if (m_Cur == m_BufEnd) {
            this->FillBuffer();
            if (m_Cur == m_BufEnd) return 0;
        }
        *b0 = m_Cur[0];
        int c = m_Cur[0];
        if (!(c & 0x80)) {
            m_Utf32Cache = static_cast<nlib_utf32_t>(c);
            m_Cur += 1;
            if (c == 0x0A) {
                m_BegLinePos = m_BasePos + (m_Cur - &m_Buf[0]);
                ++m_Line;
            }
            return 1;
        }
        m_Utf32Cache = 0;
        switch (c & 0xF0) {
        case 0xC0:
        case 0xD0:
            *b1 = m_Cur[1];
            m_Cur += 2;
            return 2;
        case 0xE0:
            *b1 = m_Cur[1];
            *b2 = m_Cur[2];
            m_Cur += 3;
            return 3;
        default:
            *b1 = m_Cur[1];
            *b2 = m_Cur[2];
            *b3 = m_Cur[3];
            m_Cur += 4;
            return 4;
        }
    }
    int ReadAsUtf16(nlib_utf16_t* upper, nlib_utf16_t* lower) NLIB_NOEXCEPT {
        int utf32 = this->Read();
        return utf32 != -1 ? nlib_utf32char_to_utf16(upper, lower, utf32) : 0;
    }
    int PeekAsUtf16(nlib_utf16_t* upper, nlib_utf16_t* lower) NLIB_NOEXCEPT {
        int utf32 = this->Peek();
        return utf32 != -1 ? nlib_utf32char_to_utf16(upper, lower, utf32) : 0;
    }
    bool Close() NLIB_NOEXCEPT;
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (m_ErrorValue == 0) m_ErrorValue = e;
    }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return m_ErrorValue; }
    InputStream* GetStream() NLIB_NOEXCEPT { return m_Stream; }
    int GetLine() const NLIB_NOEXCEPT { return m_Line; }
    int GetColumn() const NLIB_NOEXCEPT {
        return static_cast<int>((m_Cur - &m_Buf[0]) + 1 + m_BasePos - m_BegLinePos);
    }
    NLIB_SAFE_BOOL(TextReader, GetErrorValue() == 0)

 protected:
    char* GetCur() NLIB_NOEXCEPT { return m_Cur; }
    char* GetBufEnd() NLIB_NOEXCEPT { return m_BufEnd; }
    void SetBufEnd(char* p) NLIB_NOEXCEPT { m_BufEnd = p; }

    // checks UTF-8 validity, converts CRLF to LF, and convert CR to LF
    virtual void FillBuffer_() NLIB_NOEXCEPT;

 private:
    void FillBuffer() NLIB_NOEXCEPT { this->FillBuffer_(); }
    NLIB_VIS_HIDDEN void CountNewLine(const char* mcur, const char* mend) NLIB_NOEXCEPT;
    void ConstructUtf32Cache() NLIB_NOEXCEPT;
    int SkipWs_(int base) NLIB_NOEXCEPT;

 private:
    char m_Buf[512 + 3 + 1];
    nlib_utf32_t m_Utf32Cache;
    char* m_Cur;
    char* m_BufEnd;
    size_t m_NextStep;
    InputStream* m_Stream;
    mutable errno_t m_ErrorValue;
    int m_Line;
    size_t m_BegLinePos;
    size_t m_BasePos;

    NLIB_DISALLOW_COPY_AND_ASSIGN(TextReader);
};

template <class T>
bool TextReader::ReadUntil(size_t* len, char* buf, size_t n, T pred) NLIB_NOEXCEPT {
    // NOTICE: not NULL terminated
    if (!buf) {
        *len = 0;
        return false;
    }
    m_Utf32Cache = 0;
    char* p = buf;
    char* pend = buf + n;
    for (;;) {
        if (m_Cur == m_BufEnd) {
            this->FillBuffer();
            if (m_Cur == m_BufEnd) {
                *len = p - buf;
                return false;
            }
        }
        // T::operator()(const char* ptr);
        while (p != pend && m_Cur != m_BufEnd) {
            if ((*reinterpret_cast<unsigned char*>(m_Cur) & 0xC0) != 0x80 &&
                pred((const char*)m_Cur)) {  // NOLINT
                *len = p - buf;
                return true;
            }
            *p = *m_Cur;
            ++p;
            ++m_Cur;
        }
        if (p == pend) {
            if (m_Cur != m_BufEnd && (*reinterpret_cast<unsigned char*>(m_Cur) & 0xC0) == 0x80) {
                do {
                    --p;
                    --m_Cur;
                } while ((*reinterpret_cast<unsigned char*>(m_Cur) & 0xC0) == 0x80);
            }
            *len = p - buf;
            return false;
        }
    }
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_TEXTREADER_H_
