﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_UNIQUEPTR_H_
#define INCLUDE_NN_NLIB_UNIQUEPTR_H_

#include <stdlib.h>
#include <memory>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/Hash.h"
#include "nn/nlib/TypeTraits.h"

#if defined(NLIB_CXX11_UNIQUEPTR) && defined(NLIB_CXX11_TEMPLATE_ALIAS)
NLIB_NAMESPACE_BEGIN

template <class T, class DEL = std::default_delete<T> >
using UniquePtr = std::unique_ptr<T, DEL>;
template <class T>
using DefaultDelete = std::default_delete<T>;

NLIB_NAMESPACE_END
#else
#include <algorithm>  // NOLINT

NLIB_NAMESPACE_BEGIN

template <class T>
struct DefaultDelete {
    void operator()(T* p) const NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(T));
        if (p) delete p;
    }
};

template <class T>
struct DefaultDelete<T[]> {
    void operator()(T* p) const NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(T));
        if (p) delete[] p;
    }
};

NLIB_STATIC_ASSERT(IsEmpty<DefaultDelete<int> >::value&& IsEmpty<DefaultDelete<int[]> >::value);  // NOLINT

namespace detail {

template <class T, class DEL, bool DEL_EMPTY>
class UniquePtrBase {
 public:
    UniquePtrBase() : m_Ptr(NULL), m_Deleter() {}
    explicit UniquePtrBase(T* ptr) : m_Ptr(ptr), m_Deleter() {}
    UniquePtrBase(T* ptr, const DEL& del) : m_Ptr(ptr), m_Deleter(del) {}

 protected:
    DEL& get_deleter() { return m_Deleter; }
    const DEL& get_deleter() const { return m_Deleter; }
    T* m_Ptr;
    DEL m_Deleter;
    void Swap(UniquePtrBase& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(this->m_Ptr, rhs.m_Ptr);
        swap(this->m_Deleter, rhs.m_Deleter);
    }
    NLIB_DISALLOW_COPY_AND_ASSIGN(UniquePtrBase);
};

template <class T, class DEL>
class UniquePtrBase<T, DEL, true> : public DEL {
 public:
    UniquePtrBase() : m_Ptr(NULL) {}
    explicit UniquePtrBase(T* ptr) : m_Ptr(ptr) {}
    UniquePtrBase(T* ptr, const DEL& del) : DEL(del), m_Ptr(ptr) {}

 protected:
    DEL& get_deleter() { return *this; }
    const DEL& get_deleter() const { return *this; }
    T* m_Ptr;
    void Swap(UniquePtrBase& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(this->m_Ptr, rhs.m_Ptr);
    }
    NLIB_DISALLOW_COPY_AND_ASSIGN(UniquePtrBase);
};

template<class S>
struct IsCompleteType
    : public IntegralConstant<bool, (sizeof(S) != 0 || sizeof(S) == 0)> {};

template<>
struct IsCompleteType<void> : public TrueType {};

}  // namespace detail

template <class T, class DEL = DefaultDelete<T> >
class UniquePtr NLIB_FINAL : public detail::UniquePtrBase<T, DEL, IsEmpty<DEL>::value> {
    typedef detail::UniquePtrBase<T, DEL, IsEmpty<DEL>::value> BaseType;

 public:
    typedef T* pointer_type;   // simpler definition than std::unique_ptr
    typedef DEL deleter_type;  // note that DEL cannot have data members
    typedef T element_type;

    NLIB_CEXPR UniquePtr() NLIB_NOEXCEPT : BaseType() {}
    NLIB_CEXPR explicit UniquePtr(nullptr_t) NLIB_NOEXCEPT : BaseType() {}
    explicit UniquePtr(T* p) NLIB_NOEXCEPT : BaseType(p) {}
    UniquePtr(T* p, const DEL& del) : BaseType(p, del) {}
    NLIB_MOVE_MEMBER_HELPER_1(UniquePtr, BaseType)
    ~UniquePtr() NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(detail::IsCompleteType<T>::value);
        // typedef char is_complete_type[sizeof(T) ? 1 : -1];
        // (void) sizeof(is_complete_type);
        if (this->m_Ptr) get_deleter()(this->m_Ptr);
    }
    void reset(T* p = 0) NLIB_NOEXCEPT {
        // NOTE:
        // this->m_Ptr is for CTR
        if (p != this->m_Ptr) {
            T* old = this->m_Ptr;
            this->m_Ptr = p;
            if (old) get_deleter()(old);
        }
    }
    typename AddRef<T>::type operator*() const {
        NLIB_ASSERT(this->m_Ptr != NULL);
        return *this->m_Ptr;
    }
    T* operator->() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->m_Ptr != NULL);
        return this->m_Ptr;
    }
    T* get() const NLIB_NOEXCEPT { return this->m_Ptr; }
    T* release() NLIB_NOEXCEPT {
        T* p = this->m_Ptr;
        this->m_Ptr = NULL;
        return p;
    }
    void swap(UniquePtr& rhs) NLIB_NOEXCEPT {  // NOLINT
        this->Swap(rhs);
    }
    UniquePtr& operator=(nullptr_t) NLIB_NOEXCEPT {
        this->reset();
        return *this;
    }
    size_t GetHash() const { return Hash<T*>()(this->get()); }
    DEL& get_deleter() { return BaseType::get_deleter(); }
    const DEL& get_deleter() const { return BaseType::get_deleter(); }

    NLIB_SAFE_BOOL(UniquePtr, this->m_Ptr != NULL);

 private:
    template <class S, class D>
    bool operator==(const UniquePtr<S, D>& p) const;
    template <class S, class D>
    bool operator!=(const UniquePtr<S, D>& p) const;
    NLIB_DISALLOW_COPY_AND_ASSIGN(UniquePtr);
};

template <class T, class DEL>
class UniquePtr<T[], DEL> : public detail::UniquePtrBase<T, DEL, IsEmpty<DEL>::value> {
    typedef detail::UniquePtrBase<T, DEL, IsEmpty<DEL>::value> BaseType;

 public:
    typedef T* pointer_type;  // simpler definition than std::unique_ptr
    typedef DEL deleter_type;
    typedef T element_type;

    NLIB_CEXPR UniquePtr() NLIB_NOEXCEPT : BaseType() {}
    NLIB_CEXPR explicit UniquePtr(nullptr_t) NLIB_NOEXCEPT : BaseType() {}
    explicit UniquePtr(T* p) NLIB_NOEXCEPT : BaseType(p) {}
    UniquePtr(T* p, const DEL& del) : BaseType(p, del) {}
    NLIB_MOVE_MEMBER_HELPER_1(UniquePtr, BaseType)
    ~UniquePtr() NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(T));
        get_deleter()(this->m_Ptr);
    }
    void reset(T* p = 0) NLIB_NOEXCEPT {
        // NOTE:
        // this->m_Ptr is for CTR
        if (p != this->m_Ptr) {
            T* old = this->m_Ptr;
            this->m_Ptr = p;
            if (old) get_deleter()(old);
        }
    }
    T& operator*() const {
        NLIB_ASSERT(this->m_Ptr != NULL);
        return *this->m_Ptr;
    }
    T* operator->() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->m_Ptr != NULL);
        return this->m_Ptr;
    }
    T* get() const NLIB_NOEXCEPT { return this->m_Ptr; }
    T* release() NLIB_NOEXCEPT {
        T* p = this->m_Ptr;
        this->m_Ptr = NULL;
        return p;
    }
    void swap(UniquePtr& rhs) NLIB_NOEXCEPT {  // NOLINT
        this->Swap(rhs);
    }
    T& operator[](size_t n) { return get()[n]; }
    const T& operator[](size_t n) const { return get()[n]; }
    size_t GetHash() const { return Hash<T*>()(this->get()); }
    DEL& get_deleter() { return BaseType::get_deleter(); }
    const DEL& get_deleter() const { return BaseType::get_deleter(); }

    NLIB_SAFE_BOOL(UniquePtr, this->m_Ptr != NULL);

 private:
    template <class S, class D>
    bool operator==(const UniquePtr<S, D>& p) const;
    template <class S, class D>
    bool operator!=(const UniquePtr<S, D>& p) const;
    NLIB_DISALLOW_COPY_AND_ASSIGN(UniquePtr);
};

template <class T1, class D1, class T2, class D2>
bool operator==(const UniquePtr<T1, D1>& rhs, const UniquePtr<T2, D2>& lhs) NLIB_NOEXCEPT {
    return rhs.get() == lhs.get();
}

template <class T1, class D1>
bool operator==(const UniquePtr<T1, D1>& lhs, nullptr_t) NLIB_NOEXCEPT {
    return !lhs;
}

template <class T1, class D1>
bool operator==(nullptr_t, const UniquePtr<T1, D1>& rhs) NLIB_NOEXCEPT {
    return !rhs;
}

template <class T1, class D1, class T2, class D2>
bool operator!=(const UniquePtr<T1, D1>& rhs, const UniquePtr<T2, D2>& lhs) NLIB_NOEXCEPT {
    return rhs.get() != lhs.get();
}

template <class T1, class D1>
bool operator!=(const UniquePtr<T1, D1>& lhs, nullptr_t) NLIB_NOEXCEPT {
    return static_cast<bool>(lhs);
}

template <class T1, class D1>
bool operator!=(nullptr_t, const UniquePtr<T1, D1>& rhs) NLIB_NOEXCEPT {
    return static_cast<bool>(rhs);
}

NLIB_NAMESPACE_END

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#else
NLIB_DEFINE_STD_SWAP_T_BEGIN1(std)  // NOLINT
#endif

NLIB_DEFINE_STD_SWAP_T2(T, DEL, NLIB_NS::UniquePtr)  // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#else
NLIB_DEFINE_STD_SWAP_T_END1(std)  // NOLINT
#endif

#ifndef NLIB_CXX11_TEMPLATE_ALIAS
NLIB_NAMESPACE_BEGIN
#else
NLIB_DEFINE_STD_HASH_T_BEGIN1(std)  // NOLINT
#endif

NLIB_DEFINE_STD_HASH_T2(T, DEL, NLIB_NS::UniquePtr)  // NOLINT

#ifndef NLIB_CXX11_TEMPLATE_ALIAS
NLIB_NAMESPACE_END
#else
NLIB_DEFINE_STD_HASH_T_END1(std)  // NOLINT
#endif

#endif

NLIB_NAMESPACE_BEGIN

/*
template<class T, void FreeFunc(void*) = free>
struct FreeDeleterT {
    void operator()(T* p) const NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(T));
        NLIB_STATIC_ASSERT(IsPod<T>::value);
        if (p) FreeFunc(p);
    }
};
*/

namespace detail {

template <class T, class ALLOC, bool DEL_EMPTY>
class DefaultDeleteExBase {
 public:
    DefaultDeleteExBase() : al() {}
    explicit DefaultDeleteExBase(const ALLOC& al_) : al(al_) {}

 protected:
    ALLOC& get_allocator() { return al; }
    ALLOC al;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(DefaultDeleteExBase);
};

template <class T, class ALLOC>
class DefaultDeleteExBase<T, ALLOC, true> : public ALLOC {
 public:
    DefaultDeleteExBase() {}
    explicit DefaultDeleteExBase(const ALLOC& al_) { NLIB_UNUSED(al_); }

 protected:
    ALLOC& get_allocator() { return *this; }
};

}  // namespace detail

template <class T, class ALLOC = std::allocator<T> >
struct DefaultDeleteEx : public detail::DefaultDeleteExBase<T, ALLOC, IsEmpty<ALLOC>::value> {
    void operator()(T* p)NLIB_NOEXCEPT {
        NLIB_STATIC_ASSERT(sizeof(T));
        NLIB_STATIC_ASSERT(!IsArray<T>::value);
        typedef detail::DefaultDeleteExBase<T, ALLOC, IsEmpty<ALLOC>::value> BaseType;
        if (p) {
            ALLOC& al = BaseType::get_allocator();
            al.destroy(p);
            al.deallocate(p, 1);
        }
    }
};

template <class T, class ALLOC>
struct DefaultDeleteEx<T[], ALLOC>
    : public detail::DefaultDeleteExBase<T, ALLOC, IsEmpty<ALLOC>::value> {
    // mainly for using malloc variant like nmalloc
    // ex.
    // UniquePtr<
    //   unsigned char[],
    //   DefaultDeleteEx<unsigned char[], NMallocStlAllocator<unsigned char> >
    // >
    void operator()(T* p)NLIB_NOEXCEPT {
        typedef detail::DefaultDeleteExBase<T, ALLOC, IsEmpty<ALLOC>::value> BaseType;
        NLIB_STATIC_ASSERT(sizeof(T));
        NLIB_STATIC_ASSERT(IsPod<T>::value);
        if (p) {
            ALLOC& al = BaseType::get_allocator();
            // The length is unknown
            al.deallocate(p, 1);
        }
    }
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_UNIQUEPTR_H_
