﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_URI_H_
#define INCLUDE_NN_NLIB_URI_H_

#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Nlist.h"

NLIB_NAMESPACE_BEGIN

class InputStream;
class OutputStream;
class MemoryOutputStream;

class Uri NLIB_FINAL {
 public:
    // reserved    = gen-delims / sub-delims
    // gen - delims = ":" / "/" / "?" / "#" / "[" / "]" / "@"
    // sub - delims = "!" / "$" / "&" / "'" / "(" / ")" / "*" / "+" / "," / ";" / "="
    NLIB_VIS_PUBLIC static bool IsReserved(int c) NLIB_NOEXCEPT;
    // unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
    NLIB_VIS_PUBLIC static bool IsUnreserved(int c) NLIB_NOEXCEPT;
    // 0-9 / A-F / a-f
    NLIB_VIS_PUBLIC static bool IsHexDigit(int c) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC static errno_t DecodeUriComponent(InputStream* is,
                                                      OutputStream* os) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC static errno_t EncodeUriComponent(InputStream* is, OutputStream* os,
                                                      bool fragment_mode = false) NLIB_NOEXCEPT;

    NLIB_VIS_PUBLIC static errno_t DecodeUriComponent(size_t* written, char* buf, size_t n,
                                                      const char* s) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC static errno_t EncodeUriComponent(size_t* written, char* buf, size_t n,
                                                      const char* s,
                                                      bool fragment_mode) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC static errno_t DecodePath(size_t* written, char* buf, size_t n,
                                              const char* s) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC static errno_t EncodePath(size_t* written, char* buf, size_t n,
                                              const char* s) NLIB_NOEXCEPT;

    template <size_t N>
    static errno_t
        DecodeUriComponent(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, N, s);
    }
    template <size_t N>
    static errno_t
        EncodeUriComponent(size_t* written, char (&buf)[N], const char* s,
                           bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, N, s, fragment_mode);
    }
    template <size_t N>
    static errno_t
        DecodePath(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodePath(written, buf, N, s);
    }
    template <size_t N>
    static errno_t
        EncodePath(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return EncodePath(written, buf, N, s);
    }

 public:
    NLIB_VIS_PUBLIC Uri() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC ~Uri() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Parse(const char* str) NLIB_NOEXCEPT;
    // for example:
    //   http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag
    // scheme = "http"
    // userinfo = "foobar:pass"
    // host = "www.example.com"
    // port = "80"
    // path = "/my/dir/index.html"
    // query = "var=value"
    // fragment = "frag"
    NLIB_VIS_PUBLIC bool SetUri(const char* scheme, const char* userinfo, const char* host,
                                const char* port, const char* path, const char* query,
                                const char* fragment) NLIB_NOEXCEPT;

    NLIB_VIS_PUBLIC errno_t ComposeString(OutputStream* os) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool ComposeString(char* buf, size_t size) const NLIB_NOEXCEPT;
    template <size_t N>
    bool ComposeString(char (&buf)[N]) const NLIB_NOEXCEPT {
        return this->ComposeString(buf, N);
    }

    // AddBaseUri("g/", "http://a/b/c/d;p?q") -> http://a/b/c/g/
    // AddBaseUri("/g", "http://a/b/c/d;p?q") -> http://a/g
    // AddBaseUri("//g", "http://a/b/c/d;p?q") -> http://g
    NLIB_VIS_PUBLIC bool AddBaseUri(const Uri& relative, const Uri& base) NLIB_NOEXCEPT;
    // "http" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetScheme() const NLIB_NOEXCEPT { return m_Scheme.get(); }
    // "foobar:pass" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetUserInfo() const NLIB_NOEXCEPT { return m_UserInfo.get(); }
    // "www.example.com" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetHost() const NLIB_NOEXCEPT { return m_Host.get(); }
    // "80" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetPort() const NLIB_NOEXCEPT { return m_Port.get(); }
    // "/my/dir/index.html" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetPath() const NLIB_NOEXCEPT { return m_Path.get(); }
    // "var=value" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetQuery() const NLIB_NOEXCEPT { return m_Query.get(); }
    // "frag" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetFragment() const NLIB_NOEXCEPT { return m_Fragment.get(); }
    void Reset() NLIB_NOEXCEPT;

 private:
    NLIB_VIS_HIDDEN bool SetScheme_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetUserInfo_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetHost_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetPort_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetPath_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetQuery_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetFragment_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool DirectSet(UniquePtr<char[]>* field, const char* str) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool MergePath(const char* base, const char* relative) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool RemoveDotSegments(const char* path) NLIB_NOEXCEPT;

 private:
    UniquePtr<char[]> m_Scheme;    // unspecified if NULL, empty string not allowed.
    UniquePtr<char[]> m_UserInfo;  // unspecified if NULL, empty string not allowed.
    UniquePtr<char[]> m_Host;      // unspecified if NULL, empty string for empty string host.
    UniquePtr<char[]> m_Port;      // unspecified if NULL, empty string not allowed.
    UniquePtr<char[]> m_Path;      // empty path if empty string, NULL is not allowed.
    UniquePtr<char[]> m_Query;     // unspecified if NULL, empty string for empty string query
    UniquePtr<char[]> m_Fragment;  // unspecified if NULL, empty string for empty string fragment
};

NLIB_NAMESPACE_END

NLIB_NAMESPACE_BEGIN
namespace detail {

class Query NLIB_FINAL {
 public:
    typedef std::pair<const char*, const char*> KeyValue;
    Query() NLIB_NOEXCEPT {}
    NLIB_VIS_PUBLIC ~Query() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Set(const char* query) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Add(const char* key, const char* value) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Compose(char** p, char* end) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC const char* GetValue(const char* key) const NLIB_NOEXCEPT;
    size_t GetNumQuery() const NLIB_NOEXCEPT { return m_Query.size(); }
    const KeyValue* GetKeyValue(size_t i) const NLIB_NOEXCEPT {
        return (i < m_Query.size()) ? &m_Query[i] : NULL;
    }
    void Reset() NLIB_NOEXCEPT { m_Query.clear(); }

 private:
    Nlist<KeyValue> m_Query;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Query);
};

class Segments NLIB_FINAL {
    // NOTE:
    // Each segments holds the byte sequence decoded completely.
 public:
    NLIB_VIS_PUBLIC Segments() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC ~Segments() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool SetPath(const char* str) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool ComposePath(char** p, char* buf_end) const NLIB_NOEXCEPT;
    bool IsAbsolutePath() const NLIB_NOEXCEPT { return m_IsAbsolutePath; }
    bool IsDirectory() const NLIB_NOEXCEPT { return m_IsDirectory; }
    bool IsEmpty() const NLIB_NOEXCEPT {
        return m_Segment.empty() && !m_IsAbsolutePath && !m_IsDirectory;
    }
    const char* GetSegment(size_t i) const NLIB_NOEXCEPT {
        return i < m_Segment.size() ? m_Segment[i].get() : NULL;
    }
    size_t GetNumSegment() const NLIB_NOEXCEPT { return m_Segment.size(); }

 private:
    bool m_IsAbsolutePath;
    bool m_IsDirectory;
    Nlist<UniquePtr<char[]> > m_Segment;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Segments);
};

}  // namespace detail
NLIB_NAMESPACE_END


#endif  // INCLUDE_NN_NLIB_URI_H_
