﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_HEAP_FRAMEHEAP_H_
#define INCLUDE_NN_NLIB_HEAP_FRAMEHEAP_H_

#include "nn/nlib/Config.h"

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace heap {

class NLIB_VIS_PUBLIC FrameHeap NLIB_FINAL {
 public:
    typedef void (*DestructorHandler)(void* start, void* end);
    enum AllocOption {
        ALLOCOPTION_ALLOCHEAD = 0x00000000,
        ALLOCOPTION_ALLOCTAIL = 0x00000001
    };

    struct State {
        size_t _[2];
        bool operator==(const State& rhs) NLIB_NOEXCEPT {
            return _[0] == rhs._[0] && _[1] == rhs._[1];
        }
        bool operator!=(const State& rhs) NLIB_NOEXCEPT {
            return _[0] != rhs._[0] || _[1] != rhs._[1];
        }
    };
    FrameHeap() NLIB_NOEXCEPT : m_HeapOption(0), m_Start(NULL), m_End(NULL), m_Handler(NULL) {
        m_Cur.interlocked._[0] = m_Cur.interlocked._[1] = 0;
    }
    ~FrameHeap() NLIB_NOEXCEPT {
        if (m_Handler) {
            (*m_Handler)(m_Start, m_End);
        }
    }
    errno_t Initialize(void* start, size_t size, DestructorHandler handler,
                       uint32_t heap_option) NLIB_NOEXCEPT;
    void FreeAll() NLIB_NOEXCEPT {
        HeapStatus to;
        to.tb.topidx = 0;
        to.tb.botidx = m_End - m_Start;
        m_Cur = to;
    }
    void FreeTop() NLIB_NOEXCEPT { m_Cur.tb.topidx = 0; }
    void FreeBot() NLIB_NOEXCEPT {
        size_t bot = m_End - m_Start;
        m_Cur.tb.botidx = bot;
    }
    errno_t ResizeTop(void* p, size_t from_size, size_t to_size) NLIB_NOEXCEPT;
    errno_t Adjust(void** start, size_t* size) NLIB_NOEXCEPT;

    State SaveState() NLIB_NOEXCEPT { return m_Cur.interlocked; }
    void RestoreState(State state) NLIB_NOEXCEPT { m_Cur.interlocked = state; }
    void* Alloc(size_t n) NLIB_NOEXCEPT { return this->Alloc(n, 8, 0); }
    void* Alloc(size_t n, size_t algn) NLIB_NOEXCEPT { return this->Alloc(n, algn, 0); }
    void* Alloc(size_t n, size_t algn, uint32_t alloc_option) NLIB_NOEXCEPT;
    void Dump() NLIB_NOEXCEPT;

 private:
    union HeapStatus {
        State interlocked;
        struct topbot_ {
            size_t topidx;
            size_t botidx;
        } tb;
    };
    static uint8_t* RoundUp(uint8_t* ptr, size_t algn) NLIB_NOEXCEPT {
        // algn is power of 2.
        intptr_t tmp = reinterpret_cast<intptr_t>(ptr);
        return reinterpret_cast<uint8_t*>(tmp + (-tmp & (algn - 1)));
    }
    static uint8_t* RoundDown(uint8_t* ptr, size_t algn) NLIB_NOEXCEPT {
        // algn is power of 2.
        intptr_t tmp = reinterpret_cast<intptr_t>(ptr);
        return reinterpret_cast<uint8_t*>(tmp & (~algn + 1));
    }
    void ChangeHeapRange(void* start, void* end) NLIB_NOEXCEPT {
        m_Start = reinterpret_cast<uint8_t*>(start);
        m_End = reinterpret_cast<uint8_t*>(end);
    }

 private:
    HeapStatus m_Cur;

    uint32_t m_HeapOption;
    uint8_t* m_Start;
    uint8_t* m_End;
    DestructorHandler m_Handler;
};

}  // namespace heap
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_heap_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_HEAP_FRAMEHEAP_H_
