﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_
#define INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_

#include <map>
#include <string>
#include <utility>
#include <vector>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/Cstring.h"
#include "nn/nlib/Nlist.h"

#ifdef NLIB_CXX11_STDLIB_ARRAY
#include <array>  // NOLINT
#endif

#ifdef NLIB_CXX11_STDLIB_UNORDERED
#include <unordered_map>  // NOLINT
#endif

#ifdef NLIB_CXX11_STDLIB_TUPLE
#include <tuple>  // NOLINT
#endif

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {

class MpObject;

template <class T>
errno_t Box(MpObject* obj, const T& v);

template <class T>
errno_t Unbox(const MpObject* obj, T* v);

struct nil {};

template <class T>
inline bool operator==(const nil& lhs, const T& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}
template <class T>
inline bool operator==(const T& lhs, const nil& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}
inline bool operator==(const nil& lhs, const nil& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
template <class T>
inline bool operator!=(const nil& rhs, const T& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
template <class T>
inline bool operator!=(const T& rhs, const nil& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
inline bool operator!=(const nil& rhs, const nil& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}

struct MpObjectKv;

class NLIB_VIS_PUBLIC MpObject NLIB_FINAL {
 private:
    union UnionType {
        uint64_t u64;
        int64_t i64;
        float f32;
        double f64;
        Nlist<MpObject>* ary;
        Nlist<MpObjectKv>* mp;
        void* raw;
    };

 public:
    MpObject* GetArrayItem(size_t n) NLIB_NOEXCEPT {
        MpObject* obj;
        return NLIB_LIKELY(0 == this->GetArrayItem(n, &obj)) ? obj : NULL;
    }
    const MpObject* GetArrayItem(size_t n) const NLIB_NOEXCEPT {
        const MpObject* obj;
        return NLIB_LIKELY(0 == this->GetArrayItem(n, &obj)) ? obj : NULL;
    }
    MpObject& operator[](size_t n) {
        NLIB_ASSERT(this->IsArray());
        NLIB_ASSERT(n < m_Size);
        return (*m_Via.ary)[n];
    }
    MpObject* GetMapItem(const char* str) NLIB_NOEXCEPT;
    const MpObject* GetMapItem(const char* str) const NLIB_NOEXCEPT;
    MpObject* GetMapItem(const std::string& str) NLIB_NOEXCEPT;
    const MpObject* GetMapItem(const std::string& str) const NLIB_NOEXCEPT;
    errno_t Resize(uint32_t n);
    MpObject* SwapArrayItem(size_t n, MpObject* obj) NLIB_NOEXCEPT NLIB_NONNULL {
        MpObject* p = this->GetArrayItem(n);
        if (NLIB_LIKELY(p)) p->swap(*obj);
        return p;
    }
    MpObject* SwapMapItem(const std::string& key,
                                             MpObject* obj) NLIB_NOEXCEPT NLIB_NONNULL {
        MpObject* p = this->GetMapItem(key);
        if (NLIB_LIKELY(p)) p->swap(*obj);
        return p;
    }
    MpObject* SwapMapItem(const char* str,
                                             MpObject* obj) NLIB_NOEXCEPT NLIB_NONNULL {
        MpObject* p = this->GetMapItem(str);
        if (NLIB_LIKELY(p)) p->swap(*obj);
        return p;
    }
    MpObject* AppendArrayItem() NLIB_NOEXCEPT;
    MpObjectKv* AppendMapItem() NLIB_NOEXCEPT;

 public:
    MpObject() NLIB_NOEXCEPT : m_Type(NIL) {}  // only do minimum memory access
    ~MpObject() NLIB_NOEXCEPT { ReleaseIfNeeded(); }
    NLIB_MOVE_MEMBER_HELPER_1(MpObject, m_Type);

    // NOTE: implicit call was error in gcc or clang....
    template <class T>
    explicit MpObject(const T& x) {
        MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Construct(this, x);
    }
    errno_t GetArrayItem(size_t n, MpObject** obj) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetArrayItem(size_t n,
                                         const MpObject** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetMapItem(size_t n, MpObjectKv** obj) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetMapItem(size_t n,
                                       const MpObjectKv** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    void* GetRaw() {
        if (m_Type == RAW) return m_Via.raw;
        if (NLIB_LIKELY(m_Type == RAW_INLINE)) return &m_Raw14Data;
        return NULL;
    }
    const void* GetRaw() const {
        if (m_Type == RAW) return m_Via.raw;
        if (NLIB_LIKELY(m_Type == RAW_INLINE)) return &m_Raw14Data;
        return NULL;
    }
    errno_t InitArray(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitMap(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitRaw(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitRaw(const void* p, uint32_t n) NLIB_NOEXCEPT;

 public:  // Box
    errno_t Box(const char* str) NLIB_NOEXCEPT NLIB_NONNULL;
    template <uint32_t n>
    errno_t Box(const char (&str)[n]) NLIB_NOEXCEPT;
    template <class T, uint32_t n>
    errno_t Box(const T (&vec)[n]);
#ifdef NLIB_CXX11_STDLIB_ARRAY
    template <class T, size_t n>
    errno_t Box(const std::array<T, n>& vec);
#endif
    template <class T>
    errno_t Box(const T& v) {
        return MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Box(this, v);
    }
    errno_t Box(const void* p, uint32_t n) NLIB_NOEXCEPT NLIB_NONNULL;

 public:  // Unbox
    template <class T, size_t n>
    errno_t Unbox(T (&a)[n]) const {
        return this->Unbox(&a[0], n);
    }
#ifdef NLIB_CXX11_STDLIB_ARRAY
    template <class T, size_t n>
    errno_t Unbox(std::array<T, n>& a) {  // NOLINT
        return this->Unbox(reinterpret_cast<T*>(&*a.begin()), n);
    }
#endif
    template <size_t n>
    errno_t Unbox(char (&str)[n]) const NLIB_NOEXCEPT {
        return this->Unbox(&str[0], n);
    }
    template <class T>
    errno_t Unbox(T* a, size_t n) const NLIB_NONNULL;
    errno_t Unbox(char* str, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t Unbox(void* p, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    template <class T>
    errno_t Unbox(T v) const {
        // T cannot be T* though v is only pointer type
        // because of conflict against array types.
        return MyBox<typename RemovePtr<T>::type,
                     typename IsIntegral<typename RemovePtr<T>::type>::type,
                     typename IsSigned<typename RemovePtr<T>::type>::type>::Unbox(this, v);
    }

 public:
    // You can use operator=() only for basic types(never fails).
    // Use Box() otherwise.
    template <class T>
    MpObject& operator=(const T& x) {
        MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Assign(this, x);
        return *this;
    }

    MpObject* Clone() NLIB_NOEXCEPT;
    void swap(MpObject& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(m_Type, rhs.m_Type);
        swap(m_Raw14Size, rhs.m_Raw14Size);
        swap(m_Raw14Data, rhs.m_Raw14Data);
        swap(_, rhs._);
        swap(m_Size, rhs.m_Size);
        swap(m_Via.u64, rhs.m_Via.u64);
    }

 public:
    enum ObjectType {
        NIL = 0,
        BOOLEAN,
        UINT64,
        INT64,
        FLOAT,
        DOUBLE,
        RAW,
        ARRAY,
        MAP,
        RAW_INLINE = 0x80 | RAW,
        BOOLEAN_FALSE = 0x00 | BOOLEAN,
        BOOLEAN_TRUE = 0x80 | BOOLEAN
    };
    ObjectType GetType() const NLIB_NOEXCEPT {
        return static_cast<ObjectType>(m_Type & 0x7F);
    }
    uint32_t GetSize() const NLIB_NOEXCEPT {
        return (m_Type != RAW_INLINE) ? m_Size : m_Raw14Size;
    }
    bool IsNil() const NLIB_NOEXCEPT { return GetType() == NIL; }
    bool IsBoolean() const NLIB_NOEXCEPT { return GetType() == BOOLEAN; }
    bool IsInteger() const NLIB_NOEXCEPT {
        ObjectType tp = GetType();
        return tp == UINT64 || tp == INT64;
    }
    bool IsFloat() const NLIB_NOEXCEPT { return GetType() == FLOAT; }
    bool IsDouble() const NLIB_NOEXCEPT { return GetType() == DOUBLE; }
    bool IsRaw() const NLIB_NOEXCEPT { return GetType() == RAW; }
    bool IsArray() const NLIB_NOEXCEPT { return GetType() == ARRAY; }
    bool IsMap() const NLIB_NOEXCEPT { return GetType() == MAP; }

 private:
    errno_t Clone(MpObject* p) NLIB_NOEXCEPT;
    template <class T, class IS_INTEGRAL, class IS_SIGNED>
    struct MyBox {
        static errno_t Box(MpObject* This, const T& v) { return ::nlib_ns::msgpack::Box(This, v); }
        static errno_t Unbox(const MpObject* This, T* v) {
            if (!v) return EINVAL;
            return ::nlib_ns::msgpack::Unbox(This, v);
        }
        static void Assign(MpObject* This, const T& x);
        static void Construct(MpObject* This, const T& x);
    };
    void ReleaseIfNeeded() NLIB_NOEXCEPT {
        if (m_Type >= RAW && m_Type <= MAP) ReleaseIfNeeded_();
    }
    void ReleaseIfNeeded_() NLIB_NOEXCEPT;

 private:
    uint8_t m_Type;
    uint8_t m_Raw14Size;
    uint8_t m_Raw14Data;
    uint8_t _;
    uint32_t m_Size;
    UnionType m_Via;
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpObject);
    friend NLIB_VIS_PUBLIC bool operator==(const MpObject& lhs, const MpObject& rhs);
};

struct MpObjectKv {
    MpObject first;
    MpObject second;

 public:
    MpObjectKv() NLIB_NOEXCEPT {}
    ~MpObjectKv() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_2(MpObjectKv, first, second);
    void swap(MpObjectKv& rhs) NLIB_NOEXCEPT {  // NOLINT
        first.swap(rhs.first);
        second.swap(rhs.second);
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpObjectKv);
};

template <>
struct MpObject::MyBox<nil, FalseType, FalseType> {
    static errno_t Box(MpObject* This, nil v) NLIB_NOEXCEPT {
        NLIB_UNUSED(v);
        This->ReleaseIfNeeded();
        This->m_Type = NIL;
        return 0;
    }
    static void Assign(MpObject* This, nil x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, nil) NLIB_NOEXCEPT { This->m_Type = NIL; }
};

template <>
struct MpObject::MyBox<bool, TrueType, FalseType> {
    static errno_t Box(MpObject* This, bool v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->m_Type = v ? static_cast<uint8_t>(BOOLEAN_TRUE) : static_cast<uint8_t>(BOOLEAN_FALSE);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, bool* v) NLIB_NOEXCEPT {
        if (This->m_Type == BOOLEAN_FALSE) {
            *v = false;
            return 0;
        } else if (NLIB_LIKELY(This->m_Type == BOOLEAN_TRUE)) {
            *v = true;
            return 0;
        }
        *v = false;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, bool x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, bool x) NLIB_NOEXCEPT {
        This->m_Type = x ? static_cast<uint8_t>(BOOLEAN_TRUE) : static_cast<uint8_t>(BOOLEAN_FALSE);
    }
};

template <class T>
struct MpObject::MyBox<T, TrueType, TrueType> {
    static errno_t Box(MpObject* This, T v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->m_Type = INT64;
        This->m_Via.i64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, T* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        if (This->m_Type == INT64) {
            *v = static_cast<T>(This->m_Via.i64);
            return 0;
        } else if (NLIB_LIKELY(This->m_Type == UINT64)) {
            *v = static_cast<T>(This->m_Via.u64);
            return 0;
        }
        *v = 0;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, T x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, T x) NLIB_NOEXCEPT {
        This->m_Type = INT64;
        This->m_Via.i64 = x;
    }
};

template <class T>
struct MpObject::MyBox<T, TrueType, FalseType> {
    static errno_t Box(MpObject* This, T v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->m_Type = UINT64;
        This->m_Via.u64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, T* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        if (This->m_Type == INT64) {
            *v = static_cast<T>(This->m_Via.i64);
            return 0;
        } else if (NLIB_LIKELY(This->m_Type == UINT64)) {
            *v = static_cast<T>(This->m_Via.u64);
            return 0;
        }
        *v = 0;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, T x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, T x) NLIB_NOEXCEPT {
        This->m_Type = UINT64;
        This->m_Via.u64 = x;
    }
};

template <class TR1_WORKAROUND>
struct MpObject::MyBox<float, FalseType, TR1_WORKAROUND> {
    static errno_t Box(MpObject* This, float v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->m_Type = FLOAT;
        This->m_Via.f32 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, float* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        switch (This->m_Type) {
        case FLOAT:
            *v = This->m_Via.f32;
            break;
        case DOUBLE:
            *v = static_cast<float>(This->m_Via.f64);
            break;
        default:
            *v = 0;  // to suppress warnings
            return EACCES;
        }
        return 0;
    }
    static void Assign(MpObject* This, float x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, float x) NLIB_NOEXCEPT {
        This->m_Type = FLOAT;
        This->m_Via.f32 = x;
    }
};

template <class TR1_WORKAROUND>
struct MpObject::MyBox<double, FalseType, TR1_WORKAROUND> {
    static errno_t Box(MpObject* This, double v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->m_Type = DOUBLE;
        This->m_Via.f64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, double* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        switch (This->m_Type) {
        case FLOAT:
            *v = This->m_Via.f32;
            break;
        case DOUBLE:
            *v = This->m_Via.f64;
            break;
        default:
            *v = 0;  // to suppress warnings
            return EACCES;
        }
        return 0;
    }
    static void Assign(MpObject* This, double x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, double x) NLIB_NOEXCEPT {
        This->m_Type = DOUBLE;
        This->m_Via.f64 = x;
    }
};

template <>
struct MpObject::MyBox<std::string, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::string& str) NLIB_NOEXCEPT {
        uint32_t n = static_cast<uint32_t>(str.size());
        errno_t e = This->InitRaw(n);
        if (NLIB_UNLIKELY(e != 0)) return e;
        nlib_memcpy(This->GetRaw(), n, &str[0], n);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::string* str) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!str)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != RAW) && NLIB_UNLIKELY(This->m_Type != RAW_INLINE))
            return EACCES;
        const char* chr = reinterpret_cast<const char*>(This->GetRaw());
        NLIB_TRY { (*str).assign(chr, This->GetSize()); }
        NLIB_CATCH(std::bad_alloc&) { return ENOMEM; }
        return 0;
    }
};

template <class T, class ALLOC>
struct MpObject::MyBox<std::vector<T, ALLOC>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::vector<T, ALLOC>& vec) {
        uint32_t n = static_cast<uint32_t>(vec.size());
        MpObject tmp;
        errno_t e = tmp.InitArray(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        for (uint32_t i = 0; i < n; ++i) {
            e = (*tmp.m_Via.ary)[i].Box(vec[i]);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::vector<T, ALLOC>* vec) {
        if (NLIB_UNLIKELY(!vec)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != ARRAY)) return EACCES;
        NLIB_TRY { vec->resize(This->m_Size); }
        NLIB_CATCH(std::bad_alloc&) { return ENOMEM; }
        for (uint32_t i = 0; i < This->m_Size; ++i) {
            errno_t e = (*This->m_Via.ary)[i].Unbox(&(*vec)[i]);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *vec NOT RESET
        }
        return 0;
    }
};

template <class T, class ALLOC>
struct MpObject::MyBox<Nlist<T, ALLOC>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const Nlist<T, ALLOC>& vec) {
        size_t n = vec.size();
        MpObject tmp;
        errno_t e = tmp.InitArray(static_cast<uint32_t>(n));
        if (NLIB_UNLIKELY(e != 0)) return e;
        typename Nlist<T, ALLOC>::const_iterator it = vec.begin();
        for (size_t i = 0; i < n; ++it, ++i) {
            e = (*tmp.m_Via.ary)[i].Box(*it);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, Nlist<T, ALLOC>* vec) {
        if (NLIB_UNLIKELY(!vec)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != ARRAY)) return EACCES;
        vec->clear();
        if (NLIB_UNLIKELY(!vec->reserve(This->m_Size))) return ENOMEM;
        for (uint32_t i = 0; i < This->m_Size; ++i) {
            T* item = vec->push_back();
            errno_t e = (*This->m_Via.ary)[i].Unbox(item);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *vec NOT RESET
        }
        return 0;
    }
};

template <class T1, class T2>
struct MpObject::MyBox<std::pair<T1, T2>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::pair<T1, T2>& p) {
        MpObject tmp;
        errno_t e = tmp.InitArray(2);
        if (NLIB_UNLIKELY(e != 0)) return e;
        e = (*tmp.m_Via.ary)[0].Box(p.first);
        if (NLIB_UNLIKELY(e != 0)) return e;
        e = (*tmp.m_Via.ary)[1].Box(p.second);
        if (NLIB_UNLIKELY(e != 0)) return e;
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::pair<T1, T2>* p) {
        if (NLIB_UNLIKELY(!p)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != ARRAY)) return EACCES;
        if (NLIB_UNLIKELY(This->m_Size != 2)) return EACCES;
        errno_t e;
        e = (*This->m_Via.ary)[0].Unbox(&p->first);
        if (NLIB_UNLIKELY(e != 0)) return e;  // *p NOT RESET
        e = (*This->m_Via.ary)[1].Unbox(&p->second);
        if (NLIB_UNLIKELY(e != 0)) return e;  // *p NOT RESET
        return 0;
    }
};

#if defined(NLIB_CXX11_STDLIB_TUPLE) && defined(NLIB_CXX11_VARIADIC_TEMPLATES)
template <class ...Args>
struct MpObject::MyBox<std::tuple<Args...>, FalseType, FalseType > {
 private:
    template<size_t N, class HEAD>
    static errno_t BoxHelper(MpObject* tmp, const HEAD& head) {
        return (*tmp->m_Via.ary)[N].Box(head);
    }
    template<size_t N, class HEAD, class ...TAIL>
    static errno_t BoxHelper(MpObject* tmp, const HEAD& head, TAIL&... tail) {
        errno_t e = (*tmp->m_Via.ary)[N].Box(head);
        if (NLIB_UNLIKELY(e != 0)) return e;
        return BoxHelper<N + 1, TAIL...>(tmp, tail...);
    }
    template<size_t N, class HEAD>
    static errno_t UnboxHelper(const MpObject* This, HEAD& head) {  // NOLINT
        return (*This->m_Via.ary)[N].Unbox(&head);
    }
    template<size_t N, class HEAD, class ...TAIL>
    static errno_t UnboxHelper(const MpObject* This, HEAD& head, TAIL&... tail) {  // NOLINT
        errno_t e = (*This->m_Via.ary)[N].Unbox(&head);
        if (NLIB_UNLIKELY(e != 0)) return e;
        return UnboxHelper<N + 1, TAIL...>(This, tail...);
    }
    template<int ...> struct seq {};
    template<int N, int ...S> struct gens : gens<N - 1, N - 1, S...> {};
    template<int ...S>
    struct gens<0, S...> {
        typedef seq<S...> type;
    };
    template<int ...S>
    static errno_t Box(MpObject* tmp, seq<S...>, const std::tuple<Args...>& p) {
        return BoxHelper<0>(tmp, std::get<S>(p)...);
    }
    template<int ...S>
    static errno_t Unbox(const MpObject* This, seq<S...>, std::tuple<Args...>& p) {  // NOLINT
        return UnboxHelper<0>(This, std::get<S>(p)...);
    }

 public:
    static errno_t Box(MpObject* This, const std::tuple<Args...>& p) {
        size_t count = std::tuple_size<std::tuple<Args...> >::value;
        MpObject tmp;
        errno_t e = tmp.InitArray(static_cast<uint32_t>(count));
        if (NLIB_UNLIKELY(e != 0)) return e;
        typedef typename gens<sizeof...(Args)>::type MySeq;
        e = Box(&tmp, MySeq(), p);
        if (NLIB_UNLIKELY(e != 0)) return e;
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::tuple<Args...>* p) {
        if (NLIB_UNLIKELY(!p)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != ARRAY)) return EACCES;
        if (NLIB_UNLIKELY(This->m_Size != std::tuple_size<std::tuple<Args...> >::value))
            return EACCES;
        typedef typename gens<sizeof...(Args)>::type MySeq;
        return Unbox(This, MySeq(), *p);
    }
};
#endif

template <class K, class V, class Pr, class Alloc>
struct MpObject::MyBox<std::map<K, V, Pr, Alloc>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::map<K, V, Pr, Alloc>& m) {
        typedef typename std::map<K, V, Pr, Alloc> maptype;
        uint32_t n = static_cast<uint32_t>(m.size());
        MpObject tmp;
        errno_t e = tmp.InitMap(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        typename maptype::const_iterator it;
        typename maptype::const_iterator it_end = m.end();
        uint32_t i;
        for (i = 0, it = m.begin(); it != it_end; ++it, ++i) {
            MpObjectKv& kv = (*tmp.m_Via.mp)[i];
            e = kv.first.Box(it->first);
            if (NLIB_UNLIKELY(e != 0)) return e;
            e = kv.second.Box(it->second);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::map<K, V, Pr, Alloc>* m) {
        if (NLIB_UNLIKELY(!m)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != MAP)) return EACCES;
        m->clear();
        for (uint32_t i = 0; i < This->m_Size; ++i) {
            K key;
            V value;
            const MpObjectKv& kv = (*This->m_Via.mp)[i];
            errno_t e;
            e = kv.first.Unbox(&key);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            e = kv.second.Unbox(&value);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            NLIB_TRY { (*m)[NLIB_MOVE(key)] = NLIB_MOVE(value); }
            NLIB_CATCH(...) { return ENOMEM; }
        }
        return 0;
    }
};

#ifdef NLIB_CXX11_STDLIB_UNORDERED
template <class K, class V, class Hash, class Pr, class Alloc>
struct MpObject::MyBox<std::unordered_map<K, V, Hash, Pr, Alloc>, FalseType, FalseType > {
    static errno_t Box(MpObject* This, const std::unordered_map<K, V, Hash, Pr, Alloc>& m) {
        typedef typename std::unordered_map<K, V, Hash, Pr, Alloc> maptype;
        uint32_t n = static_cast<uint32_t>(m.size());
        MpObject tmp;
        errno_t e = tmp.InitMap(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        typename maptype::const_iterator it;
        typename maptype::const_iterator it_end = m.end();
        uint32_t i;
        for (i = 0, it = m.begin(); it != it_end; ++it, ++i) {
            MpObjectKv& kv = (*tmp.m_Via.mp)[i];
            e = kv.first.Box(it->first);
            if (NLIB_UNLIKELY(e != 0)) return e;
            e = kv.second.Box(it->second);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::unordered_map<K, V, Hash, Pr, Alloc>* m) {
        if (NLIB_UNLIKELY(!m)) return EINVAL;
        if (NLIB_UNLIKELY(This->m_Type != MAP)) return EACCES;
        m->clear();
        for (uint32_t i = 0; i < This->m_Size; ++i) {
            K key;
            V value;
            const MpObjectKv& kv = (*This->m_Via.mp)[i];
            errno_t e;
            e = kv.first.Unbox(&key);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            e = kv.second.Unbox(&value);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            NLIB_TRY { (*m)[NLIB_MOVE(key)] = NLIB_MOVE(value); }
            NLIB_CATCH(...) { return ENOMEM; }
        }
        return 0;
    }
};
#endif

NLIB_VIS_PUBLIC bool operator==(const MpObject& lhs, const MpObject& rhs);
inline bool operator==(const MpObjectKv& lhs, const MpObjectKv& rhs) NLIB_NOEXCEPT {
    return lhs.first == rhs.first && lhs.second == rhs.second;
}
inline bool operator!=(const MpObject& lhs, const MpObject& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}
inline bool operator!=(const MpObjectKv& lhs, const MpObjectKv& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}

template <uint32_t n>
errno_t MpObject::Box(const char (&str)[n]) NLIB_NOEXCEPT {
    uint32_t nn = static_cast<uint32_t>(nlib_strlen(str));
    errno_t e = this->InitRaw(nn);
    if (NLIB_UNLIKELY(e != 0)) return e;
    nlib_memcpy(this->GetRaw(), nn, reinterpret_cast<const void*>(&str[0]), nn);
    return 0;
}

template <class T, uint32_t n>
errno_t MpObject::Box(const T (&vec)[n]) {
    MpObject tmp;
    errno_t e = tmp.InitArray(n);
    if (NLIB_UNLIKELY(e != 0)) return e;
    for (uint32_t i = 0; i < n; ++i) {
        e = (*tmp.m_Via.ary)[i].Box(vec[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    this->swap(tmp);
    return 0;
}

#ifdef NLIB_CXX11_STDLIB_ARRAY
template <class T, size_t n>
errno_t MpObject::Box(const std::array<T, n>& vec) {
    MpObject tmp;
    errno_t e = tmp.InitArray(n);
    if (NLIB_UNLIKELY(e != 0)) return e;
    for (uint32_t i = 0; i < n; ++i) {
        e = (*tmp.m_Via.ary)[i].Box(vec[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    this->swap(tmp);
    return 0;
}
#endif

template <class T>
errno_t MpObject::Unbox(T* a, size_t n) const {
    NLIB_EINVAL_IFNULL(a);
    if (NLIB_UNLIKELY(n == 0)) return EINVAL;
    if (NLIB_UNLIKELY(m_Type != ARRAY)) return EACCES;
    if (NLIB_UNLIKELY(n < m_Size)) return ENOMEM;
    for (uint32_t i = 0; i < m_Size; ++i) {
        errno_t e = (*m_Via.ary)[i].Unbox(&a[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    return 0;
}

}  // namespace msgpack
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::MpObject)
NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::MpObjectKv)

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_
