﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_MPREADER_H_
#define INCLUDE_NN_NLIB_MSGPACK_MPREADER_H_

#include <map>
#include <string>
#include <utility>
#include <vector>

#include "nn/nlib/InputStream.h"
#include "nn/nlib/msgpack/MpObject.h"
#include "nn/nlib/Nlist.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {

class MpReader;
template <class T>
bool MpRead(MpReader* obj, T* v);

struct NLIB_VIS_PUBLIC MpReaderSettings {
    size_t maxArraySize;
    size_t maxMapSize;
    size_t maxRawSize;

 public:
    NLIB_CEXPR MpReaderSettings(size_t max_array_size_, size_t max_map_size_,
                                size_t max_raw_size_) NLIB_NOEXCEPT
                                : maxArraySize(max_array_size_),
                                  maxMapSize(max_map_size_),
                                  maxRawSize(max_raw_size_) {}
    NLIB_CEXPR MpReaderSettings() NLIB_NOEXCEPT : maxArraySize(1024 * 64),
                                                  maxMapSize(1024),
                                                  maxRawSize(1024 * 512) {}
};

class MpReader NLIB_FINAL {
 public:
    NLIB_VIS_PUBLIC MpReader() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC ~MpReader() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Init(InputStream* istr,
                              const MpReaderSettings& settings) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Init(InputStream* istr) NLIB_NOEXCEPT NLIB_NONNULL {
        return this->Init(istr, MpReaderSettings());
    }
    bool Close() NLIB_NOEXCEPT {
        m_Istr = NULL;
        return true;
    }

 public:
    static NLIB_VIS_PUBLIC bool Read(MpObject* obj, const void* p, size_t n,
                                     const MpReaderSettings& settings) NLIB_NOEXCEPT NLIB_NONNULL;
    static bool Read(MpObject* obj, const void* p, size_t n) NLIB_NOEXCEPT NLIB_NONNULL {
        return Read(obj, p, n, MpReaderSettings());
    }
    void SetError(errno_t e) NLIB_NOEXCEPT {
        if (m_ErrorValue == 0) m_ErrorValue = e;
    }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return m_ErrorValue; }
    NLIB_VIS_PUBLIC bool GetNextValue() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool GetNextArrayNum(size_t* num) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool GetNextMapNum(size_t* num) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool GetNextRawNum(size_t* num) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool GetNextRawBody(void* p, size_t num) NLIB_NOEXCEPT;

    template <class T>
    errno_t Convert(T* v) {
        return this->Convert_(v);
    }

    template <class T>
    bool Read(T v) {
        // T cannot be T* though v is only pointer type
        // because of conflict against array types.
        return Read_(v);
    }
    template <size_t n>
    bool Read(char (&v)[n]) NLIB_NOEXCEPT {
        return this->Read(&v[0], n);  // n is OK
    }
    template <class T, size_t n>
    bool Read(T (&v)[n]) {
        return this->Read(&v[0], n);
    }

    NLIB_VIS_PUBLIC bool Read(char* v, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    NLIB_VIS_PUBLIC bool Read(void* v, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    template <class T>
    bool Read(T* v, size_t n) NLIB_NONNULL;
    bool Read(MpObject* obj) NLIB_NOEXCEPT NLIB_NONNULL {
        return this->Read_(&obj);
    }

    NLIB_VIS_PUBLIC bool SkipMpObject() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool CheckString(const char* key, size_t keylen) NLIB_NOEXCEPT NLIB_NONNULL;
    bool CheckString(const char* key) NLIB_NOEXCEPT NLIB_NONNULL {
        return CheckString(key, nlib_strlen(key));
    }
    NLIB_SAFE_BOOL(MpReader, GetErrorValue() == 0);

 private:
    NLIB_VIS_HIDDEN bool GetU32(uint32_t* pv) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool GetU64(uint64_t* pv) NLIB_NOEXCEPT;

    NLIB_VIS_PUBLIC errno_t Convert_(int8_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(int16_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(int32_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(int64_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(uint8_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(uint16_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(uint32_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(uint64_t* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(float* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(double* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(bool* v) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC errno_t Convert_(nil* v) NLIB_NOEXCEPT;
    template <class T>
    errno_t Convert_(T* v);

#define NLIB_EXEC_READ return GetNextValue() ? (void)this->Convert(v), true : false
    NLIB_VIS_PUBLIC bool Read_(int8_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(int16_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(int32_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(int64_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(uint8_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(uint16_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(uint32_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(uint64_t* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(float* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(double* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(bool* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(nil* v) NLIB_NOEXCEPT { NLIB_EXEC_READ; }
    NLIB_VIS_PUBLIC bool Read_(std::string* v);
    NLIB_VIS_PUBLIC bool Read_(MpObject** o) NLIB_NOEXCEPT;
    template <class T>
    bool Read_(T* v);
#undef NLIB_EXEC_READ

    NLIB_VIS_HIDDEN bool ReadMap(MpObject* obj, uint32_t n) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool ReadArray(MpObject* obj, uint32_t n) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool ReadRaw(MpObject* obj, uint32_t n) NLIB_NOEXCEPT;

 private:
    struct ValueObj {
        enum Category {
            CAT_NIL = 0,
            CAT_FALSE,
            CAT_TRUE,
            CAT_UINT,
            CAT_INT,
            CAT_FLOAT,
            CAT_UINT64,
            CAT_INT64,
            CAT_DOUBLE
        };
        Category category;
        union {
            uint32_t u;
            int32_t i;
            float f;
            uint64_t u64;
            int64_t i64;
            double d;
        } value;
        ValueObj() NLIB_NOEXCEPT : category(CAT_NIL) {}
    };
    size_t m_MaxArraySize;
    size_t m_MaxMapSize;
    size_t m_MaxRawSize;
    InputStream* m_Istr;
    errno_t m_ErrorValue;
    ValueObj m_Obj;
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpReader);
};

template <class T>
bool MpReader::Read(T* v, size_t n) {
    size_t num;
    if (!this->GetNextArrayNum(&num)) return false;
    if (n < num) {
        SetError(E2BIG);
        return false;
    }
    size_t i;
    for (i = 0; i < num; ++i) {
        if (!this->Read(&v[i])) return false;
    }
    // The case of n > num is not error.
    for (; i < n; ++i) {
        v[i] = T();
    }
    return true;
}

template <class T1, class T2>
inline bool MpRead(MpReader* r, std::pair<T1, T2>* v) {
    size_t num;
    if (!r->GetNextArrayNum(&num)) return false;
    if (num != 2) {
        r->SetError(E2BIG);  // E2BIG even if num is 1
        return false;
    }
    if (!r->Read(&v->first)) return false;
    if (!r->Read(&v->second)) return false;
    return true;
}

template <class T, class Alloc>
bool MpRead(MpReader* r, std::vector<T, Alloc>* v) {
    size_t num;
    if (!r->GetNextArrayNum(&num)) return false;
    v->resize(num);
    for (size_t i = 0; i < num; ++i) {
        if (!r->Read(&(*v)[i])) return false;
    }
    return true;
}

template <class T, class Alloc>
bool MpRead(MpReader* r, Nlist<T, Alloc>* v) {
    size_t num;
    if (!r->GetNextArrayNum(&num)) return false;
    v->clear();
    if (!v->reserve(num)) return false;
    for (size_t i = 0; i < num; ++i) {
        T* item = v->push_back();
        if (!r->Read(item)) return false;
    }
    return true;
}

template <class K, class V, class Pr, class Alloc>
bool MpRead(MpReader* r, std::map<K, V, Pr, Alloc>* m) {
    size_t num;
    if (!r->GetNextMapNum(&num)) return false;
    K key;
    V value;
    for (size_t i = 0; i < num; ++i) {
        if (!r->Read(&key)) return false;
        if (!r->Read(&value)) return false;
        (*m)[key] = value;
    }
    return true;
}

template <class T>
bool MpReader::Read_(T* v) {
    return ::nlib_ns::msgpack::MpRead(this, v);
}

#define NLIB_MPREADER_READ_(tp, coerce)              \
    \
template<> inline bool MpReader::Read_<tp>(tp * v) { \
        coerce vv;                                   \
        if (!Read_(&vv)) return false;               \
        *v = static_cast<tp>(vv);                    \
        return true;                                 \
    \
}

NLIB_MPREADER_READ_(char, int8_t)
NLIB_MPREADER_READ_(unsigned char, uint8_t)
NLIB_MPREADER_READ_(int, int32_t)
NLIB_MPREADER_READ_(unsigned int, uint32_t)
NLIB_MPREADER_READ_(short, int16_t)                          // NOLINT
NLIB_MPREADER_READ_(unsigned short, uint16_t)                // NOLINT
NLIB_MPREADER_READ_(long, nlib_long_compatible_t)            // NOLINT
NLIB_MPREADER_READ_(unsigned long, nlib_ulong_compatible_t)  // NOLINT
NLIB_MPREADER_READ_(long long, int64_t)                      // NOLINT
NLIB_MPREADER_READ_(unsigned long long, uint64_t)            // NOLINT

#undef NLIB_MPREADER_READ_

#define NLIB_MPREADER_CONVERT_(tp, coerce)                   \
    \
template<> inline errno_t MpReader::Convert_<tp>(tp * v) {   \
        return this->Convert_(reinterpret_cast<coerce*>(v)); \
    \
}

NLIB_MPREADER_CONVERT_(char, int8_t)
NLIB_MPREADER_CONVERT_(unsigned char, uint8_t)
NLIB_MPREADER_CONVERT_(int, int32_t)
NLIB_MPREADER_CONVERT_(unsigned int, uint32_t)
NLIB_MPREADER_CONVERT_(short, int16_t)                          // NOLINT
NLIB_MPREADER_CONVERT_(unsigned short, uint16_t)                // NOLINT
NLIB_MPREADER_CONVERT_(long, nlib_long_compatible_t)            // NOLINT
NLIB_MPREADER_CONVERT_(unsigned long, nlib_ulong_compatible_t)  // NOLINT
NLIB_MPREADER_CONVERT_(long long, int64_t)                      // NOLINT
NLIB_MPREADER_CONVERT_(unsigned long long, uint64_t)            // NOLINT

#undef NLIB_MPREADER_CONVERT_

}  // namespace msgpack
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_MPREADER_H_
