﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_MPWRITER_H_
#define INCLUDE_NN_NLIB_MSGPACK_MPWRITER_H_

#include <map>
#include <string>
#include <utility>
#include <vector>

#include "nn/nlib/msgpack/MpObject.h"
#include "nn/nlib/OutputStream.h"
#include "nn/nlib/BinaryWriter.h"
#include "nn/nlib/Cstring.h"
#include "nn/nlib/Nlist.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {

class MpWriter;

template <class T>
bool MpWrite(MpWriter* obj, const T& v);

class NLIB_VIS_PUBLIC MpWriter NLIB_FINAL {
 public:
    MpWriter() NLIB_NOEXCEPT : m_ErrorValue(0) {}
    ~MpWriter() NLIB_NOEXCEPT { this->Close(); }
    bool Init(OutputStream* stream) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Flush() NLIB_NOEXCEPT {
        if (!m_Writer.Flush()) {
            this->SetError(EIO);
            return false;
        }
        return this->GetErrorValue() == 0;
    }
    bool Close() NLIB_NOEXCEPT {
        // MpWriter does not own m_Ostr, and do not close m_Ostr.
        if (!m_Writer.Close()) {
            this->SetError(EIO);
        }
        return this->GetErrorValue() == 0;
    }
    void SetError(errno_t e) NLIB_NOEXCEPT {
        if (m_ErrorValue == 0) m_ErrorValue = e;
    }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return m_ErrorValue; }

 public:
    bool Write(const char* str) NLIB_NOEXCEPT NLIB_NONNULL {
        return this->Write(&str[0], nlib_strlen(str));
    }
    template <uint32_t n>
    bool Write(const char (&str)[n]) NLIB_NOEXCEPT {
        return this->Write(&str[0], nlib_strlen(str));
    }
    template <class T, size_t n>
    bool Write(const T (&vec)[n]);
    template <class T>
    bool Write(const T& tp);

    bool Write(const void* p, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool WriteArrayCount(uint32_t n) NLIB_NOEXCEPT;
    bool WriteMapCount(uint32_t n) NLIB_NOEXCEPT;
    bool WriteRawCount(uint32_t n) NLIB_NOEXCEPT;
    bool WriteRawBody(const void* p, uint32_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Write(const MpObject& obj) NLIB_NOEXCEPT;
    OutputStream* GetStream() NLIB_NOEXCEPT { return m_Writer.GetStream(); }
    NLIB_SAFE_BOOL(MpWriter, GetErrorValue() == 0);

 private:  // See: http://wiki.msgpack.org/display/MSGPACK/Format+specification
    bool WriteU8(uint8_t v) NLIB_NOEXCEPT;
    bool WriteU16(uint16_t v) NLIB_NOEXCEPT;
    bool WriteU32(uint32_t v) NLIB_NOEXCEPT;
    bool WriteU64(uint64_t v) NLIB_NOEXCEPT;
    bool WriteI8(int8_t v) NLIB_NOEXCEPT;
    bool WriteI16(int16_t v) NLIB_NOEXCEPT;
    bool WriteI32(int32_t v) NLIB_NOEXCEPT;
    bool WriteI64(int64_t v) NLIB_NOEXCEPT;
    bool Write_(uint8_t v) NLIB_NOEXCEPT { return this->WriteU8(v); }
    bool Write_(uint16_t v) NLIB_NOEXCEPT { return this->WriteU16(v); }
    bool Write_(uint32_t v) NLIB_NOEXCEPT { return this->WriteU32(v); }
    bool Write_(uint64_t v) NLIB_NOEXCEPT { return this->WriteU64(v); }
    bool Write_(int8_t v) NLIB_NOEXCEPT { return this->WriteI8(v); }
    bool Write_(int16_t v) NLIB_NOEXCEPT { return this->WriteI16(v); }
    bool Write_(int32_t v) NLIB_NOEXCEPT { return this->WriteI32(v); }
    bool Write_(int64_t v) NLIB_NOEXCEPT { return this->WriteI64(v); }
    bool Write_(float v) NLIB_NOEXCEPT;
    bool Write_(double v) NLIB_NOEXCEPT;
    bool Write_(bool v) NLIB_NOEXCEPT {
        return m_Writer.Write<uint8_t>(v ? 0xC3 : 0xC2);  // true : false
    }
    bool Write_(nil v) NLIB_NOEXCEPT {
        NLIB_UNUSED(v);
        return m_Writer.Write<uint8_t>(0xC0);  // nil
    }
    bool Write_(const std::string& str) NLIB_NOEXCEPT { return this->Write(&str[0], str.size()); }
    template <class T>
    bool Write_(const T& v);

 private:
    BinaryWriter m_Writer;
    errno_t m_ErrorValue;
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpWriter);
};

#define NLIB_MPWRITER_CHK_NSTREAM \
    if (!m_Ostr) {                \
        SetError(ENOENT);         \
        return false;             \
    }
#define NLIB_MPWRITER_SET_ERROR \
    SetError(EIO);              \
    return false

template <class T, size_t n>
bool MpWriter::Write(const T (&vec)[n]) {
    if (!this->WriteArrayCount(static_cast<uint32_t>(n))) return false;
    for (size_t i = 0; i < n; ++i) {
        if (!this->Write(vec[i])) {
            NLIB_MPWRITER_SET_ERROR;
        }
    }
    return true;
}

template <class T>
inline bool MpWriter::Write(const T& tp) {
    if (this->Write_(tp)) {
        return true;
    }
    NLIB_MPWRITER_SET_ERROR;
}

#undef NLIB_MPWRITER_SET_ERROR
#undef NLIB_MPWRITER_CHK_NSTREAM

#define NLIB_MPWRITE(tp, coerce)                             \
template<> inline bool MpWrite(MpWriter* obj, const tp& v) { \
        return obj->Write(static_cast<coerce>(v));           \
}

NLIB_MPWRITE(char, int32_t)
NLIB_MPWRITE(unsigned char, uint32_t)
NLIB_MPWRITE(int, int32_t)
NLIB_MPWRITE(unsigned int, uint32_t)
NLIB_MPWRITE(short, int32_t)                          // NOLINT
NLIB_MPWRITE(unsigned short, uint32_t)                // NOLINT
NLIB_MPWRITE(long long, int64_t)                      // NOLINT
NLIB_MPWRITE(unsigned long long, uint64_t)            // NOLINT
NLIB_MPWRITE(long, nlib_long_compatible_t)            // NOLINT
NLIB_MPWRITE(unsigned long, nlib_ulong_compatible_t)  // NOLINT

#undef NLIB_MPWRITE

template <class T1, class T2>
bool MpWrite(MpWriter* w, const std::pair<T1, T2>& v) {
    return w->WriteArrayCount(2) && w->Write(v.first) && w->Write(v.second);
}

inline bool MpWrite(MpWriter* w, const char* str) {
    uint32_t n = static_cast<uint32_t>(StrLen(str));
    return w->WriteRawCount(n) && w->WriteRawBody(str, n);
}

template <class T, class Alloc>
bool MpWrite(MpWriter* w, const std::vector<T, Alloc>& vec) {
    size_t n = vec.size();
    if (!w->WriteArrayCount(static_cast<uint32_t>(n))) return false;
    typename std::vector<T, Alloc>::const_iterator it = vec.begin();
    typename std::vector<T, Alloc>::const_iterator it_end = vec.end();
    for (; it != it_end; ++it) {
        if (!w->Write(*it)) return false;
    }
    return true;
}

template <class T, class Alloc>
bool MpWrite(MpWriter* w, const Nlist<T, Alloc>& vec) {
    size_t n = vec.size();
    if (!w->WriteArrayCount(static_cast<uint32_t>(n))) return false;
    typename Nlist<T, Alloc>::const_iterator it = vec.begin();
    typename Nlist<T, Alloc>::const_iterator it_end = vec.end();
    for (; it != it_end; ++it) {
        if (!w->Write(*it)) return false;
    }
    return true;
}

template <class K, class V, class Pr, class Alloc>
bool MpWrite(MpWriter* w, const std::map<K, V, Pr, Alloc>& m) {
    size_t n = m.size();
    if (!w->WriteMapCount(static_cast<uint32_t>(n))) return false;
    typename std::map<K, V, Pr, Alloc>::const_iterator it = m.begin();
    typename std::map<K, V, Pr, Alloc>::const_iterator end = m.end();
    for (; it != end; ++it) {
        if (!w->Write(it->first)) return false;
        if (!w->Write(it->second)) return false;
    }
    return true;
}

template <class T>
bool MpWriter::Write_(const T& v) {
    // This function must be at the end of this file.
    return ::nlib_ns::msgpack::MpWrite(this, v);
}

}  // namespace msgpack
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_MPWRITER_H_
