﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_AHOCORASICKBUILDER_H_
#define INCLUDE_NN_NLIB_SUCCINCT_AHOCORASICKBUILDER_H_

#include <string.h>
#include <utility>

#include "nn/nlib/succinct/AhoCorasick.h"
#include "nn/nlib/ReallocVec.h"
#include "nn/nlib/threading/ThreadPool.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class AhoCorasickBuilder;

namespace detail {

class StringHolder NLIB_FINAL {
 public:
    NLIB_VIS_PUBLIC StringHolder() NLIB_NOEXCEPT : m_Cur(kBufSize) {}
    NLIB_VIS_PUBLIC ~StringHolder() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN char* MakeString(const void* ptr, size_t len) NLIB_NOEXCEPT;
 private:
    static const size_t kBufSize = 1024 * 1024 * 4U;

    ReallocVec<char*> m_Data;
    ReallocVec<char*> m_Buf;
    size_t m_Cur;
    NLIB_DISALLOW_COPY_AND_ASSIGN(StringHolder);
};

class NLIB_VIS_PUBLIC PrefixString NLIB_FINAL {
 public:
    NLIB_VIS_HIDDEN PrefixString() NLIB_NOEXCEPT : m_Rbegin(NULL) {}
    NLIB_VIS_HIDDEN ~PrefixString() NLIB_NOEXCEPT {}
    NLIB_VIS_HIDDEN
        PrefixString(const unsigned char* first, const unsigned char* last) NLIB_NOEXCEPT {
        NLIB_UNUSED(first);
        NLIB_ASSERT(last);
        m_Rbegin = last - 1;
    }
    NLIB_VIS_HIDDEN bool ReverseLexOrderLess(const PrefixString& rhs) const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Rbegin && rhs.m_Rbegin);
        // NOTE:
        // lexicographical_compare between reversed strings
        // lexicographical_compare is very slow in DEBUG build of MSVC
        const unsigned char* rbeg = m_Rbegin;
        const unsigned char* rbeg_rhs = rhs.m_Rbegin;
        for (;;) {
            unsigned char c = *rbeg;
            unsigned char c_rhs = *rbeg_rhs;
            if (c < c_rhs) return true;
            if (c > c_rhs) return false;
            if (c == 0) return false;
            --rbeg;
            --rbeg_rhs;
        }
    }
    NLIB_VIS_HIDDEN const unsigned char* begin() const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Rbegin);
        return m_Rbegin - length() + 1;
    }
    NLIB_VIS_HIDDEN const unsigned char* end() const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Rbegin);
        return m_Rbegin + 1;
    }
    NLIB_VIS_HIDDEN bool empty() const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Rbegin);
        return *m_Rbegin == 0;
    }
    NLIB_VIS_HIDDEN size_t length() const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Rbegin);
        const unsigned char* p;
        const unsigned char* rbeg = m_Rbegin;
        for (p = rbeg; *p; --p) {
        }
        return rbeg - p;
    }
    NLIB_VIS_HIDDEN bool IsTerminal() const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Rbegin);
        return !*(m_Rbegin + 1);
    }

 private:
    const unsigned char* m_Rbegin;
};

class NLIB_VIS_PUBLIC AcNode NLIB_FINAL {
 public:
    struct EdgeList {
        unsigned char label;
        AcNode* dest;
        EdgeList* next;
    };

 public:
    NLIB_VIS_HIDDEN AcNode() NLIB_NOEXCEPT : m_NodeId(0), m_Failure(NULL), m_Edges(NULL) {}
    NLIB_VIS_HIDDEN ~AcNode() NLIB_NOEXCEPT {}
    NLIB_VIS_HIDDEN bool operator<(const AcNode& rhs) const NLIB_NOEXCEPT {
        return m_Prefix.ReverseLexOrderLess(rhs.m_Prefix);
    }
    NLIB_VIS_HIDDEN unsigned char GetLevelValue(size_t level) const NLIB_NOEXCEPT {
        return level < m_NodeId ? *(m_Prefix.end() - 1 - level) : 0;
    }
    NLIB_VIS_HIDDEN uint32_t GetNodeId() const NLIB_NOEXCEPT { return m_NodeId; }
    NLIB_VIS_HIDDEN bool IsTerminal() const NLIB_NOEXCEPT {
        return m_NodeId != 0 && m_Prefix.IsTerminal();
    }

    NLIB_VIS_HIDDEN bool AddGoto(unsigned char c, AcNode* next,
                                 Nlist<AcNode::EdgeList>* holder) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN void print() NLIB_NOEXCEPT;

    NLIB_VIS_HIDDEN const AcNode* Goto(const unsigned char c) const NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN AcNode* Find(const unsigned char* first,
                                 const unsigned char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN AcNode* GetReportArc() const NLIB_NOEXCEPT {
        AcNode* failure;
        for (failure = m_Failure;
             failure != NULL && !failure->IsTerminal();
             failure = failure->m_Failure) {
        }
        return failure;
    }
    NLIB_VIS_HIDDEN AcNode* GetFailureArc() const NLIB_NOEXCEPT { return m_Failure; }
    NLIB_VIS_HIDDEN const EdgeList* GetEdgeList() const NLIB_NOEXCEPT { return m_Edges; }
    NLIB_VIS_HIDDEN AcNode* FindArc(unsigned char c) const NLIB_NOEXCEPT {
        for (EdgeList* e = m_Edges; e != NULL; e = e->next) {
            if (e->label == c) return e->dest;
        }
        return NULL;
    }
    NLIB_VIS_HIDDEN PrefixString GetPrefix() const NLIB_NOEXCEPT { return m_Prefix; }

 private:
    NLIB_VIS_HIDDEN void SetNodeId(uint32_t nodeid) NLIB_NOEXCEPT { m_NodeId = nodeid; }
    NLIB_VIS_HIDDEN void SetFailureArc(AcNode* failure) NLIB_NOEXCEPT { m_Failure = failure; }
    NLIB_VIS_HIDDEN void SetPrefix(PrefixString prefix) NLIB_NOEXCEPT {
        m_Prefix = prefix;
        m_NodeId = static_cast<uint32_t>(m_Prefix.length());
    }

 private:
    uint32_t m_NodeId;  // suffix-lexicographic order, m_NodeId == 0 if root
    PrefixString m_Prefix;
    AcNode* m_Failure;
    EdgeList* m_Edges;

    friend class ::nlib_ns::succinct::AhoCorasickBuilder;
    NLIB_DISALLOW_COPY_AND_ASSIGN(AcNode);
};

}  // namespace detail

class AhoCorasickBuilder NLIB_FINAL {
    typedef ReallocVec<detail::AcNode*> AcNodeHolder;

 public:
    AhoCorasickBuilder() NLIB_NOEXCEPT : m_NumWords(0), m_NumBytes(0) {}
    NLIB_VIS_PUBLIC ~AhoCorasickBuilder() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Init() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool AddWord(const char* str) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool AddPattern(const void* p, size_t n) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool AddWords(const char* str, size_t len) NLIB_NOEXCEPT;
    bool AddWords(const char* str) NLIB_NOEXCEPT {
        return AddWords(str, strlen(str));
    }
    NLIB_VIS_PUBLIC AhoCorasick* Build() NLIB_NOEXCEPT;
    typedef bool (*MatchCallback)(const char* first, const char* last, uint32_t nodeid,
                                  void* user_obj);
    NLIB_VIS_PUBLIC void MatchByBuilder(const char* doc, MatchCallback callback,
                                        void* user_obj = NULL) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void print() NLIB_NOEXCEPT;
    size_t GetNumWords() const NLIB_NOEXCEPT { return m_NumWords; }
    size_t GetNumBytes() const NLIB_NOEXCEPT { return m_NumBytes; }
    size_t GetNumNodes() const NLIB_NOEXCEPT { return m_AcNodeHolder.size(); }

 private:
    NLIB_VIS_HIDDEN bool SortNodes() NLIB_NOEXCEPT;
    struct NLIB_VIS_HIDDEN BuildFailureArcTh {
        bool operator()() NLIB_NOEXCEPT;
        AhoCorasickBuilder* This;
    };
    struct NLIB_VIS_HIDDEN BuildReportTreeHolderTh {
        bool operator()() NLIB_NOEXCEPT;
        AhoCorasickBuilder* This;
        AhoCorasick* obj;
    };
    struct NLIB_VIS_HIDDEN BuildFailureTreeHolderTh {
        bool operator()() NLIB_NOEXCEPT;
        AhoCorasickBuilder* This;
        AhoCorasick* obj;
    };
    struct NLIB_VIS_HIDDEN BuildGotoArcHolderTh {
        bool operator()() NLIB_NOEXCEPT;
        AhoCorasickBuilder* This;
        AhoCorasick* obj;
    };
    struct NLIB_VIS_HIDDEN BuildLenHolderTh {
        bool operator()() NLIB_NOEXCEPT;
        AhoCorasickBuilder* This;
        AhoCorasick* obj;
    };
    struct NLIB_VIS_HIDDEN SortNodesTh {
        void operator()() NLIB_NOEXCEPT;
        AhoCorasickBuilder* This;
        int from;
        int to;
        const int* count;
    };

 private:
    size_t m_NumWords;
    size_t m_NumBytes;
    AcNodeHolder m_AcNodeHolder;
    Nlist<detail::AcNode> m_AcNodeMem;
    Nlist<detail::AcNode::EdgeList> m_EdgeMem;
    detail::StringHolder m_StringHolder;
    nlib_ns::threading::ThreadPool m_ThreadPool;

    NLIB_DISALLOW_COPY_AND_ASSIGN(AhoCorasickBuilder);
};

}  // namespace succinct
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_AHOCORASICKBUILDER_H_
