﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_BP_H_
#define INCLUDE_NN_NLIB_SUCCINCT_BP_H_

#include "nn/nlib/Swap.h"
#include "nn/nlib/succinct/Sbv.h"
#include "nn/nlib/ReallocVec.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

class Bp_;

class Bp NLIB_FINAL {
 public:
    Bp() NLIB_NOEXCEPT : m_Bp(NULL) {}
    ~Bp() NLIB_NOEXCEPT { this->Reset(); }
    NLIB_MOVE_MEMBER_HELPER_2(Bp, m_Bp, m_Select1);
    void swap(Bp& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_Bp, rhs.m_Bp);
        swap(m_Select1, rhs.m_Select1);
    }
    template <class TREEITER>
    bool Init(TREEITER& root, uint32_t num_nodes) NLIB_NOEXCEPT;  // NOLINT
    NLIB_VIS_PUBLIC bool InitDirect(const uint32_t* bv, uint32_t bv_size) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int ToNodeId(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int ToPos(uint32_t nodeid) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int FindOpen(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int FindClose(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int Enclose(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int FirstChild(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int LastChild(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int NextSibling(uint32_t pos) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int PrevSibling(uint32_t pos) const NLIB_NOEXCEPT;
    int Parent(uint32_t pos) const NLIB_NOEXCEPT { return this->Enclose(pos); }
    NLIB_VIS_PUBLIC size_t MemSize() const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    Set* Get1stLevelSet() NLIB_NOEXCEPT;
    bool Build() NLIB_NOEXCEPT;

 private:
    Bp_* m_Bp;
    detail::SbvSelect m_Select1;  // used to get pos from nodeid
    NLIB_DISALLOW_COPY_AND_ASSIGN(Bp);
};

template <class TREEITER>
bool Bp::Init(TREEITER& iter, uint32_t num_nodes) NLIB_NOEXCEPT {  // NOLINT
    // TREEITER has following methods:
    //   const void* NodePtr();        // ptr to node
    //   const void* ParentNodePtr();  // ptr to parent node
    //   bool MoveNext();  // move to the next node by depth first search
    if (num_nodes >= 0x80000000) return false;
    if (m_Bp) return false;
    Set* bv = Get1stLevelSet();  // bv = &m_Bp.m_B
    bv->Init(num_nodes * 2);

    uint32_t idx = 0;
    ReallocVec<const void*> stack;
    do {
        const void* parent = iter.ParentNodePtr();
        const void* node = iter.NodePtr();
        if (stack.empty() || stack.back() == parent) {
            bv->TurnOn(idx++);
            if (!stack.push_back(node)) return false;
        } else {
            while (stack.back() != parent) {
                stack.pop_back();
                ++idx;
            }
            bv->TurnOn(idx++);
            if (!stack.push_back(node)) return false;
        }
    } while (iter.MoveNext());

    if (!this->Build() || !m_Select1.Build(bv->GetBitVector(), bv->GetBitVectorSize())) {
        this->Reset();
        return false;
    }
    return true;
}

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Bp)

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_BP_H_
