﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_SBV_H_
#define INCLUDE_NN_NLIB_SUCCINCT_SBV_H_

#include <string.h>
#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/SmartBitmap.h"
#include "nn/nlib/BinaryReader.h"
#include "nn/nlib/BinaryWriter.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/TypeTraits.h"
#include "nn/nlib/ReallocVec.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN

namespace succinct {

class Set;

namespace detail {

// BIT must be uint32_t or uint64_t

static const unsigned int BLK_SIZE = 32;

template<class BIT>
NLIB_ALWAYS_INLINE uint32_t GetBlockCount(uint32_t num_bits) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(BIT) * 8;  // NOLINT
    return (num_bits + kBitCount - 1) / kBitCount;
}

template<class BIT>
NLIB_ALWAYS_INLINE uint64_t GetBlockCount(uint64_t num_bits) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(BIT) * 8;  // NOLINT
    return (num_bits + kBitCount - 1) / kBitCount;
}

template<class BIT>
NLIB_ALWAYS_INLINE void SetBit(BIT* bv, uint32_t idx) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(*bv) * 8;
    uint32_t blk = idx / kBitCount;
    uint32_t ofs = idx % kBitCount;
    bv[blk] |= static_cast<BIT>(1) << ofs;
}

template<class BIT>
NLIB_ALWAYS_INLINE void SetBit(BIT* bv, uint64_t idx) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(*bv) * 8;
    uint64_t blk = idx / kBitCount;
    uint32_t ofs = static_cast<uint32_t>(idx % kBitCount);
    bv[blk] |= static_cast<BIT>(1) << ofs;
}

template<class BIT>
NLIB_ALWAYS_INLINE void ResetBit(BIT* bv, uint32_t idx) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(*bv) * 8;
    uint32_t blk = idx / kBitCount;
    uint32_t ofs = idx % kBitCount;
    bv[blk] &= ~(static_cast<BIT>(1) << ofs);
}

template<class BIT>
NLIB_ALWAYS_INLINE void ResetBit(BIT* bv, uint64_t idx) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(*bv) * 8;
    uint64_t blk = idx / kBitCount;
    uint32_t ofs = static_cast<uint32_t>(idx % kBitCount);
    bv[blk] &= ~(static_cast<BIT>(1) << ofs);
}

template<class BIT>
NLIB_ALWAYS_INLINE bool GetBit(const BIT* bv, uint32_t idx) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(*bv) * 8;
    uint32_t blk = idx / kBitCount;
    uint32_t ofs = idx % kBitCount;
    return 0 != (bv[blk] & (static_cast<BIT>(1) << ofs));
}

template<class BIT>
NLIB_ALWAYS_INLINE bool GetBit(const BIT* bv, uint64_t idx) NLIB_NOEXCEPT {
    static const int kBitCount = sizeof(*bv) * 8;
    uint64_t blk = idx / kBitCount;
    uint32_t ofs = static_cast<uint32_t>(idx % kBitCount);
    return 0 != (bv[blk] & (static_cast<BIT>(1) << ofs));
}

template <class T>
class Resetter NLIB_FINAL {
 public:
    NLIB_ALWAYS_INLINE explicit Resetter(T* obj) NLIB_NOEXCEPT
        : m_Obj(obj), m_Ok(false) {}
    NLIB_ALWAYS_INLINE ~Resetter() NLIB_NOEXCEPT {
        if (!m_Ok) m_Obj->Reset();
    }
    NLIB_ALWAYS_INLINE bool Ok() NLIB_NOEXCEPT {
        m_Ok = true;
        return true;
    }

 private:
    T* m_Obj;
    bool m_Ok;
};

template <class T>
class PlainArray32 NLIB_FINAL {
 public:
    PlainArray32() NLIB_NOEXCEPT : m_Count(0), m_Value() {
        NLIB_STATIC_ASSERT(IsPod<T>::value);
    }
    ~PlainArray32() NLIB_NOEXCEPT { this->Reset(); }
    NLIB_MOVE_MEMBER_HELPER_2(PlainArray32, m_Count, m_Value);
    void swap(PlainArray32& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_Count, rhs.m_Count);
        swap(m_Value, rhs.m_Value);
    }

    bool Init(uint32_t count) NLIB_NOEXCEPT;
    T& operator[](size_t idx) {
        NLIB_ASSERT(idx < m_Count);
        return m_Value[idx];
    }
    const T& operator[](size_t idx) const {
        NLIB_ASSERT(idx < m_Count);
        return m_Value[idx];
    }
    size_t Size() const NLIB_NOEXCEPT { return m_Count; }
    size_t MemSize() const NLIB_NOEXCEPT {
        size_t array_size = sizeof(T) * m_Count;
        return sizeof(m_Count) + sizeof(m_Value) + array_size;
    }
    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    uint32_t m_Count;
    UniquePtr<T[]> m_Value;
    NLIB_DISALLOW_COPY_AND_ASSIGN(PlainArray32);
};

template <class T>
bool PlainArray32<T>::Init(uint32_t count) NLIB_NOEXCEPT {
    if (m_Value) return false;
    if (count == 0) {
        m_Count = 0;
        return true;
    }
    m_Value.reset(new (std::nothrow) T[count]);
    if (!m_Value) return false;
    m_Count = count;
    memset(m_Value.get(), 0, sizeof(T) * count);
    return true;
}

template <class T>
NLIB_ALWAYS_INLINE void PlainArray32<T>::Reset() NLIB_NOEXCEPT {
    m_Count = 0;
    m_Value.reset();
}

template <class T>
bool PlainArray32<T>::Export(BinaryWriter* w) const NLIB_NOEXCEPT {
    if (!w->Write(m_Count)) return false;
    if (!w->WriteArray(m_Value.get(), m_Count)) return false;
    return true;
}

template <class T>
bool PlainArray32<T>::Import(BinaryReader* r) NLIB_NOEXCEPT {
    Resetter<PlainArray32<T> > rst(this);
    if (!r->Read(&m_Count)) return false;
    if (m_Count == 0) {
        m_Value.reset();
        return true;
    }
    m_Value.reset(new (std::nothrow) T[m_Count]);
    if (!m_Value) return false;
    if (m_Count != r->ReadArray(m_Value.get(), m_Count)) return false;
    rst.Ok();
    return true;
}

class BitVector32 NLIB_FINAL {
 public:
    BitVector32() NLIB_NOEXCEPT : m_Size(0), m_BitVector() {}
    NLIB_VIS_PUBLIC ~BitVector32() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_2(BitVector32, m_Size, m_BitVector);
    void swap(BitVector32& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_Size, rhs.m_Size);
        swap(m_BitVector, rhs.m_BitVector);
    }
    NLIB_VIS_PUBLIC bool Init(uint32_t size) NLIB_NOEXCEPT;
    bool TurnOn(uint32_t nth) NLIB_NOEXCEPT {
        if (nth >= m_Size) return false;
        SetBit(m_BitVector.get(), nth);
        return true;
    }
    bool TurnOff(uint32_t nth) NLIB_NOEXCEPT {
        if (nth >= m_Size) return false;
        ResetBit(m_BitVector.get(), nth);
        return true;
    }
    bool operator[](uint32_t nth) const NLIB_NOEXCEPT {
        return nth < m_Size ? GetBit(m_BitVector.get(), nth) : false;
    }
    uint32_t Size() const NLIB_NOEXCEPT { return m_Size; }
    const uint32_t* Data() const NLIB_NOEXCEPT {
        return m_BitVector.get();
    }
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC size_t MemSize() const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    uint32_t m_Size;  // m_Size is bit length
    UniquePtr<uint32_t[]> m_BitVector;
    NLIB_DISALLOW_COPY_AND_ASSIGN(BitVector32);
};

// classic rank
// See http://codezine.jp/article/detail/260
class SbvRank NLIB_FINAL {
    // LV-A has cumulative popCounts for each 256bits block.
    // LV-B has popCounts in a 256bits block.
    // We can calcuate 'rank' in O(1).
 public:
    SbvRank() NLIB_NOEXCEPT : m_LvA(), m_LvB() {}
    NLIB_VIS_PUBLIC ~SbvRank() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_2(SbvRank, m_LvA, m_LvB);
    void swap(SbvRank& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_LvA, rhs.m_LvA);
        swap(m_LvB, rhs.m_LvB);
    }
    NLIB_VIS_PUBLIC bool Build(const uint32_t* bv, uint32_t size) NLIB_NOEXCEPT;
    uint32_t Rank1(const uint32_t* bv, uint32_t pos) const NLIB_NOEXCEPT {
        using ::nlib_ns::detail::CalcRank1;
        return CalcRank1(pos, bv, m_LvA.get(), m_LvB.get());
    }
    uint32_t Rank0(const uint32_t* bv, uint32_t pos) const NLIB_NOEXCEPT {
        return pos + 1 - this->Rank1(bv, pos);
    }
    int32_t Select1(const uint32_t* bv, uint32_t size, uint32_t nth) const NLIB_NOEXCEPT {
        using ::nlib_ns::detail::CalcSelect1;
        if (nth >= this->Rank1(bv, size - 1)) return -1;
        int32_t ans = CalcSelect1(nth, size, bv, m_LvA.get(), m_LvB.get());
        NLIB_ASSERT(ans < static_cast<int>(size));
        return ans;
    }
    int32_t Select0(const uint32_t* bv, uint32_t size, uint32_t nth) const NLIB_NOEXCEPT {
        using ::nlib_ns::detail::CalcSelect0;
        if (nth >= this->Rank0(bv, size - 1)) return -1;
        int32_t ans = CalcSelect0(nth, size, bv, m_LvA.get(), m_LvB.get());
        NLIB_ASSERT(ans < static_cast<int>(size));
        return ans;
    }
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC size_t MemSize(uint32_t size) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w, uint32_t size) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r, uint32_t size) NLIB_NOEXCEPT;

 private:
    UniquePtr<uint32_t[]> m_LvA;
    UniquePtr<uint8_t[]> m_LvB;
    NLIB_DISALLOW_COPY_AND_ASSIGN(SbvRank);
};

// classic select:
// See http://codezine.jp/article/detail/260
class SbvSelect NLIB_FINAL {
 public:
    // We can calcuate 'select' in O(1).
    SbvSelect() NLIB_NOEXCEPT {}
    NLIB_VIS_PUBLIC ~SbvSelect() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_5(SbvSelect, m_ClumpArray, m_R, m_Q, m_RankR, m_RankQ);
    void swap(SbvSelect& rhs) NLIB_NOEXCEPT {
        using std::swap;
        m_ClumpArray.swap(rhs.m_ClumpArray);
        m_R.swap(rhs.m_R);
        m_Q.swap(rhs.m_Q);
        m_RankR.swap(rhs.m_RankR);
        m_RankQ.swap(rhs.m_RankQ);
    }
    bool Build(const uint32_t* bv, uint32_t size) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int32_t Select(const Set& set, uint32_t nth) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC size_t MemSize() const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    PlainArray32<uint32_t> m_ClumpArray;
    BitVector32 m_R;
    BitVector32 m_Q;
    SbvRank m_RankR;
    SbvRank m_RankQ;
    NLIB_DISALLOW_COPY_AND_ASSIGN(SbvSelect);
};

}  // namespace detail

class Set NLIB_FINAL {
 public:
    typedef uint32_t IdxType;
    Set() NLIB_NOEXCEPT : m_BVec(), m_Rank() {}
    NLIB_VIS_PUBLIC ~Set() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_2(Set, m_BVec, m_Rank);
    void swap(Set& rhs) NLIB_NOEXCEPT {
        using std::swap;
        m_BVec.swap(rhs.m_BVec);
        m_Rank.swap(rhs.m_Rank);
    }
    bool Init(uint32_t bv_size) NLIB_NOEXCEPT { return m_BVec.Init(bv_size); }
    bool TurnOn(uint32_t idx) NLIB_NOEXCEPT { return m_BVec.TurnOn(idx); }
    bool TurnOff(uint32_t idx) NLIB_NOEXCEPT { return m_BVec.TurnOff(idx); }
    NLIB_VIS_PUBLIC bool Build() NLIB_NOEXCEPT;
    bool Has(uint32_t idx) const NLIB_NOEXCEPT { return m_BVec[idx]; }
    uint32_t Rank1(uint32_t idx) const NLIB_NOEXCEPT {
        // The values beyond the bit-vector are all 0.
        uint32_t pos = idx < m_BVec.Size() ? idx : m_BVec.Size() - 1;
        return m_Rank.Rank1(m_BVec.Data(), pos);
    }
    uint32_t Rank0(uint32_t idx) const NLIB_NOEXCEPT {
        uint32_t rank1 = this->Rank1(idx);
        return idx + 1 - rank1;
    }
    int32_t Select1(uint32_t nth) const NLIB_NOEXCEPT {
        return m_Rank.Select1(m_BVec.Data(), m_BVec.Size(), nth);
    }
    int32_t Select0(uint32_t nth) const NLIB_NOEXCEPT {
        return m_Rank.Select0(m_BVec.Data(), m_BVec.Size(), nth);
    }
    NLIB_VIS_PUBLIC size_t MemSize() const NLIB_NOEXCEPT;
    uint32_t GetBitVectorSize() const NLIB_NOEXCEPT { return m_BVec.Size(); }
    const uint32_t* GetBitVector() const NLIB_NOEXCEPT { return m_BVec.Data(); }
    void Reset() NLIB_NOEXCEPT { m_BVec.Reset(); }
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    detail::BitVector32 m_BVec;
    detail::SbvRank m_Rank;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Set);
};

class Sbv NLIB_FINAL {
 public:
    typedef uint32_t IdxType;
    Sbv() NLIB_NOEXCEPT {}
    NLIB_VIS_PUBLIC ~Sbv() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_2(Sbv, m_Set, m_Select1);
    void swap(Sbv& rhs) NLIB_NOEXCEPT {
        using std::swap;
        m_Set.swap(rhs.m_Set);
        m_Select1.swap(rhs.m_Select1);
    }
    bool Init(uint32_t bv_size) NLIB_NOEXCEPT { return m_Set.Init(bv_size); }
    bool TurnOn(uint32_t idx) NLIB_NOEXCEPT { return m_Set.TurnOn(idx); }
    bool TurnOff(uint32_t idx) NLIB_NOEXCEPT { return m_Set.TurnOff(idx); }
    NLIB_VIS_PUBLIC bool Build() NLIB_NOEXCEPT;
    bool Has(uint32_t idx) const NLIB_NOEXCEPT { return m_Set.Has(idx); }
    uint32_t Rank1(uint32_t idx) const NLIB_NOEXCEPT { return m_Set.Rank1(idx); }
    uint32_t Rank0(uint32_t idx) const NLIB_NOEXCEPT { return m_Set.Rank0(idx); }
    int32_t Select1(uint32_t nth) const NLIB_NOEXCEPT {
        return m_Select1.Select(m_Set, nth);
    }
    int32_t Select0(uint32_t nth) const NLIB_NOEXCEPT {
        return m_Set.Select0(nth);
    }
    size_t MemSize() const NLIB_NOEXCEPT;
    uint32_t GetBitVectorSize() const NLIB_NOEXCEPT {
        return m_Set.GetBitVectorSize();
    }
    const uint32_t* GetBitVector() const NLIB_NOEXCEPT {
        return m_Set.GetBitVector();
    }
    void Reset() NLIB_NOEXCEPT {
        m_Set.Reset();
        m_Select1.Reset();
    }
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    Set m_Set;
    detail::SbvSelect m_Select1;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Sbv);
};

class SparseSet NLIB_FINAL {
 public:
    // index is 64-bit and # of on-bits is 32-bit.
    // In other words, rank is 64-bit index, and select is 32-bit index
    typedef uint64_t IdxType;
    SparseSet() NLIB_NOEXCEPT : m_BitVectorSize(0),
                                m_LowestIdx(0xFFFFFFFF),
                                m_HighestIdx(0),
                                m_NumOnBits(0),
                                m_Width(0),
                                m_BLen(0),
                                m_B() {}
    NLIB_VIS_PUBLIC ~SparseSet() NLIB_NOEXCEPT;
#ifdef NLIB_CXX11_RVALUE_REFERENCES
    SparseSet(SparseSet&& rhs) NLIB_NOEXCEPT :
        m_BitVectorSize(rhs.m_BitVectorSize), m_LowestIdx(rhs.m_LowestIdx),
        m_HighestIdx(rhs.m_HighestIdx), m_NumOnBits(rhs.m_NumOnBits),
        m_Width(rhs.m_Width), m_BLen(rhs.m_BLen),
        m_B(std::move(rhs.m_B)), m_Sbv(std::move(rhs.m_Sbv)) {}
    SparseSet& operator=(SparseSet&& rhs) NLIB_NOEXCEPT {
        m_BitVectorSize = rhs.m_BitVectorSize;
        m_LowestIdx = rhs.m_LowestIdx;
        m_HighestIdx = rhs.m_HighestIdx;
        m_NumOnBits = rhs.m_NumOnBits;
        m_Width = rhs.m_Width;
        m_BLen = rhs.m_BLen;
        m_B = std::move(rhs.m_B);
        m_Sbv = std::move(rhs.m_Sbv);
        return *this;
    }
#endif
    NLIB_VIS_PUBLIC bool Init(uint64_t bv_size) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool TurnOn(uint64_t idx) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Build() NLIB_NOEXCEPT;
    bool Has(uint64_t idx, uint32_t* rank) const NLIB_NOEXCEPT {
        // called after Build()
        uint32_t rank1 = this->Rank1(idx);
        if (rank) *rank = rank1;
        if (rank1 == 0) return false;
        int64_t select1 = this->Select1Ex(rank1 - 1);
        NLIB_ASSERT(select1 >= 0);
        // if (select1 < 0) return false;  // not possible
        return idx == static_cast<uint64_t>(select1);
    }
    bool Has(uint64_t idx) const NLIB_NOEXCEPT {
        return this->Has(idx, NULL);
    }
    NLIB_VIS_PUBLIC uint32_t Rank1(uint64_t idx) const NLIB_NOEXCEPT;
    uint32_t Rank0(uint64_t idx) const NLIB_NOEXCEPT {
        uint32_t rank1 = this->Rank1(idx);
        return static_cast<uint32_t>(idx + 1 - rank1);
    }
    NLIB_VIS_PUBLIC int32_t Select1(uint32_t nth) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC int64_t Select1Ex(uint32_t nth) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC size_t MemSize() const NLIB_NOEXCEPT;
    uint64_t GetBitVectorSize() const NLIB_NOEXCEPT {
        return m_BitVectorSize;
    }
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void swap(SparseSet& rhs) NLIB_NOEXCEPT;  // NOLINT

 private:
    uint32_t GetBits(uint64_t pos, size_t num) const NLIB_NOEXCEPT {
        NLIB_ASSERT(num <= 32);
        uint64_t idx64 = pos / detail::BLK_SIZE;
        NLIB_ASSERT(idx64 <= SIZE_MAX);
        size_t idx = static_cast<size_t>(idx64);
        size_t ofs = static_cast<size_t>(pos % detail::BLK_SIZE);
        if (ofs + num > 32) {
            uint64_t b = m_B[idx + 1];
            b <<= 32;
            b += m_B[idx];
            return static_cast<uint32_t>((b >> ofs) & ((1 << num) - 1));
        } else {
            return (m_B[idx] >> ofs) & ((1 << num) - 1);
        }
    }

 private:
    uint64_t m_BitVectorSize;
    uint64_t m_LowestIdx;
    uint64_t m_HighestIdx;
    uint32_t m_NumOnBits;
    uint32_t m_Width;
    uint32_t m_BLen;
    UniquePtr<uint32_t[]> m_B;
    Sbv m_Sbv;  // Set m_Sbv;
    NLIB_DISALLOW_COPY_AND_ASSIGN(SparseSet);
};

/*
// GammaArray may be removed.
class GammaArray NLIB_FINAL {
public:
    GammaArray() : m_Len(0), m_U(NULL), m_B(NULL) {}
    ~GammaArray() { this->Reset(); }
    bool Init(const uint32_t* values, uint32_t n);
    uint32_t operator[](uint32_t idx) const;
    size_t MemSize() const;
    void Reset();
    bool Export(BinaryWriter* w) const;
    bool Import(BinaryReader* r);
    void swap(GammaArray& rhs) {  // NOLINT
        using std::swap;
        swap(m_Len, rhs.m_Len);
        swap(m_U, rhs.m_U);
        swap(m_B, rhs.m_B);
    }

private:
    typedef SparseSet SetType;
    uint32_t m_Len;
    SetType* m_U;
    detail::BitVector32* m_B;
    NLIB_DISALLOW_COPY_AND_ASSIGN(GammaArray);
};
*/

class CompressedArray NLIB_FINAL {
    static const size_t UNIT_SIZE = 64;

 public:
    // cppcheck-suppress uninitMemberVar
    CompressedArray() NLIB_NOEXCEPT : m_Header(), m_Data(), m_Idx(0) {}
    NLIB_VIS_PUBLIC ~CompressedArray() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_3(CompressedArray, m_Header, m_Data, m_Idx);
    void swap(CompressedArray& rhs) NLIB_NOEXCEPT {
        using std::swap;
        swap(m_Header, rhs.m_Header);
        swap(m_Data, rhs.m_Data);
        swap(m_Idx, rhs.m_Idx);
        for (size_t i = 0; i < UNIT_SIZE; ++i) {
            swap(m_Work[i], rhs.m_Work[i]);
        }
    }
    bool PushBack(uint32_t x) NLIB_NOEXCEPT {
        m_Work[m_Idx++] = x;
        if (m_Idx == UNIT_SIZE) return this->Build();
        return true;
    }
    NLIB_VIS_PUBLIC uint32_t operator[](size_t idx) const NLIB_NOEXCEPT;
    size_t Size() const NLIB_NOEXCEPT {
        return m_Idx + m_Header.size() * UNIT_SIZE / 2;
    }
    NLIB_VIS_PUBLIC void Reset() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC size_t MemSize() const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    NLIB_VIS_PUBLIC bool Build() NLIB_NOEXCEPT;
    // Header:
    //  baseidx: index of m_Data(lower 27bits)
    //  k: bit width of diff data(upper 5bits)
    //  base: base value(uint32_t)

    // Data:
    //  UNIT_SIZE * kbit: diff data
    ReallocVec<uint32_t> m_Header;
    ReallocVec<uint32_t> m_Data;
    uint8_t m_Idx;
    uint32_t m_Work[UNIT_SIZE];
    NLIB_DISALLOW_COPY_AND_ASSIGN(CompressedArray);
};

class Map NLIB_FINAL {
    typedef CompressedArray ArrayType;
    typedef SparseSet SetType;

 public:
    typedef SetType::IdxType IdxType;
    typedef std::pair<bool, uint32_t> FindType;
    Map() NLIB_NOEXCEPT : m_Set(), m_Value() {}
    ~Map() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_2(Map, m_Set, m_Value);
    void swap(Map& rhs) NLIB_NOEXCEPT {
        m_Set.swap(rhs.m_Set);
        m_Value.swap(rhs.m_Value);
    }
    bool Init(IdxType bv_size) NLIB_NOEXCEPT {
        if (m_Value.Size() != 0) return false;
        return m_Set.Init(bv_size);
    }
    bool TurnOn(IdxType idx, uint32_t data) NLIB_NOEXCEPT {
        if (!m_Set.TurnOn(idx)) return false;
        m_Value.PushBack(data);
        return true;
    }
    bool Build() NLIB_NOEXCEPT { return m_Set.Build(); }
    FindType Find(IdxType idx) NLIB_NOEXCEPT {
        if (m_Set.Has(idx))
            return std::make_pair(true, m_Value[m_Set.Rank1(idx) - 1]);
        else
            return std::make_pair(false, 0);
    }
    const FindType Find(IdxType idx) const NLIB_NOEXCEPT {
        if (m_Set.Has(idx))
            return std::make_pair(true, m_Value[m_Set.Rank1(idx) - 1]);
        else
            return std::make_pair(false, 0);
    }
    size_t MemSize() const NLIB_NOEXCEPT {
        size_t value_size = m_Value.MemSize();
        size_t set_size = m_Set.MemSize();
        return value_size + set_size;
    }
    const SetType& GetKeys() const NLIB_NOEXCEPT { return m_Set; }
    const ArrayType& GetValues() const NLIB_NOEXCEPT { return m_Value; }
    void Reset() NLIB_NOEXCEPT {
        m_Value.Reset();
        m_Set.Reset();
    }
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT {
        if (!m_Value.Export(w)) return false;
        if (!m_Set.Export(w)) return false;
        return true;
    }
    bool Import(BinaryReader* r) NLIB_NOEXCEPT {
        if (!m_Value.Import(r) || !m_Set.Import(r)) {
            this->Reset();
            return false;
        }
        return true;
    }

 private:
    SetType m_Set;
    ArrayType m_Value;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Map);
};

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::detail::BitVector32)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::detail::SbvRank)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::detail::SbvSelect)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Set)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Sbv)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::SparseSet)
// NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::GammaArray)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::CompressedArray)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Map)

NLIB_DEFINE_STD_SWAP_T_BEGIN4(nn, nlib, succinct, detail)               // NOLINT
NLIB_DEFINE_STD_SWAP_T1(T, ::nlib_ns::succinct::detail::PlainArray32)  // NOLINT
NLIB_DEFINE_STD_SWAP_T_END4(nn, nlib, succinct, detail)                 // NOLINT

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_SBV_H_
