﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TESTING_PARAMVALUETYPES_H_
#define INCLUDE_NN_NLIB_TESTING_PARAMVALUETYPES_H_

#include <memory>
#include <vector>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"

NLIB_NAMESPACE_BEGIN
namespace testing {
namespace detail {

class ValuesBool NLIB_FINAL {
 public:
    template <class T>
    std::vector<T>* Gen() const {
        bool a[] = {false, true};
        return new std::vector<T>(&a[0], &a[0] + 2);
    }

 private:
    ValuesBool& operator=(const ValuesBool&);
};

template <class C, size_t N>
class ValuesInObjArray NLIB_FINAL {
 public:
    explicit ValuesInObjArray(const C (&a)[N]) : m_Array(a) {}

    template <class T>
    std::vector<T>* Gen() const {
        return new std::vector<T>(&m_Array[0], &m_Array[0] + N);
    }

 private:
    const C (&m_Array)[N];
    ValuesInObjArray& operator=(const ValuesInObjArray&);
};

template <class C>
class ValuesInObjContainer NLIB_FINAL {
 public:
    explicit ValuesInObjContainer(const C& c) : m_Container(&c) {}

    template <class T>
    std::vector<T>* Gen() const {
        return new std::vector<T>(m_Container->begin(), m_Container->end());
    }

 private:
    const C* m_Container;
    ValuesInObjContainer& operator=(const ValuesInObjContainer&);
};

template <class Iterator>
class ValuesInObjIterator NLIB_FINAL {
 public:
    ValuesInObjIterator(Iterator first, Iterator last) : m_First(first), m_Last(last) {}

    template <class T>
    std::vector<T>* Gen() const {
        return new std::vector<T>(m_First, m_Last);
    }

 private:
    Iterator m_First;
    Iterator m_Last;
    ValuesInObjIterator& operator=(const ValuesInObjIterator&);
};

template <class T1, class T2>
class ValuesRange NLIB_FINAL {
 public:
    ValuesRange(T1 first, T1 last, T2 step) : m_First(first), m_Last(last), m_Step(step) {}

    template <class T>
    std::vector<T>* Gen() const;

 private:
    T1 m_First;
    T1 m_Last;
    T2 m_Step;
    ValuesRange& operator=(const ValuesRange&);
};

template <class T1, class T2>
template <class T>
std::vector<T>* ValuesRange<T1, T2>::Gen() const {
    // T must have:
    // - default constructor
    // - copy constructor
    // - operator+, operator<
    T first(m_First);
    T last(m_Last);
    T2 step(m_Step);
    T tmp(m_First);

    // counting
    size_t N = 0;
    while (tmp < last) {
        // first += step invokes a warning if T is char.
        tmp = tmp + step;
        ++N;
    }
    UniquePtr<std::vector<T> > vec(new std::vector<T>(N));
    if (vec.get()) {
        size_t n = 0;
        while (first < last) {
            (*vec)[n] = first;
            first = first + step;
            ++n;
        }
    }
    return vec.release();
}

template <class T1>
class Values1 NLIB_FINAL {
    typedef Values1 SelfType;

 public:
    explicit Values1(T1 v1) : m_V1(v1) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1};
        return new std::vector<T>(&a[0], &a[0] + 1);
    }

 private:
    const T1 m_V1;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2>
class Values2 NLIB_FINAL {
    typedef Values2 SelfType;

 public:
    Values2(T1 v1, T2 v2) : m_V1(v1), m_V2(v2) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2};
        return new std::vector<T>(&a[0], &a[0] + 2);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3>
class Values3 NLIB_FINAL {
    typedef Values3 SelfType;

 public:
    Values3(T1 v1, T2 v2, T3 v3) : m_V1(v1), m_V2(v2), m_V3(v3) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3};
        return new std::vector<T>(&a[0], &a[0] + 3);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4>
class Values4 NLIB_FINAL {
    typedef Values4 SelfType;

 public:
    Values4(T1 v1, T2 v2, T3 v3, T4 v4) : m_V1(v1), m_V2(v2), m_V3(v3), m_V4(v4) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4};
        return new std::vector<T>(&a[0], &a[0] + 4);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5>
class Values5 NLIB_FINAL {
    typedef Values5 SelfType;

 public:
    Values5(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5) : m_V1(v1), m_V2(v2), m_V3(v3), m_V4(v4), m_V5(v5) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4, m_V5};
        return new std::vector<T>(&a[0], &a[0] + 5);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    const T5 m_V5;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6>
class Values6 NLIB_FINAL {
    typedef Values6 SelfType;

 public:
    Values6(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6)
        : m_V1(v1), m_V2(v2), m_V3(v3), m_V4(v4), m_V5(v5), m_V6(v6) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4, m_V5, m_V6};
        return new std::vector<T>(&a[0], &a[0] + 6);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    const T5 m_V5;
    const T6 m_V6;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7>
class Values7 NLIB_FINAL {
    typedef Values7 SelfType;

 public:
    Values7(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7)
        : m_V1(v1), m_V2(v2), m_V3(v3), m_V4(v4), m_V5(v5), m_V6(v6), m_V7(v7) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4, m_V5, m_V6, m_V7};
        return new std::vector<T>(&a[0], &a[0] + 7);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    const T5 m_V5;
    const T6 m_V6;
    const T7 m_V7;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8>
class Values8 NLIB_FINAL {
    typedef Values8 SelfType;

 public:
    Values8(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8)
        : m_V1(v1), m_V2(v2), m_V3(v3), m_V4(v4), m_V5(v5), m_V6(v6), m_V7(v7), m_V8(v8) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4, m_V5, m_V6, m_V7, m_V8};
        return new std::vector<T>(&a[0], &a[0] + 8);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    const T5 m_V5;
    const T6 m_V6;
    const T7 m_V7;
    const T8 m_V8;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9>
class Values9 NLIB_FINAL {
    typedef Values9 SelfType;

 public:
    Values9(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8, T9 v9)
        : m_V1(v1),
          m_V2(v2),
          m_V3(v3),
          m_V4(v4),
          m_V5(v5),
          m_V6(v6),
          m_V7(v7),
          m_V8(v8),
          m_V9(v9) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4, m_V5, m_V6, m_V7, m_V8, m_V9};
        return new std::vector<T>(&a[0], &a[0] + 9);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    const T5 m_V5;
    const T6 m_V6;
    const T7 m_V7;
    const T8 m_V8;
    const T9 m_V9;
    SelfType& operator=(const SelfType&);
};

template <class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9,
          class T10>  // NOLINT
class Values10 NLIB_FINAL {
    typedef Values10 SelfType;

 public:
    Values10(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8, T9 v9, T10 v10)
        : m_V1(v1),
          m_V2(v2),
          m_V3(v3),
          m_V4(v4),
          m_V5(v5),
          m_V6(v6),
          m_V7(v7),
          m_V8(v8),
          m_V9(v9),
          m_V10(v10) {}

    template <class T>
    std::vector<T>* Gen() const {
        T a[] = {m_V1, m_V2, m_V3, m_V4, m_V5, m_V6, m_V7, m_V8, m_V9, m_V10};
        return new std::vector<T>(&a[0], &a[0] + 10);
    }

 private:
    const T1 m_V1;
    const T2 m_V2;
    const T3 m_V3;
    const T4 m_V4;
    const T5 m_V5;
    const T6 m_V6;
    const T7 m_V7;
    const T8 m_V8;
    const T9 m_V9;
    const T10 m_V10;
    SelfType& operator=(const SelfType&);
};

}  // namespace detail
}  // namespace testing
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_TESTING_PARAMVALUETYPES_H_
