﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_CRITICALSECTION_H_
#define INCLUDE_NN_NLIB_THREADING_CRITICALSECTION_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"

NLIB_NAMESPACE_BEGIN
class TimeSpan;
class DateTime;
namespace threading {

class NLIB_LOCKABLE SimpleCriticalSection {
 public:
#ifdef NLIB_CXX11_CONSTEXPR
    constexpr SimpleCriticalSection() NLIB_NOEXCEPT
        : m_Mutex(NLIB_MUTEX_INITIALIZER) {}
#else
    SimpleCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_init(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~SimpleCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_destroy(&m_Mutex);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    void lock() NLIB_NOEXCEPT NLIB_LOCK_FUNC() {
        errno_t e = nlib_mutex_lock(&m_Mutex);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    void unlock() NLIB_NOEXCEPT NLIB_UNLOCK_FUNC() {
        errno_t e = nlib_mutex_unlock(&m_Mutex);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    bool try_lock() NLIB_NOEXCEPT NLIB_TRYLOCK_FUNC(true) {
        errno_t e = nlib_mutex_trylock(&m_Mutex);
        if (e == 0) return true;
        NLIB_ASSERT(e == EBUSY);
        return false;
    }
    typedef nlib_mutex* native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT { return &m_Mutex; }

 private:
    nlib_mutex m_Mutex;
};

#ifndef NLIB_CRITICALSECTION_USE_FALLBACK
class CriticalSection {
 public:
#if defined(NLIB_CXX11_CONSTEXPR)
    constexpr CriticalSection() NLIB_NOEXCEPT
        : m_Mutex(NLIB_RECURSIVE_MUTEX_INITIALIZER) {}
#else
    CriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_recursive_init(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~CriticalSection() NLIB_NOEXCEPT { nlib_mutex_destroy(&m_Mutex); }
    void lock() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_lock(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    void unlock() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_unlock(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    bool try_lock() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_trylock(&m_Mutex);
        if (e == 0) return true;
        NLIB_ASSERT(e == EBUSY);
        return false;
    }
    typedef nlib_mutex* native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT { return &m_Mutex; }

 private:
    nlib_mutex m_Mutex;
};
#endif

#ifndef NLIB_TIMEDCRITICALSECTION_USE_FALLBACK
class TimedCriticalSection {
 public:
#if defined(NLIB_CXX11_CONSTEXPR)
    constexpr TimedCriticalSection() NLIB_NOEXCEPT
        : m_Mutex(NLIB_RECURSIVE_TIMED_MUTEX_INITIALIZER) {}
#else
    TimedCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_recursive_timed_init(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~TimedCriticalSection() NLIB_NOEXCEPT { nlib_mutex_destroy(&m_Mutex); }
    NLIB_VIS_PUBLIC bool try_lock_for(const TimeSpan& span) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool try_lock_until(const DateTime& abstime) NLIB_NOEXCEPT;
    void lock() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_lock(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    void unlock() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_unlock(&m_Mutex);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    bool try_lock() NLIB_NOEXCEPT {
        errno_t e = nlib_mutex_trylock(&m_Mutex);
        if (e == 0) return true;
        NLIB_ASSERT(e == EBUSY);
        return false;
    }
    typedef nlib_mutex* native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT { return &m_Mutex; }

 private:
    nlib_mutex m_Mutex;
};
#endif

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_NAMESPACE_BEGIN
namespace threading {

class DummyCriticalSection {
 public:
    NLIB_CEXPR DummyCriticalSection() NLIB_NOEXCEPT {}
    ~DummyCriticalSection() NLIB_NOEXCEPT {}
    void lock() NLIB_NOEXCEPT {}
    void unlock() NLIB_NOEXCEPT {}
    bool try_lock() NLIB_NOEXCEPT { return true; }
    bool try_lock_for(const TimeSpan&) NLIB_NOEXCEPT { return true; }    // NOLINT
    bool try_lock_until(const DateTime&) NLIB_NOEXCEPT { return true; }  // NOLINT
    typedef int native_handle_type;
    native_handle_type native_handle() NLIB_NOEXCEPT { return 0; }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(DummyCriticalSection);
};

class CriticalSectionFallback {
 public:
    // NOTE:
    // -1 must not be a thread-id
    CriticalSectionFallback() NLIB_NOEXCEPT : m_Counter(0) {
        m_Tid = -1;
    }
    ~CriticalSectionFallback() NLIB_NOEXCEPT {}
    NLIB_VIS_PUBLIC void lock() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC void unlock() NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool try_lock() NLIB_NOEXCEPT;
    typedef SimpleCriticalSection::native_handle_type native_handle_type;
    native_handle_type native_handle() { return m_Lock.native_handle(); }

 private:
    SimpleCriticalSection m_Lock;
    int32_t m_Counter;
    nlib_thread_id m_Tid;
};

#ifdef NLIB_CRITICALSECTION_USE_FALLBACK
typedef CriticalSectionFallback CriticalSection;
#endif

class TimedCriticalSectionFallback {
 public:
    TimedCriticalSectionFallback() NLIB_NOEXCEPT {}
    ~TimedCriticalSectionFallback() NLIB_NOEXCEPT {}
    void lock() NLIB_NOEXCEPT { m_Lock.lock(); }
    void unlock() NLIB_NOEXCEPT { m_Lock.unlock(); }
    bool try_lock() NLIB_NOEXCEPT { return m_Lock.try_lock(); }
    NLIB_VIS_PUBLIC bool try_lock_for(const TimeSpan& span) NLIB_NOEXCEPT;
    NLIB_VIS_PUBLIC bool try_lock_until(const DateTime& abstime) NLIB_NOEXCEPT;
    typedef CriticalSection::native_handle_type native_handle_type;
    native_handle_type native_handle() { return m_Lock.native_handle(); }

 private:
    CriticalSection m_Lock;
};

#ifdef NLIB_TIMEDCRITICALSECTION_USE_FALLBACK
typedef TimedCriticalSectionFallback TimedCriticalSection;
#endif

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_NAMESPACE_BEGIN
namespace threading {

struct AdoptLockType {};
struct TryToLockType {};
struct DeferLockType {};

NLIB_CEXPR const AdoptLockType adoptLock = {};
NLIB_CEXPR const TryToLockType tryToLock = {};
NLIB_CEXPR const DeferLockType deferLock = {};

template <class T>
class NLIB_SCOPED_LOCKABLE ScopedLock {
 public:
    typedef T mutex_type;
    typedef typename mutex_type::native_handle_type native_handle_type;
    explicit ScopedLock(mutex_type& m) NLIB_LOCK_FUNC(m_Lock) : m_Lock(m) {  // NOLINT
        m_Lock.lock();
    }
    ScopedLock(mutex_type& m, AdoptLockType) NLIB_NOEXCEPT NLIB_LOCK_FUNC(m_Lock)  // NOLINT
        : m_Lock(m) {}
    ~ScopedLock() NLIB_NOEXCEPT NLIB_UNLOCK_FUNC(m_Lock) { m_Lock.unlock(); }
    native_handle_type native_handle() { return m_Lock.native_handle(); }

 private:
    mutex_type& m_Lock;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedLock);
};

template <>
class NLIB_SCOPED_LOCKABLE ScopedLock<SimpleCriticalSection> {
 public:
    typedef SimpleCriticalSection mutex_type;
    typedef mutex_type::native_handle_type native_handle_type;
    explicit ScopedLock(mutex_type& m) NLIB_LOCK_FUNC(m_Lock) : m_Lock(m) {  // NOLINT
        m_Lock.lock();
    }
    ScopedLock(mutex_type& m, AdoptLockType) NLIB_NOEXCEPT NLIB_LOCK_REQUIRED(m)  // NOLINT
        NLIB_LOCK_FUNC(m_Lock)
        : m_Lock(m) {}
    ~ScopedLock() NLIB_NOEXCEPT NLIB_UNLOCK_FUNC(m_Lock) { m_Lock.unlock(); }
    native_handle_type native_handle() { return m_Lock.native_handle(); }

 private:
    mutex_type& m_Lock;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedLock);
};

template<>
class NLIB_SCOPED_LOCKABLE ScopedLock<nlib_mutex> {
 public:
    typedef nlib_mutex mutex_type;
    typedef nlib_mutex* native_handle_type;
    explicit ScopedLock(mutex_type& m) NLIB_LOCK_FUNC(m_Lock) : m_Lock(m) {  // NOLINT
        errno_t e = nlib_mutex_lock(&m_Lock);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    ScopedLock(mutex_type& m, AdoptLockType) NLIB_NOEXCEPT NLIB_LOCK_REQUIRED(m)  // NOLINT
        NLIB_LOCK_FUNC(m_Lock)
        : m_Lock(m) {}
    ~ScopedLock() NLIB_NOEXCEPT NLIB_UNLOCK_FUNC(m_Lock) {
        errno_t e = nlib_mutex_unlock(&m_Lock);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    native_handle_type native_handle() { return &m_Lock; }

 private:
    nlib_mutex& m_Lock;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedLock);
};

template <class T>
class UniqueLock {
 public:
    typedef T mutex_type;
    typedef typename mutex_type::native_handle_type native_handle_type;
    UniqueLock() NLIB_NOEXCEPT : m_Locker(NULL), m_Locked(false) {}
    explicit UniqueLock(mutex_type& rhs) : m_Locker(&rhs), m_Locked(true) {  // NOLINT
        rhs.lock();
    }
    UniqueLock(mutex_type& rhs, AdoptLockType)  // NOLINT
        : m_Locker(&rhs),
          m_Locked(true) {}
    UniqueLock(mutex_type& rhs, DeferLockType)  // NOLINT
        NLIB_NOEXCEPT : m_Locker(&rhs),
                        m_Locked(false) {}
    UniqueLock(mutex_type& rhs, TryToLockType)  // NOLINT
        : m_Locker(&rhs),
          m_Locked(rhs.try_lock()) {}
    NLIB_MOVE_MEMBER_HELPER_2(UniqueLock, m_Locker, m_Locked)
    ~UniqueLock() NLIB_NOEXCEPT {
        if (m_Locked) m_Locker->unlock();
    }
    void lock() {
        NLIB_ASSERT(m_Locker && !m_Locked);
        if (!m_Locker || m_Locked) {
            return;
        }
        m_Locker->lock();
        m_Locked = true;
    }
    void unlock() NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Locker && m_Locked);
        if (!m_Locker || !m_Locked) {
            return;
        }
        m_Locker->unlock();
        m_Locked = false;
    }
    bool try_lock() NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Locker && !m_Locked);
        if (!m_Locker || m_Locked) {
            return false;
        }
        m_Locked = m_Locker->try_lock();
        return m_Locked;
    }
    bool try_lock_for(const TimeSpan& timeout) NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Locker && !m_Locked);
        if (!m_Locker || m_Locked) {
            return false;
        }
        m_Locked = m_Locker->try_lock_for(timeout);
        return m_Locked;
    }
    bool try_lock_until(const DateTime& abstime) NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Locker && !m_Locked);
        if (!m_Locker || m_Locked) {
            return false;
        }
        m_Locked = m_Locker->try_lock_until(abstime);
        return m_Locked;
    }
    void swap(UniqueLock& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(m_Locker, rhs.m_Locker);
        swap(m_Locked, rhs.m_Locked);
    }
    mutex_type* release() NLIB_NOEXCEPT {
        mutex_type* p = m_Locker;
        m_Locker = NULL;
        m_Locked = false;
        return p;
    }
    mutex_type* mutex() const NLIB_NOEXCEPT { return m_Locker; }
    native_handle_type native_handle() { return m_Locker->native_handle(); }
    bool owns_lock() const NLIB_NOEXCEPT { return m_Locked; }
    NLIB_SAFE_BOOL(UniqueLock, owns_lock());

 private:
    mutex_type* m_Locker;
    bool m_Locked;
    NLIB_DISALLOW_COPY_AND_ASSIGN(UniqueLock);
};

namespace detail {

template<class T>
NLIB_ALWAYS_INLINE nlib_mutex* GetRawMutex(T& obj) {  // NOLINT
    return obj.native_handle();
}

template<>
NLIB_ALWAYS_INLINE nlib_mutex* GetRawMutex(nlib_mutex& obj) {  // NOLINT
    return &obj;
}

}  // namespace detail

}  // namespace threading
NLIB_NAMESPACE_END

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#endif

NLIB_DEFINE_STD_SWAP_T_BEGIN1(threading)  // NOLINT
NLIB_DEFINE_STD_SWAP_T1(T, NLIB_NS::threading::UniqueLock)  // NOLINT
NLIB_DEFINE_STD_SWAP_T_END1(threading)               // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#endif


#endif  // INCLUDE_NN_NLIB_THREADING_CRITICALSECTION_H_
