﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_
#define INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_

#include <new>

#include "nn/nlib/Swap.h"
#include "nn/nlib/UniquePtr.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

namespace internal {

class LspStatus {
 public:
    NLIB_CEXPR LspStatus() NLIB_NOEXCEPT : m_V(NULL) {}
    ~LspStatus() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_1(LspStatus, m_V)
    errno_t ToShare(LspStatus* ptr) NLIB_NOEXCEPT {
        if (!m_V) {
            m_V = new (std::nothrow) int32_t(2);
            if (!m_V) return ENOMEM;
        } else {
            nlib_atomic_add_fetch32(m_V, 1, NLIB_ATOMIC_ACQUIRE);
        }
        NLIB_ASSERT(!ptr->m_V);
        ptr->m_V = m_V;
        return 0;
    }
    bool ToUnshare() NLIB_NOEXCEPT {
        if (!m_V) return false;
        if (nlib_atomic_sub_fetch32(m_V, 1, NLIB_ATOMIC_ACQUIRE) == 0) {
            delete m_V;
            m_V = NULL;
            return false;
        }
        m_V = NULL;
        return true;
    }
    bool IsUnique() NLIB_NOEXCEPT {
        if (!m_V) return true;
        return nlib_atomic_load32(m_V, NLIB_ATOMIC_ACQUIRE) == 1;
    }
    void swap(LspStatus& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(m_V, rhs.m_V);
    }

 private:
    int32_t* m_V;
    NLIB_DISALLOW_COPY_AND_ASSIGN(LspStatus);
};

}  // namespace internal

// Use of shared_ptr is discouraged because of unexpected overhead,
// and You have to share the pointer by calling MakeSharedFrom() method.
template <class T, class DEL = DefaultDelete<T> >
class LimitedSharedPtr {
 public:
    LimitedSharedPtr() NLIB_NOEXCEPT : m_Ptr(NULL) {}
    ~LimitedSharedPtr() NLIB_NOEXCEPT { release(); }
    explicit LimitedSharedPtr(nullptr_t) NLIB_NOEXCEPT : m_Ptr(NULL) {}
    explicit LimitedSharedPtr(T* p) NLIB_NOEXCEPT : m_Ptr(p) {
        // Do not count up LspStatus because there is the risk of alloc failure.
    }
    NLIB_MOVE_MEMBER_HELPER_2(LimitedSharedPtr, m_Ptr, m_Status)
    void reset(T* p = 0) NLIB_NOEXCEPT { LimitedSharedPtr<T>(p).swap(*this); }
    T& operator*() const {
        NLIB_ASSERT(m_Ptr != NULL);
        return *m_Ptr;
    }
    T* operator->() const NLIB_NOEXCEPT {
        NLIB_ASSERT(m_Ptr != NULL);
        return m_Ptr;
    }
    T* get() const NLIB_NOEXCEPT { return m_Ptr; }
    errno_t MakeSharedFrom(const LimitedSharedPtr& p) NLIB_NOEXCEPT;  // NOLINT
    bool IsUnique() const NLIB_NOEXCEPT { return m_Status.IsUnique(); }
    bool MakeUniquePtr(UniquePtr<T>& p) NLIB_NOEXCEPT;  // NOLINT
    void swap(LimitedSharedPtr& rhs) NLIB_NOEXCEPT {    // NOLINT
        using std::swap;
        swap(m_Ptr, rhs.m_Ptr);
        m_Status.swap(rhs.m_Status);
    }
    LimitedSharedPtr& operator=(nullptr_t) NLIB_NOEXCEPT {
        release();
        return *this;
    }
    NLIB_SAFE_BOOL(LimitedSharedPtr, m_Ptr != NULL);

 private:
    void release() NLIB_NOEXCEPT {
        if (!m_Status.ToUnshare()) DEL()(m_Ptr);
        m_Ptr = NULL;
    }

 private:
    T* m_Ptr;
    mutable internal::LspStatus m_Status;
    //    NLIB_DISALLOW_COPY_AND_ASSIGN(LimitedSharedPtr);
};

template <class T, class DEL>
errno_t LimitedSharedPtr<T, DEL>::MakeSharedFrom(const LimitedSharedPtr& p)
    NLIB_NOEXCEPT {  // NOLINT
    LimitedSharedPtr tmp;
    if (p) {
        errno_t e = p.m_Status.ToShare(&tmp.m_Status);
        if (e != 0) return e;
        tmp.m_Ptr = p.m_Ptr;
    }
    tmp.swap(*this);
    return 0;
}

template <class T, class DEL>
bool LimitedSharedPtr<T, DEL>::MakeUniquePtr(UniquePtr<T>& p) NLIB_NOEXCEPT {  // NOLINT
    if (IsUnique()) {
        p.reset(m_Ptr);
        m_Ptr = NULL;
        release();
        return true;
    } else {
        return false;
    }
}

template <class T, class DEL>
class LimitedSharedPtr<T[], DEL> {
 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(LimitedSharedPtr);
};  // not allowed for now

/*
template <class T1, class D1, class T2, class D2>
bool operator==(const LimitedSharedPtr<T1, D1>& rhs, const LimitedSharedPtr<T2, D2>& lhs) {
    return rhs.get() == lhs.get();
}

template <class T1, class D1>
bool operator==(const LimitedSharedPtr<T1, D1>& lhs, nullptr_t) {
    return !lhs;
}

template <class T1, class D1>
bool operator==(nullptr_t, const LimitedSharedPtr<T1, D1>& rhs) {
    return !rhs;
}

template <class T1, class D1, class T2, class D2>
bool operator!=(const LimitedSharedPtr<T1, D1>& rhs, const LimitedSharedPtr<T2, D2>& lhs) {
    return rhs.get() != lhs.get();
}

template <class T1, class D1>
bool operator!=(const LimitedSharedPtr<T1, D1>& lhs, nullptr_t) {
    return static_cast<bool>(lhs);
}

template <class T1, class D1>
bool operator!=(nullptr_t, const LimitedSharedPtr<T1, D1>& rhs) {
    return static_cast<bool>(rhs);
}
*/

}  // namespace threading
NLIB_NAMESPACE_END

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#endif

NLIB_DEFINE_STD_SWAP_T_BEGIN1(threading)  // NOLINT
NLIB_DEFINE_STD_SWAP_T2(T, DEL, NLIB_NS::threading::LimitedSharedPtr)  // NOLINT
NLIB_DEFINE_STD_SWAP_T_END1(threading)               // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#endif

#endif  // INCLUDE_NN_NLIB_THREADING_LIMITEDSHAREDPTR_H_
