﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_SHAREDCRITICALSECTION_H_
#define INCLUDE_NN_NLIB_THREADING_SHAREDCRITICALSECTION_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/DateTime.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

class NLIB_LOCKABLE SharedCriticalSection NLIB_FINAL {
 public:
#ifdef NLIB_CXX11_CONSTEXPR
    constexpr SharedCriticalSection() NLIB_NOEXCEPT
        : rwock_(NLIB_RWLOCK_INITIALIZER) {}
#else
    SharedCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_rwlock_init(&rwock_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
#endif
    ~SharedCriticalSection() NLIB_NOEXCEPT {
        errno_t e = nlib_rwlock_destroy(&rwock_);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }

    void lock_shared() NLIB_NOEXCEPT NLIB_SHARED_LOCK_FUNC() {
        errno_t e = nlib_rwlock_rdlock(&rwock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    bool try_lock_shared() NLIB_NOEXCEPT NLIB_SHARED_TRYLOCK_FUNC(true) {
        errno_t e = nlib_rwlock_tryrdlock(&rwock_);
        NLIB_ASSERT(e == 0 || e == EBUSY);
        return e == 0;
    }
    bool try_lock_shared_for(const TimeSpan& timeout) NLIB_NOEXCEPT
        NLIB_SHARED_TRYLOCK_FUNC(true) {
        errno_t e = nlib_rwlock_tryrdlock_for(&rwock_, timeout.ToTimeValue().tick);
        NLIB_ASSERT(e == 0 || e == ETIMEDOUT);
        return e == 0;
    }
    bool try_lock_shared_until(const DateTime& abstime) NLIB_NOEXCEPT
        NLIB_SHARED_TRYLOCK_FUNC(true) {
        errno_t e = nlib_rwlock_tryrdlock_until(&rwock_, abstime.ToTimeValue().tick);
        NLIB_ASSERT(e == 0 || e == ETIMEDOUT);
        return e == 0;
    }
    void unlock_shared() NLIB_NOEXCEPT NLIB_SHARED_UNLOCK_FUNC() {
        errno_t e = nlib_rwlock_rdunlock(&rwock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }

    void lock() NLIB_NOEXCEPT NLIB_LOCK_FUNC() {
        errno_t e = nlib_rwlock_wrlock(&rwock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    bool try_lock() NLIB_NOEXCEPT NLIB_TRYLOCK_FUNC(true) {
        errno_t e = nlib_rwlock_trywrlock(&rwock_);
        NLIB_ASSERT(e == 0 || e == EBUSY);
        return e == 0;
    }
    bool try_lock_for(const TimeSpan& timeout) NLIB_NOEXCEPT NLIB_TRYLOCK_FUNC(true) {
        errno_t e = nlib_rwlock_trywrlock_for(&rwock_, timeout.ToTimeValue().tick);
        NLIB_ASSERT(e == 0 || e == ETIMEDOUT);
        return e == 0;
    }
    bool try_lock_until(const DateTime& abstime) NLIB_NOEXCEPT NLIB_TRYLOCK_FUNC(true) {
        errno_t e = nlib_rwlock_trywrlock_until(&rwock_, abstime.ToTimeValue().tick);
        NLIB_ASSERT(e == 0 || e == ETIMEDOUT);
        return e == 0;
    }
    void unlock() NLIB_NOEXCEPT NLIB_UNLOCK_FUNC() {
        errno_t e = nlib_rwlock_wrunlock(&rwock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }

 private:
    nlib_rwlock rwock_;
    friend class CondVarForSharedCriticalSection;
};

class CondVarForSharedCriticalSection NLIB_FINAL {
 public:
#ifdef NLIB_CXX11_CONSTEXPR
     constexpr CondVarForSharedCriticalSection() NLIB_NOEXCEPT
         : condrwlock_(NLIB_CONDRWLOCK_INITIALIZER) {}
#else
    CondVarForSharedCriticalSection() NLIB_NOEXCEPT {
         errno_t e = nlib_condrwlock_init(&condrwlock_);
         NLIB_ASSERT_NOERR(e);
         NLIB_UNUSED(e);
    }
#endif
    ~CondVarForSharedCriticalSection() NLIB_NOEXCEPT {
         errno_t e = nlib_condrwlock_destroy(&condrwlock_);
         NLIB_ASSERT_NOERR(e);
         NLIB_UNUSED(e);
    }
    void Notify() NLIB_NOEXCEPT {
        errno_t e = nlib_condrwlock_signal(&condrwlock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }
    void NotifyAll() NLIB_NOEXCEPT {
        errno_t e = nlib_condrwlock_broadcast(&condrwlock_);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
    }

    errno_t WaitReader(SharedCriticalSection& lock) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait(&condrwlock_, &lock.rwock_, 1);
    }
    errno_t WaitReader(nlib_rwlock& lock) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait(&condrwlock_, &lock, 1);
    }
    errno_t Wait(SharedCriticalSection& lock) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait(&condrwlock_, &lock.rwock_, 0);
    }
    errno_t Wait(nlib_rwlock& lock) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait(&condrwlock_, &lock, 0);
    }

    errno_t WaitReaderFor(SharedCriticalSection& lock, const TimeSpan& timeout) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_for(&condrwlock_, &lock.rwock_,
                                        timeout.ToTimeValue().tick, 1);
    }
    errno_t WaitReaderFor(nlib_rwlock& lock, const TimeSpan& timeout) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_for(&condrwlock_, &lock,
                                        timeout.ToTimeValue().tick, 1);
    }
    errno_t WaitFor(SharedCriticalSection& lock, const TimeSpan& timeout) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_for(&condrwlock_, &lock.rwock_,
                                        timeout.ToTimeValue().tick, 0);
    }
    errno_t WaitFor(nlib_rwlock& lock, const TimeSpan& timeout) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_for(&condrwlock_, &lock,
                                        timeout.ToTimeValue().tick, 0);
    }

    errno_t WaitReaderUntil(SharedCriticalSection& lock, const DateTime& datetime) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_until(&condrwlock_, &lock.rwock_,
                                          datetime.ToTimeValue().tick, 1);
    }
    errno_t WaitReaderUntil(nlib_rwlock& lock, const DateTime& datetime) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_until(&condrwlock_, &lock,
                                          datetime.ToTimeValue().tick, 1);
    }
    errno_t WaitUntil(SharedCriticalSection& lock, const DateTime& datetime) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_until(&condrwlock_, &lock.rwock_,
                                          datetime.ToTimeValue().tick, 0);
    }
    errno_t WaitUntil(nlib_rwlock& lock, const DateTime& datetime) NLIB_NOEXCEPT {  // NOLINT
        return nlib_condrwlock_wait_until(&condrwlock_, &lock,
                                          datetime.ToTimeValue().tick, 0);
    }

 private:
    nlib_condrwlock condrwlock_;
};


}  // namespace threading
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_THREADING_SHAREDCRITICALSECTION_H_
