﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_THREADPOOL_H_
#define INCLUDE_NN_NLIB_THREADING_THREADPOOL_H_

#include "nn/nlib/Nqueue.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/threading/Future.h"

NLIB_NAMESPACE_BEGIN
namespace threading {

namespace detail {
class NLIB_VIS_PUBLIC ThPoolTaskBase {
 public:
    ThPoolTaskBase() NLIB_NOEXCEPT {}
    virtual ~ThPoolTaskBase() NLIB_NOEXCEPT {}
    virtual void operator()() NLIB_NOEXCEPT = 0;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(ThPoolTaskBase);
};

template <class R>
class ThPoolTask : public ThPoolTaskBase {
 public:
    explicit ThPoolTask(PackagedTask<R()>& task) NLIB_NOEXCEPT {  // NOLINT
        task.swap(m_Task);
    }
    virtual ~ThPoolTask() NLIB_NOEXCEPT NLIB_OVERRIDE {}
    virtual void operator()() NLIB_NOEXCEPT NLIB_OVERRIDE { m_Task(); }

 private:
    PackagedTask<R()> m_Task;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ThPoolTask);
};
}  // namespace detail

class ThreadPool NLIB_FINAL {
 public:
    ThreadPool() NLIB_NOEXCEPT : m_Status(0) {}
    NLIB_VIS_PUBLIC errno_t Init(size_t numThreads, const ThreadSettings& settings) NLIB_NOEXCEPT;
    errno_t Init(size_t numThreads) NLIB_NOEXCEPT {
        ThreadSettings settings;
        return this->Init(numThreads, settings);
    }
    errno_t Init() NLIB_NOEXCEPT {
        return this->Init(GetHardwareConcurrency());
    }
    NLIB_VIS_PUBLIC ~ThreadPool() NLIB_NOEXCEPT;
    template <class R, class FUNC>
    errno_t Submit(Future<R>* future, FUNC& func) NLIB_NOEXCEPT;  // NOLINT
    size_t Count() NLIB_NOEXCEPT { return m_ThreadList.size(); }

 private:
    NLIB_VIS_HIDDEN void JoinThreads(Nlist<nlib_thread>& L) NLIB_NOEXCEPT;  // NOLINT
    static NLIB_VIS_HIDDEN void WorkerThread(void* p) NLIB_NOEXCEPT;

    int32_t m_Status;
    nlib_mutex m_Lock;
    Nlist<nlib_thread> m_ThreadList;
    Nqueue<UniquePtr<detail::ThPoolTaskBase> > m_WorkQueue;
    nlib_cond m_Cond;
    NLIB_DISALLOW_COPY_AND_ASSIGN(ThreadPool);
};

template <class R, class FUNC>
inline errno_t ThreadPool::Submit(Future<R>* future, FUNC& func) NLIB_NOEXCEPT {  // NOLINT
    PackagedTask<R()> task;
    errno_t e = task.Init(func);
    if (e != 0) return e;
    if (future) {
        e = task.GetFuture(future);
        if (e != 0) return e;
    }
    UniquePtr<detail::ThPoolTaskBase> pTask(new (std::nothrow) detail::ThPoolTask<R>(task));
    if (pTask) {
        e = nlib_mutex_lock(&m_Lock);
        NLIB_UNUSED(e);
        NLIB_ASSERT_NOERR(e);
        bool result = m_WorkQueue.push_back_swap(&pTask);
        if (result) {
            // only need to wake up one thread
            e = nlib_cond_signal(&m_Cond);
            NLIB_ASSERT_NOERR(e);
        }
        e = nlib_mutex_unlock(&m_Lock);
        NLIB_ASSERT_NOERR(e);
        return result ? 0 : ENOMEM;
    } else {
        return ENOMEM;
    }
}

}  // namespace threading
NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_THREADING_THREADPOOL_H_
