﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#include <vector>

#include "./xmlrpc_clientserver.h"
#include "./xmlrpc_value.h"
#include "nn/nlib/exi/XmlStreamWriter.h"
#include "nn/nlib/exi/XmlStreamReader.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/ReallocOutputStream.h"
#include "nn/nlib/MemoryInputStream.h"
#include "nn/nlib/MemoryOutputStream.h"

using nlib_ns::MemoryInputStream;
using nlib_ns::MemoryOutputStream;
using nlib_ns::ReallocOutputStream;
using nlib_ns::exi::ExiAllocator;
using nlib_ns::UniquePtr;
using nlib_ns::exi::XmlStreamReader;
using nlib_ns::exi::XmlStreamReaderSettings;
using nlib_ns::exi::XmlStreamWriter;
using nlib_ns::exi::XmlStreamWriterSettings;
using nlib_ns::exi::ExiChar;
using nlib_ns::StrCmp;

XmlRpcServerMethod::XmlRpcServerMethod(const StdString& name, XmlRpcServer* server)
    : m_Name(name), m_Server(server) {
    m_Server->AddServerMethod(this);
}

XmlRpcServerMethod::~XmlRpcServerMethod() { m_Server->RemoveServerMethod(this); }

namespace {
inline bool ReadEvent(XmlStreamReader* reader, XmlStreamReader::XmlStreamConstants e) {
    return reader->Next() == e;
}

#define SE \
    if (!ReadEvent(reader, XmlStreamReader::START_ELEMENT)) return false
#define EE \
    if (!ReadEvent(reader, XmlStreamReader::END_ELEMENT)) return false
#define CH \
    if (!ReadEvent(reader, XmlStreamReader::CHARACTERS)) return false

bool CheckSE(XmlStreamReader* reader, const ExiChar* name) {
    XmlStreamReader::XmlStreamConstants e = reader->Next();
    if (e != XmlStreamReader::START_ELEMENT) return false;
    return StrCmp(name, reader->GetLocalName()) == 0;
}

bool CheckEE(XmlStreamReader* reader, const ExiChar* /* name */) {
    XmlStreamReader::XmlStreamConstants e = reader->Next();
    if (e != XmlStreamReader::END_ELEMENT) return false;
    return true;
}

#define CHK_SE(name) \
    if (!CheckSE(reader, name)) return false
#define CHK_EE(name) \
    if (!CheckEE(reader, name)) return false
}  // namespace

bool XmlRpcServer::Process(XmlStreamReader* reader, XmlStreamWriter* writer) {
    XmlRpcValue params;
    const ExiChar* methodName;

    reader->Next();  // StartDocument
    CHK_SE(N("methodCall"));
    CHK_SE(N("methodName"));
    CH;
    methodName = reader->GetText();
    CHK_EE(N("methodName"));

    CHK_SE(N("params"));
    params.SetArray();
    XmlRpcArray* ar = params.AsArray();
    XmlStreamReader::XmlStreamConstants e;
    while ((e = reader->Next()) == XmlStreamReader::START_ELEMENT) {
        if (StrCmp(N("param"), reader->GetLocalName()) != 0) return false;
        XmlRpcValue* item = ar->Append();
        if (!item->Read(reader)) return false;
        CHK_EE(N("param"));
    }

    if (e != XmlStreamReader::END_ELEMENT) return false;

    CHK_EE(N("methodCall"));
    reader->Next();  // EndDocument
    reader->Close();

    XmlRpcValue result;
    XmlRpcServerMethod* method = this->FindServerMethod(methodName);
    if (!method) return false;
    method->Execute(params, &result);

    writer->WriteStartDocument();
    writer->WriteStartElement(N("methodResponse"));
    writer->WriteStartElement(N("params"));

    if (result.GetValueType() != XmlRpcValue::ARRAY) {
        writer->WriteStartElement(N("param"));
        result.Write(writer);
        writer->WriteEndElement();
    } else {
        XmlRpcArray* ar2 = result.AsArray();
        size_t n = ar2->Size();
        for (size_t i = 0; i < n; ++i) {
            writer->WriteStartElement(N("param"));
            ar2->Get(i)->Write(writer);
            writer->WriteEndElement();
        }
    }

    writer->WriteEndElement();  // params
    writer->WriteEndElement();  // methodResponse
    writer->WriteEndDocument();
    writer->Close();
    return writer->IsOk();
}

class UseNexi {
    unsigned char* m_Buf;

 public:
    UseNexi() {
        m_Buf = new (std::nothrow) unsigned char[1024 * 128];
        ExiAllocator::Initialize(m_Buf, 1024 * 128);
    }
    ~UseNexi() {
        ExiAllocator::Finalize();
        delete[] m_Buf;
    }

 private:
    UseNexi(const UseNexi&);
    UseNexi& operator=(const UseNexi&);
};

bool XmlRpcCallResult::GetResult(XmlRpcValue* result) {
    UseNexi useNexi;
    ReallocOutputStream os;

    // Instead of connecting to the server,
    // This sample uses g_XmlRpcServer.
    {
        MemoryInputStream is(&m_ToServer[0], m_ToServer.size());

        // client -> server stream
        UniquePtr<XmlStreamReader> r(XmlStreamReader::Create(&is));
        if (!r.get()) return false;

        // server -> client stream
        UniquePtr<XmlStreamWriter> w(XmlStreamWriter::Create(&os));
        if (!w.get()) return false;

        // Ignores HTTP headers or something like this.
        if (!m_Server->Process(r.get(), w.get())) return false;

        r.reset(NULL);
        w.reset(NULL);
    }

    ReallocOutputStream::UniquePtrType data;
    size_t data_size = os.Release(&data);

    MemoryInputStream is2(&data[0], data_size);
    UniquePtr<XmlStreamReader> r(XmlStreamReader::Create(&is2));
    if (!r.get()) return false;

    XmlStreamReader* reader = r.get();
    result->SetArray();
    XmlRpcArray* ar = result->AsArray();

    // XML in response into XmlRpcValue
    r->Next();  // StartDocument
    CHK_SE(N("methodResponse"));
    CHK_SE(N("params"));

    XmlStreamReader::XmlStreamConstants e;
    while ((e = reader->Next()) == XmlStreamReader::START_ELEMENT) {
        if (StrCmp(N("param"), reader->GetLocalName()) != 0) return false;

        XmlRpcValue* item = ar->Append();
        if (!item->Read(reader)) return false;

        CHK_EE(N("param"));
    }

    if (e != XmlStreamReader::END_ELEMENT) return false;

    CHK_EE(N("methodResponse"));
    r->Next();  // EndDocument
    r.reset(NULL);
    return true;
}

bool XmlRpcClient::Execute(const ExiChar* methodName, const XmlRpcValue& params,
                           XmlRpcCallResult* result) {
    UseNexi useNexi;

    ReallocOutputStream os;
    UniquePtr<XmlStreamWriter> w(XmlStreamWriter::Create(&os));
    if (!w.get()) {
        nlib_printf("FAILED\n");
        return false;
    }

    w->WriteStartDocument();

    w->WriteStartElement(N("methodCall"));
    w->WriteStartElement(N("methodName"));
    w->WriteCharacters(methodName);
    w->WriteEndElement();
    w->WriteStartElement(N("params"));
    w->Flush();

    if (params.AsArray() == NULL) {
        // only one parameter supported
        w->WriteStartElement(N("param"));
        params.Write(w.get());
        w->WriteEndElement();
    } else {
        const XmlRpcArray* ar = params.AsArray();
        size_t len = ar->Size();
        for (size_t i = 0; i < len; ++i) {
            w->WriteStartElement(N("param"));
            ar->Get(i)->Write(w.get());
            w->WriteEndElement();
        }
    }

    w->WriteEndElement();  // params
    w->WriteEndElement();  // methodCall

    w->WriteEndDocument();
    w->Close();
    if (w->IsError()) {
        nlib_printf("Error in XmlRpcClient::Execute\n");
        return false;
    }
    w.reset(NULL);

    ReallocOutputStream::UniquePtrType data;
    size_t data_size = os.Release(&data);
    result->SetData(std::vector<unsigned char>(data.get(), data.get() + data_size));
    result->SetServer(m_Server);
    return true;
}
