﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#ifndef SAMPLES_XML_RPC_XMLRPC_VALUE_H_  // NOLINT
#define SAMPLES_XML_RPC_XMLRPC_VALUE_H_

#include <map>
#include <string>
#include <vector>

#include "nn/nlib/exi/exi.h"

#define N(x) NLIB_EXI_LITERAL(x)
#define M(x) NLIB_EXI_UTF8(x)
typedef std::basic_string< ::nlib_ns::exi::ExiChar> StdString;

class XmlRpcArray;
class XmlRpcStruct;

class XmlRpcValue {
 public:
    enum ValueType {
        NONE = 0,
        INT,
        BOOLEAN,
        DOUBLE,
        STRING,
        NIL,
        // DATETIME and BASE64 are omitted
        // DATETIME,
        // BASE64
        ARRAY,
        STRUCT
    };

 private:
    union {
        int valueInt;
        bool valueBool;
        double valueDouble;
        StdString* valueString;
        XmlRpcArray* valueArray;
        XmlRpcStruct* valueStruct;
    } m_ValueHolder;
    ValueType m_ValueType;
    char _[7];

 public:
    XmlRpcValue() {
        m_ValueType = NONE;
        m_ValueHolder.valueInt = 0;
    }
    ~XmlRpcValue() { this->Invalidate(); }
    void SetInt(const int& rhs) {
        this->Invalidate();
        m_ValueType = INT;
        m_ValueHolder.valueInt = rhs;
    }
    void SetBool(const bool& rhs) {
        this->Invalidate();
        m_ValueType = BOOLEAN;
        m_ValueHolder.valueBool = rhs;
    }
    void SetDouble(const double& rhs) {
        this->Invalidate();
        m_ValueType = DOUBLE;
        m_ValueHolder.valueDouble = rhs;
    }
    void SetString(const StdString& rhs) {
        this->Invalidate();
        m_ValueType = STRING;
        m_ValueHolder.valueString = new StdString(rhs);
    }
    void SetString(const nlib_ns::exi::ExiChar* rhs) {
        this->Invalidate();
        m_ValueType = STRING;
        m_ValueHolder.valueString = new StdString(rhs);
    }
    void SetArray();
    void SetStruct();

 public:
    ValueType GetValueType() const { return m_ValueType; }

    int* AsInt() { return (m_ValueType == INT) ? &m_ValueHolder.valueInt : NULL; }
    const int* AsInt() const { return (m_ValueType == INT) ? &m_ValueHolder.valueInt : NULL; }

    bool* AsBool() { return (m_ValueType == BOOLEAN) ? &m_ValueHolder.valueBool : NULL; }
    const bool* AsBool() const {
        return (m_ValueType == BOOLEAN) ? &m_ValueHolder.valueBool : NULL;
    }

    double* AsDouble() { return (m_ValueType == DOUBLE) ? &m_ValueHolder.valueDouble : NULL; }
    const double* AsDouble() const {
        return (m_ValueType == DOUBLE) ? &m_ValueHolder.valueDouble : NULL;
    }

    StdString* AsString() { return (m_ValueType == STRING) ? m_ValueHolder.valueString : NULL; }
    const StdString* AsString() const {
        return (m_ValueType == STRING) ? m_ValueHolder.valueString : NULL;
    }

    XmlRpcArray* AsArray() { return (m_ValueType == ARRAY) ? m_ValueHolder.valueArray : NULL; }
    const XmlRpcArray* AsArray() const {
        return (m_ValueType == ARRAY) ? m_ValueHolder.valueArray : NULL;
    }

    XmlRpcStruct* AsStruct() { return (m_ValueType == STRUCT) ? m_ValueHolder.valueStruct : NULL; }
    const XmlRpcStruct* AsStruct() const {
        return (m_ValueType == STRUCT) ? m_ValueHolder.valueStruct : NULL;
    }

    bool IsValid() const { return m_ValueType != NONE; }
    bool IsNil() const { return m_ValueType == NIL; }
    bool Read(nlib_ns::exi::XmlStreamReader* reader) {
        if (nlib_ns::exi::XmlStreamReader::START_ELEMENT != reader->Next()) return false;
        return this->Read_(reader);
    }
    void Write(nlib_ns::exi::XmlStreamWriter* writer) const;

 private:
    XmlRpcValue(const XmlRpcValue& rhs);             // forbidden
    XmlRpcValue& operator=(const XmlRpcValue& rhs);  // forbidden
    void Invalidate();
    bool Read_(nlib_ns::exi::XmlStreamReader* reader);

    friend class XmlRpcArray;
    friend class XmlRpcStruct;
};

class XmlRpcArray {
    typedef std::vector<XmlRpcValue*> ArrayValue;
    ArrayValue m_ArrayValue;

 public:
    XmlRpcValue* Append() {
        XmlRpcValue* p = new XmlRpcValue();
        m_ArrayValue.push_back(p);
        return p;
    }
    std::size_t Size() const { return m_ArrayValue.size(); }
    XmlRpcValue* Get(std::size_t idx) {
        return idx < m_ArrayValue.size() ? m_ArrayValue[idx] : NULL;
    }
    const XmlRpcValue* Get(std::size_t idx) const {
        return idx < m_ArrayValue.size() ? m_ArrayValue[idx] : NULL;
    }
    XmlRpcValue* operator[](std::size_t idx) { return this->Get(idx); }
    const XmlRpcValue* operator[](std::size_t idx) const { return this->Get(idx); }
    bool Read(nlib_ns::exi::XmlStreamReader* reader);
    void Write(nlib_ns::exi::XmlStreamWriter* writer) const;

 private:
    XmlRpcArray() {}
    ~XmlRpcArray();
    XmlRpcArray(const XmlRpcArray& rhs);             // forbidden
    XmlRpcArray& operator=(const XmlRpcArray& rhs);  // forbidden

    friend class XmlRpcValue;
};

class XmlRpcStruct {
    typedef std::map<StdString, XmlRpcValue*> StructValue;
    StructValue m_StructValue;

 public:
    XmlRpcValue& operator[](const StdString& rhs);
    XmlRpcValue* Insert(const StdString& rhs) { return &operator[](rhs); }
    XmlRpcValue* Get(const StdString& rhs) const {
        StructValue::const_iterator it = m_StructValue.find(rhs);
        return it != m_StructValue.end() ? it->second : NULL;
    }
    bool HasMember(const StdString& rhs) const {
        return m_StructValue.end() != m_StructValue.find(rhs);
    }
    bool Read(nlib_ns::exi::XmlStreamReader* reader);
    void Write(nlib_ns::exi::XmlStreamWriter* writer) const;

 private:
    XmlRpcStruct() {}
    ~XmlRpcStruct();
    XmlRpcStruct(const XmlRpcStruct& rhs);             // forbidden
    XmlRpcStruct& operator=(const XmlRpcStruct& rhs);  // forbidden

    friend class XmlRpcValue;
};

#endif  // SAMPLES_XML_RPC_XMLRPC_VALUE_H_  // NOLINT
