﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#include <stdlib.h>

#include "nn/nlib/heap/NMalloc.h"
#include "nn/nlib/threading/Thread.h"
#include "nn/nlib/DateTime.h"

using ::nlib_ns::threading::Thread;

#if defined(NN_PLATFORM_CTR)
static const int kRepeatConst = 1;
#elif defined(NLIB_CAFE_PPC)
static const int kRepeatConst = 10;
#else
static const int kRepeatConst = 100;  // others
#endif

//
// alloc/free 8bytes regions
//
class malloc_8bytes {
public:
    malloc_8bytes(void* (*malloc_func)(size_t), void (*free_func)(void*))
        : m_Malloc(malloc_func), m_Free(free_func) {}
    bool operator()() {
        bool success = true;
        for (int loop_cnt = 0; loop_cnt < 50; ++loop_cnt) {
            const size_t n = kRepeatConst * 1000;
            void** ptrs = (void**)m_Malloc(n * sizeof(void*));  // NOLINT
            if (!ptrs) return false;
            memset(ptrs, 0, n * sizeof(void*));  // NOLINT

            for (size_t i = 0; i < n; ++i) {
                ptrs[i] = m_Malloc(8);
                if (!ptrs[i]) {
                    success = false;
                    break;
                }
            }
            for (size_t i = 0; i < n; ++i) {
                m_Free(ptrs[i]);
            }
            m_Free(ptrs);
            if (!success) return false;
        }
        return true;
    }

private:
    void* (*m_Malloc)(size_t);
    void (*m_Free)(void*);
};

//
// it causes severe fragmentation if naive malloc/free is used
//
class malloc_fragmentation {
public:
    malloc_fragmentation(void* (*malloc_func)(size_t), void (*free_func)(void*))
        : m_Malloc(malloc_func), m_Free(free_func) {}
    bool operator()() {
        const int nloop = kRepeatConst * 30;
        bool success = true;
        for (int loop_cnt = 0; loop_cnt < nloop; ++loop_cnt) {
            const size_t n = 10000;
            void** ptrs = (void**)m_Malloc(n * sizeof(void*));  // NOLINT
            if (!ptrs) return false;
            memset(ptrs, 0, n * sizeof(void*));  // NOLINT

            for (size_t i = 0; i < n; ++i) {
                ptrs[i] = m_Malloc(8);
                if (!ptrs[i]) {
                    success = false;
                    break;
                }
            }
            for (size_t i = 0; i < n; i += 2) {
                m_Free(ptrs[i]);
                ptrs[i] = NULL;
            }
            for (size_t i = 0; i < n; i += 2) {
                ptrs[i] = m_Malloc(16);
                if (!ptrs[i]) {
                    success = false;
                    break;
                }
            }
            for (size_t i = 0; i < n; ++i) {
                m_Free(ptrs[i]);
            }
            m_Free(ptrs);
            if (!success) return false;
        }
        return true;
    }

private:
    void* (*m_Malloc)(size_t);
    void (*m_Free)(void*);
};

//
// realloc from 1 byte to 100000 bytes
//
class malloc_realloc {
public:
    malloc_realloc(void* (*realloc_func)(void*, size_t), void (*free_func)(void*))
        : m_Realloc(realloc_func), m_Free(free_func) {}
    bool operator()() {
        for (int loop_cnt = 0; loop_cnt < kRepeatConst; ++loop_cnt) {
            const size_t n = 100000;
            void* p = NULL;
            for (size_t i = 1; i < n; ++i) {
                void* pnew = m_Realloc(p, i);
                if (!pnew) {
                    m_Free(p);
                    return false;
                }
                p = pnew;
            }
            m_Free(p);
        }
        return true;
    }

private:
    void* (*m_Realloc)(void*, size_t);
    void (*m_Free)(void*);
};

const int kNumThread = 10;
Thread g_Th[kNumThread];
volatile bool g_Success;

class mt_smallmem {
public:
    mt_smallmem(void* malloc_func(size_t), void free_func(void*))
        : m_Malloc(malloc_func), m_Free(free_func) {}
    void operator()() {
        const int n = 1000;
        void* p[n];
        for (int j = 0; j < kRepeatConst * 10; ++j) {
            for (int i = 0; i < n; ++i) {
                p[i] = m_Malloc(8);
                if (!p[i]) g_Success = false;
            }
            for (int i = 0; i < n; ++i) {
                m_Free(p[i]);
            }
        }
    }

private:
    void* (*m_Malloc)(size_t);
    void (*m_Free)(void*);
};

class mt_largemem {
public:
    mt_largemem(void* malloc_func(size_t), void free_func(void*))
        : m_Malloc(malloc_func), m_Free(free_func) {}
    void operator()() {
        const int n = 40;
        void* p[n];
        for (int j = 0; j < kRepeatConst * 10; ++j) {
            for (int i = 0; i < n; ++i) {
                p[i] = m_Malloc(16384);
                if (!p[i]) g_Success = false;
            }
            for (int i = 0; i < n; ++i) {
                m_Free(p[i]);
            }
        }
    }

private:
    void* (*m_Malloc)(size_t);
    void (*m_Free)(void*);
};

//
// multithread performance
//
template<class FUNC>
class malloc_mt {
public:
    explicit malloc_mt(const FUNC& obj) : m_Obj(obj) {}
    bool operator()() {
        g_Success = true;
        for (int i = 0; i < kNumThread; ++i) {
            g_Th[i].Start(m_Obj);
        }
        for (int i = 0; i < kNumThread; ++i) {
            g_Th[i].Join();
        }
        return g_Success;
    }
private:
    FUNC m_Obj;
};

//
// lock / unock mutex
//
static nlib_mutex g_Mutex = NLIB_MUTEX_INITIALIZER;
class mutex_loop {
public:
    bool operator()() {
        for (int i = 0; i < kRepeatConst * 50000; ++i) {
            nlib_mutex_lock(&g_Mutex);
            nlib_mutex_unlock(&g_Mutex);
        }
        return true;
    }
};

static void* g_Ptr = NULL;
class nmalloc_loop {
public:
    bool operator()() {
        for (int i = 0; i < kRepeatConst * 50000; ++i) {
            g_Ptr = nmalloc(8);
            nfree(g_Ptr);
        }
        return true;
    }
};

template<class FUNC>
class do_test {
public:
    do_test(const char* title, const FUNC& func) : m_Title(title), m_Func(func) {}
    bool operator()() {
        uint64_t from = ::nlib_ns::GetTickTime();
        bool result = m_Func();
        uint64_t to = ::nlib_ns::GetTickTime();
        nlib_printf("%s: %" PRIu64 " msec\n", m_Title, to - from);
        return result;
    }
private:
    const char* m_Title;
    FUNC m_Func;
    NLIB_DISALLOW_COPY_AND_ASSIGN(do_test);
};

static bool clear_cache() {
    // free thread-cache nmalloc has, returning them to CentralHeap.
    nmalloc_finalize_tls();
    if (!nmalloc_isclean()) {
        nmalloc_dumpex(NMALLOC_DUMP_ALL);
        return false;
    }
    return true;
}

static bool compare_speed() {
    bool result;
    //
    // 8 bytes alloc/free
    //
    result = do_test<malloc_8bytes>("malloc: malloc_8bytes() loop",
                                    malloc_8bytes(malloc, free))();
    if (!result) return false;
    result = do_test<malloc_8bytes>("nmalloc: malloc_8bytes() loop",
                                    malloc_8bytes(nmalloc, nfree))();
    if (!result) return false;
    if (!clear_cache()) return false;

    //
    // fragmentation
    //
    result = do_test<malloc_fragmentation>("malloc: malloc_fragmentation() loop",
                                           malloc_fragmentation(malloc, free))();
    if (!result) return false;
    result = do_test<malloc_fragmentation>("nmalloc: malloc_fragmentation() loop",
                                           malloc_fragmentation(nmalloc, nfree))();
    if (!result) return false;
    if (!clear_cache()) return false;

    //
    // realloc
    //
    result = do_test<malloc_realloc>("malloc: realloc() loop",
                                     malloc_realloc(realloc, free))();
    if (!result) return false;
    result = do_test<malloc_realloc>("nmalloc: realloc() loop",
                                     malloc_realloc(nrealloc, nfree))();
    if (!result) return false;
    if (!clear_cache()) return false;

    //
    // multi-thread malloc/free small mem
    //
    result = do_test<malloc_mt<mt_smallmem> >(
        "malloc: mt-small loop",
        malloc_mt<mt_smallmem>(mt_smallmem(malloc, free)))();
    if (!result) return false;
    result = do_test<malloc_mt<mt_smallmem> >(
        "nmalloc: mt-small loop",
        malloc_mt<mt_smallmem>(mt_smallmem(nmalloc, nfree)))();
    if (!result) return false;
    if (!clear_cache()) return false;

    //
    // multi-thread malloc/free big mem
    //
    result = do_test<malloc_mt<mt_largemem> >(
        "malloc: mt-large loop",
        malloc_mt<mt_largemem>(mt_largemem(malloc, free)))();
    if (!result) return false;
    result = do_test<malloc_mt<mt_largemem> >(
        "nmalloc: mt-large loop",
        malloc_mt<mt_largemem>(mt_largemem(nmalloc, nfree)))();
    if (!result) return false;
    if (!clear_cache()) return false;

    //
    // compare mutex_lock/mutex_unlock loop and nmalloc/nfree loop
    //
    result = do_test<mutex_loop>("nlib_mutex: lock/unlock loop", mutex_loop())();
    result = do_test<nmalloc_loop>("nmalloc: nmalloc/nfree loop", nmalloc_loop())();
    return true;
}

#ifdef NLIB_HAS_VIRTUALMEMORY
extern "C" void nmalloc_get_settings(NMallocSettings* settings) {
    settings->addr = NULL;
    settings->size = 1024 * 1024 * 10;
    settings->heap_option = 0;
}
#else
const size_t heapmem_size = 1024 * 1024 * 10;
NLIB_ALIGNAS(4096) static char heapmem[heapmem_size];
extern "C" void nmalloc_get_settings(NMallocSettings* settings) {
    settings->addr = heapmem;
    settings->size = heapmem_size;
    settings->heap_option = 0;
}
#endif

static bool SampleMain(int, char**) { return compare_speed(); }

NLIB_MAINFUNC
