﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#include "nn/nlib/threading/CondVar.h"
#include "nn/nlib/threading/Thread.h"
#include "nn/nlib/threading/CriticalSection.h"

using ::nlib_ns::threading::ScopedLock;
using ::nlib_ns::threading::UniqueLock;
using ::nlib_ns::threading::CondVar;
using ::nlib_ns::threading::SimpleCriticalSection;
using ::nlib_ns::threading::Thread;

const int NUM_THREADS = 6;  // NUM_THREADS must be even
Thread g_Th[NUM_THREADS];

struct CondVarState {
    CondVar cond;
    SimpleCriticalSection lock;
    bool flag;

 public:
    CondVarState() : flag(false) {}
} g_Cond1, g_Cond2;

class ThreadParam {
 public:
    ThreadParam() : m_ThreadNum(0), m_Status(false), m_Wait(NULL), m_Signal(NULL) {}
    void Initialize(int num, CondVarState* w, CondVarState* s) {
        m_ThreadNum = num;
        m_Wait = w;
        m_Signal = s;
        m_Status = true;
    }
    void Exec();
    static void ThreadFunc(ThreadParam* param) { param->Exec(); }
    bool IsSuccessful() { return m_Status; }

 private:
    int m_ThreadNum;
    bool m_Status;
    CondVarState* m_Wait;
    CondVarState* m_Signal;
};

void ThreadParam::Exec() {
    {
        UniqueLock<SimpleCriticalSection> lock(m_Wait->lock);

        // Note that the thread is already notified at this point.
        while (!m_Wait->flag) {
            nlib_printf("  Thread %d waiting\n", m_ThreadNum);

            // releases 'lock' and waits notification
            errno_t e = m_Wait->cond.Wait(lock);
            if (e != 0) {
                m_Status = false;
                return;
            }
            // 'lock' is acquired when it returns from Wait(lock)
            nlib_printf("  Thread %d may be notified and acquires lock\n", m_ThreadNum);
        }
        nlib_printf("   Thread %d is notified, and do its own job\n", m_ThreadNum);
        m_Wait->flag = false;  // turn off the flag, note that 'lock' is acquired
    }

    {
        ScopedLock<SimpleCriticalSection> lock(m_Signal->lock);

        // you have to set m_Signal->flag within the criticalsection
        m_Signal->flag = true;
        nlib_printf("   Thread %d notifies\n", m_ThreadNum);
        m_Signal->cond.NotifyAll();
    }
}

bool CondVarDemo() {
    errno_t e;
    ThreadParam param[NUM_THREADS];

    for (int i = 0; i < NUM_THREADS; ++i) {
        if (i % 2 == 0) {
            param[i].Initialize(i, &g_Cond1, &g_Cond2);
            e = g_Th[i].Start(ThreadParam::ThreadFunc, &param[i]);
        } else {
            param[i].Initialize(i, &g_Cond2, &g_Cond1);
            e = g_Th[i].Start(ThreadParam::ThreadFunc, &param[i]);
        }
        if (e != 0) return false;
    }

    g_Cond1.lock.lock();
    nlib_printf(" Main thread notifies\n");

    // At this point, the threads may have already started to wait or not.
    g_Cond1.flag = true;
    g_Cond1.cond.NotifyAll();
    g_Cond1.lock.unlock();

    for (int i = 0; i < NUM_THREADS; ++i) {
        e = g_Th[i].Join();
        if (e != 0) return false;
    }
    for (int i = 0; i < NUM_THREADS; ++i) {
        if (!param[i].IsSuccessful()) return false;
    }
    return true;
}

static bool SampleMain(int, char**) { return CondVarDemo(); }

NLIB_MAINFUNC
