﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2015 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#include <algorithm>
#include <map>
#include <string>
#include <vector>

#include "jsonrpc.h"  // NOLINT

#include "nn/nlib/DateTime.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/threading/Thread.h"
#include "nn/nlib/msgpack/MpObject.h"
#include "nn/nlib/msgpack/jsonrpc/JsonRpcServerExec.h"

#ifndef DONT_USE_SOCKET
#include "nn/nlib/TcpInputStream.h"
#include "nn/nlib/TcpOutputStream.h"
#include "nn/nlib/TextWriter.h"
#endif

using nlib_ns::DateTime;
using nlib_ns::ReallocOutputStream;
using nlib_ns::UniquePtr;
using nlib_ns::msgpack::MpObject;
using nlib_ns::msgpack::MpObjectKv;

using nlib_ns::msgpack::jsonrpc::JsonRpcServerExec;
using nlib_ns::msgpack::jsonrpc::JsonRpcServerFuncCallError;
using nlib_ns::msgpack::jsonrpc::JsonRpcResponse;
using nlib_ns::msgpack::jsonrpc::reqid_t;
using nlib_ns::msgpack::jsonrpc::JSONSERVER_OK;
using nlib_ns::msgpack::jsonrpc::JSONSERVER_INVALID_PARAMS;
using nlib_ns::msgpack::jsonrpc::JSONSERVER_INTERNAL_ERROR;

namespace {

#ifdef DONT_USE_SOCKET
class ServerEndByMemory {
 public:
    errno_t Init() { return 0; }
    errno_t Accept() { return 0; }
    errno_t Send(nlib_ns::ReallocOutputStream::UniquePtrType* data, size_t n) {
        g_s2c.Send(data, n);
        return 0;
    }
    errno_t Receive(nlib_ns::ReallocOutputStream::UniquePtrType* data, size_t* n) {
        while (!g_c2s.Receive(data, n)) {
            if (g_ShutdownServer) {
                *n = 0;
                data->reset();
                return 0;
            }
            nlib_ns::threading::this_thread::YieldThread();
        }
        return 0;
    }
    void Close() {}
};
#else
class ServerEndByTcp {
 public:
    errno_t Init() {
        errno_t e;
        e = nlib_socket(&m_Acceptor, NLIB_AF_INET, NLIB_SOCK_STREAM, 0);
        if (e != 0) return e;
        int yes = 1;
        e = nlib_setsockopt(m_Acceptor, SOL_SOCKET, NLIB_SO_REUSEADDR, &yes, sizeof(yes));
        if (e != 0) return e;

        nlib_sockaddr_in addr;
        addr.sin_family = NLIB_AF_INET;
        addr.sin_port = nlib_htons(NLIB_SOCKPORT_SAMPLE);
        addr.sin_addr.s_addr = nlib_htonl(NLIB_INADDR_ANY);
        e = nlib_bind(m_Acceptor, reinterpret_cast<nlib_sockaddr*>(&addr), sizeof(addr));
        if (e != 0) return e;

        e = nlib_listen(m_Acceptor, 1);
        return e;
    }
    errno_t Accept() {
        errno_t e = nlib_accept(&m_Socket, m_Acceptor, NULL, NULL);
        if (e != 0) return e;
        m_Is.Init(m_Socket);
        return 0;
    }

    // This sample sends and receives the data in netstring method.
    // you can also close a socket every time.
    // JSON-RPC does not refer to them.
    errno_t Send(nlib_ns::ReallocOutputStream::UniquePtrType* data, size_t n) {
        using nlib_ns::threading::SimpleCriticalSection;
        using nlib_ns::threading::ScopedLock;

        ScopedLock<SimpleCriticalSection> l(m_LockS);
        nlib_ns::TcpOutputStream os;
        os.Init(m_Socket);
        {
            nlib_ns::TextWriter writer;
            writer.Init(&os);
            writer.WriteFormat("%" PRIuS ":", n);
        }
        os.Write(data->get(), n);
        os.Write(',');
        os.Flush();
        return os.IsOk() ? 0 : os.GetErrorValue();
    }
    errno_t Receive(nlib_ns::ReallocOutputStream::UniquePtrType* data, size_t* n) {
        using nlib_ns::threading::SimpleCriticalSection;
        using nlib_ns::threading::ScopedLock;
        ScopedLock<SimpleCriticalSection> l(m_LockR);
        if (m_Is.Peek() < 0) {
            if (!m_Is.IsOk()) return EIO;
            *n = 0;
            return 0;
        }
        int c;
        size_t nBytes = 0;
        for (;;) {
            c = m_Is.Read();
            if (c == ':') break;
            if (!nlib_ns::IsDigit(c)) return EILSEQ;
            size_t newVal = nBytes * 10 + c - '0';
            if (newVal < nBytes) return EILSEQ;
            nBytes = newVal;
        }
        data->reset(reinterpret_cast<uint8_t*>(nlib_malloc(nBytes)));
        if (!*data) return ENOMEM;
        if (nBytes != m_Is.Read(data->get(), nBytes)) return EILSEQ;
        if (m_Is.Read() != ',') return EILSEQ;
        *n = nBytes;
        return 0;
    }
    void Close() {
        nlib_closesocket(m_Socket);
        nlib_closesocket(m_Acceptor);
    }

 private:
    nlib_sock m_Acceptor;
    nlib_sock m_Socket;
    nlib_ns::TcpInputStream m_Is;
    nlib_ns::threading::SimpleCriticalSection m_LockS;
    nlib_ns::threading::SimpleCriticalSection m_LockR;
};
#endif

#ifndef DONT_USE_SOCKET
typedef ServerEndByTcp ServerEnd;
#else
typedef ServerEndByMemory ServerEnd;
#endif
UniquePtr<ServerEnd> g_ServerEnd;

int g_Counter = 0;

struct Msg {
    std::string uri;
    DateTime datetime;
    std::string text;

 public:
    Msg(DateTime datetime_, const char* text_) : datetime(datetime_), text(text_) {
        char buf[128];
        ::nlib_ns::SnPrintf(buf, "http://example.com/board/%d", g_Counter);
        ++g_Counter;
        uri = buf;
    }
    bool operator<(const Msg& rhs) const { return datetime < rhs.datetime; }
    bool ToMpObject(MpObject* obj) {
        // { "datetime": ...., "uri": ...,  "text": .... }
        char dt[32];
        datetime.ToW3cDtf(dt);
        std::map<std::string, std::string> m;
        m["datetime"] = dt;
        m["uri"] = uri;
        m["text"] = text;
        MpObject tmp;
        if (tmp.Box(m) != 0) return false;
        tmp.swap(*obj);
        return true;
    }
};

class MsgDb {
 public:
    ~MsgDb() {
        std::vector<Msg*>::iterator it = m_Data.begin();
        std::vector<Msg*>::iterator end = m_Data.end();
        for (; it != end; ++it) {
            delete *it;
        }
    }

 private:
    Msg* FindByUri(const char* uri) {
        std::string uristr(uri);
        std::vector<Msg*>::iterator it = m_Data.begin();
        std::vector<Msg*>::iterator end = m_Data.end();
        for (; it != end; ++it) {
            if ((*it)->uri == uristr) {
                return *it;
            }
        }
        return NULL;
    }
    bool DeleteByUri(const char* uri) {
        std::string uristr(uri);
        std::vector<Msg*>::iterator it = m_Data.begin();
        std::vector<Msg*>::iterator end = m_Data.end();
        for (; it != end; ++it) {
            if ((*it)->uri == uristr) {
                delete *it;
                m_Data.erase(it);
                return true;
            }
        }
        return false;
    }
    static bool LessFunc(Msg* lhs, Msg* rhs) { return *lhs < *rhs; }

 public:
    std::string Create(DateTime datetime, const char* text) {
        m_Data.push_back(new Msg(datetime, text));
        std::string rval = m_Data.back()->uri;
        std::sort(m_Data.begin(), m_Data.end(), LessFunc);
        return rval;
    }
    bool Read(MpObject* obj) const {
        uint32_t n = static_cast<uint32_t>(m_Data.size());
        errno_t e = obj->InitArray(n);
        if (e != 0) return false;
        for (uint32_t i = 0; i < n; ++i) {
            m_Data[i]->ToMpObject(obj->GetArrayItem(i));
        }
        return true;
    }
    bool Update(const std::string& uri, DateTime datetime, const std::string& text) {
        Msg* msg = FindByUri(uri.c_str());
        if (!msg) return false;
        msg->uri = uri;
        msg->datetime = datetime;
        msg->text = text;
        std::sort(m_Data.begin(), m_Data.end(), LessFunc);
        return true;
    }
    bool Delete(const char* uri) { return DeleteByUri(uri); }

 private:
    std::vector<Msg*> m_Data;
} g_MsgDb;

bool GetDateTimeField(const MpObject& param, DateTime* dt) {
    const MpObject* datetimeObj = param.GetMapItem("datetime");
    if (!datetimeObj) return false;
    std::string dtstr;
    if (datetimeObj->Unbox(&dtstr) != 0) return false;
    nlib_ns::TimeSpan span;
    if (DateTime::Parse(dtstr.c_str(), dt, &span) != 0) return false;
    return true;
}

bool GetTextField(const MpObject& param, std::string* text) {
    const MpObject* txtObj = param.GetMapItem("text");
    if (!txtObj) return false;
    if (txtObj->Unbox(text) != 0) return false;
    return true;
}

bool GetUriField(const MpObject& param, std::string* uri) {
    const MpObject* uriObj = param.GetMapItem("uri");
    if (!uriObj) return false;
    if (uriObj->Unbox(uri) != 0) return false;
    return true;
}

JsonRpcServerFuncCallError Create(MpObject& param, JsonRpcResponse& response) {  // NOLINT
    // STRING create({datetime:STRING, text:STRING});
    DateTime dt;
    if (!GetDateTimeField(param, &dt)) return JSONSERVER_INVALID_PARAMS;
    std::string text;
    if (!GetTextField(param, &text)) return JSONSERVER_INVALID_PARAMS;

    // Creates a message and receives URI as the key for it.
    std::string uri = g_MsgDb.Create(dt, text.c_str());
    MpObject obj;
    if (obj.Box(uri) != 0) return JSONSERVER_INTERNAL_ERROR;

    response.MoveResultFrom(obj);

    // You can make the client timeout by delaying the response.
    // ::nlib_ns::threading::this_thread::SleepMilliSeconds(500);
    return JSONSERVER_OK;
}

JsonRpcServerFuncCallError Read(MpObject&, JsonRpcResponse& response) {  // NOLINT
    // [{uri:STRING, datetime:STRING, text:STRING}, ...] read()

    // This function simplly responds all the messages which the server stores.
    // You can implement extended version of this function
    // by analyzing the parameters in MpObject.
    MpObject obj;
    if (!g_MsgDb.Read(&obj)) return JSONSERVER_INTERNAL_ERROR;

    response.MoveResultFrom(obj);
    return JSONSERVER_OK;
}

JsonRpcServerFuncCallError Update(MpObject& param, JsonRpcResponse& response) {  // NOLINT
    // void update([{uri:STRING, datetime:STRING, text:STRING}, ...])
    std::string uri;
    if (!GetUriField(param, &uri)) return JSONSERVER_INVALID_PARAMS;
    std::string text;
    if (!GetTextField(param, &text)) return JSONSERVER_INVALID_PARAMS;
    DateTime dt;
    if (!GetDateTimeField(param, &dt)) return JSONSERVER_INVALID_PARAMS;
    if (!g_MsgDb.Update(uri, dt, text)) {
        // Returns JSONSERVER_OK if you set non-predefined error.
        response.SetError(-32000, "uri not found");
    }
    return JSONSERVER_OK;
}

JsonRpcServerFuncCallError Delete(MpObject& param, JsonRpcResponse& response) {  // NOLINT
    // void delete({uri:STRING})
    std::string uritext;
    if (!GetUriField(param, &uritext)) return JSONSERVER_INVALID_PARAMS;
    if (!g_MsgDb.Delete(uritext.c_str())) {
        // Returns JSONSERVER_OK if you set non-predefined error.
        response.SetError(-32000, "uri not found");
    }
    return JSONSERVER_OK;
}

void ProcessRequest() {
    ReallocOutputStream::UniquePtrType reqBytes;
    size_t n = 0;
    ReallocOutputStream::UniquePtrType responseBytes;
    errno_t e = g_ServerEnd->Receive(&reqBytes, &n);
    if (e != 0 || n == 0) return;
    size_t rn = JsonRpcServerExec(reqBytes.get(), n, &responseBytes);
    g_ServerEnd->Send(&responseBytes, rn);
}

}  // namespace

errno_t InitServer() {
    g_ServerEnd.reset(new ServerEnd());
    if (!g_ServerEnd) {
        return ENOMEM;
    }
    errno_t e;
    e = g_ServerEnd->Init();
    if (e != 0) {
        nlib_printf("SERVER ERROR: %d\n", e);
        return e;
    }
    using nlib_ns::msgpack::jsonrpc::AddJsonRpcServerMethod;
    e = AddJsonRpcServerMethod("create", Create);
    if (e != 0) {
        nlib_printf("SERVER ERROR: %d\n", e);
        return e;
    }
    e = AddJsonRpcServerMethod("read", Read);
    if (e != 0) {
        nlib_printf("SERVER ERROR: %d\n", e);
        return e;
    }
    e = AddJsonRpcServerMethod("update", Update);
    if (e != 0) {
        nlib_printf("SERVER ERROR: %d\n", e);
        return e;
    }
    e = AddJsonRpcServerMethod("delete", Delete);
    if (e != 0) {
        nlib_printf("SERVER ERROR: %d\n", e);
        return e;
    }
    return 0;
}

volatile bool g_ShutdownServer = false;
void ServerThread() {
    errno_t e;
    // NOTE:
    // This sample uses single connection.
    e = g_ServerEnd->Accept();
    if (e != 0) {
        nlib_printf("SERVER ERROR: %d\n", e);
        return;
    }
    for (;;) {
        ProcessRequest();
        if (g_ShutdownServer) break;
    }
    g_ServerEnd->Close();
    g_ServerEnd.reset();
    nlib_printf("ServerThread() end\n");
}
